function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_Brockett(B, N, method, fns, params)
% driver for Brockett cost function on the Stiefel manifold.
% X^* = arg min_{X in St(p, n)} trace(X^T B X N), where B is an n by n matrix.
%
% INPUT:
% B: the symmetric matrix in the cost function
% N: the diagonal matrix in the cost function
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang

    fns.f = @(x)f(x, B, N);
    if(params.retraction == 2 || params.retraction == 4 || params.retraction == 6)
        fprintf('gradient of test function: stiefel manifold is considered as embedded space\n')
        fns.gf = @(x)gf_embedded(x, B, N);
        fns.hessianv = @(x, v)hessianv_newton_embedded(x, v, B, N);
        if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
            fns.f = @(x)f_intr(x, B, N);
            fns.gf = @(x)gf_embedded_intr(x, B, N);
            fns.hessianv = @(x, v)hessianv_newton_embedded_intr(x, v, B, N);
        end
    else
        fprintf('gradient of quotient function: stiefel manifold is considered as quotient space\n')
        fns.gf = @(x)gf_quotient(x, B, N);
        fns.hessianv = @(x, v)hessianv_newton_quotient(x, v, B, N);
        if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
            fns.f = @(x)f_intr(x, B, N);
            fns.gf = @(x)gf_quotient_intr(x, B, N);
            fns.hessianv = @(x, v)hessianv_newton_quotient_intr(x, v, B, N);
        end
    end
    fns.gf_exists = @(x)gf_exists(x, B, N);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RWRBFGS(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    elseif(method == 13)
        [X, F, G, timecost, nf, ng, T, nR] = check_grad_Hess(fns, params);
    end
end

function [output, x] = f(x, B, N)
    temp = B * (x.U * N);
    output = sum(sum(x.U .* temp));
    x.BUN = temp;
end

function [output, x] = f_intr(x, B, N)
    temp = B * (x.U * N);
    output = sum(sum(x.U .* temp));
    x.BUN = temp;
end

function [output, x] = gf_embedded(x, B, N)  % gradient of f(x)
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp = x.UtQ;
    
    output.TV = (Q - x.U * (temp + temp') / 2);
end

function [output, x] = gf_embedded_intr(x, B, N)  % gradient of f(x)
    [n, p] = size(x.U);
%     Q = 2 * B * (x1 * N);
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    K = x.Up' * Q;
    omega = 0.5 * x.U' * Q;
    omega = omega - omega';
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(p, p), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);%nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function [output, x] = gf_quotient(x, B, N)  % gradient of f(x)
    if(~isfield(x, 'BUN'))
        x.BUN =  B * (x.U * N);
    end
    Q = 2 * x.BUN;
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp = x.UtQ;
    output.TV = Q - x.U * temp';
end

function [output, x] = gf_quotient_intr(x, B, N)  % gradient of f(x)
    [n, p] = size(x.U);
    if(~isfield(x, 'BUN'))
        x.BUN =  B * (x.U * N);
    end
    Q = 2 * x.BUN;
    omega = x.U' * Q;
    omega = omega' - omega;
    K = x.Up' * Q;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(p, p), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);%%nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function [output, x] = hessianv_newton_embedded(x, v, B, N)
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp1 = x.UtQ;
    Q2 = 2 * B * (v.TV * N);
    temp2 = x.U' * Q2;
    Dgradfv = Q2 - v.TV * (temp1 + temp1') / 2;% - x * (temp2 + temp2') / 2;
    temp3 = x.U' * Dgradfv;
    output.TV = Dgradfv - x.U * (temp3 + temp3') / 2;
end

function output = hessianv_newton_embedded_intr(x, v, B, N)
    v = embedded_full_c(x, v);
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp1 = x.UtQ;
    Q2 = 2 * B * (v.TV * N);
    temp2 = x.U' * Q2;
    Dgradfv = Q2 - v.TV * (temp1 + temp1') / 2;% - x1 * (temp2 + temp2') / 2;
    temp3 = x.U' * Dgradfv;
    output.TV = Dgradfv - x.U * (temp3 + temp3') / 2;
    output = embedded_intr_c(x, output);
end

function output = hessianv_newton_quotient(x, v, B, N)
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp1 = Q' * v.TV;
    temp2 = v.TV * x.UtQ;
    temp3 = B * (v.TV * N);
    Fxx = 2 * (temp3 - x.U * (temp3' * x.U));
    output.TV = Fxx - x.U * (temp1  - temp1') / 2 - (v.TV * x.UtQ' - Q * (v.TV' * x.U)) / 2 - 0.5 * (temp2 - x.U * (x.U' * temp2));
end

function output = hessianv_newton_quotient_intr(x, v, B, N)
    v = quotient_full_c(x, v);
    if(~isfield(x, 'BUN'))
        x.BUN = B * (x.U * N);
    end
    Q = 2 * x.BUN;
    if(~isfield(x, 'UtQ'))
        x.UtQ = x.U' * Q;
    end
    temp1 = Q' * v.TV;
    temp2 = v.TV * x.UtQ;
    temp3 = B * (v.TV * N);
    Fxx = 2 * (temp3 - x.U * (temp3' * x.U));
    output.TV = Fxx - x.U * (temp1  - temp1') / 2 - (v.TV * x.UtQ' - Q * (v.TV' * x.U)) / 2 - 0.5 * (temp2 - x.U * (x.U' * temp2));
    output = quotient_intr_c(x, output);
end

function output = gf_exists(x, B, N)
    output = 1;
end

function output = embedded_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = quotient_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;
end

function output = quotient_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1));
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;
end

% function output = embedded_intr_c(x, eta)
%     x1 = x.U;
%     x2 = x.Up;
%     r2 = sqrt(2);
%     [n, p] = size(x1);
%     omega = x1' * eta; % omega
%     K = x2' * eta; % K
%     omega = - sqrt(2) * omega;
%     output = zeros(n * p - 0.5 * p * (p + 1), 1);
%     indx = find(tril(ones(size(omega)), -1));
%     output(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
% end
% 
% function output = quotient_intr_c(x, eta)
%     x1 = x.U;
%     x2 = x.Up;
%     [n, p] = size(x1);
%     omega = x1' * eta; % omega
%     K = x2' * eta; % K
%     omega = - omega;
%     output = zeros(n * p - 0.5 * p * (p + 1), 1);
%     indx = find(tril(ones(size(omega)), -1));
%     output(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
% end
% 
% function output = embedded_full_c(x, v)
%     x1 = x.U;
%     x2 = x.Up;
%     r2 = sqrt(2);
%     [n, p] = size(x1);
%     omega = tril(ones(p, p), -1);
%     indx = find(omega);
%     omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
%     omega = omega' - omega;
%     K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
%     output = x1 * omega + x2 * K;
% end
% 
% function output = quotient_full_c(x, v)
%     x1 = x.U;
%     x2 = x.Up;
%     [n, p] = size(x1);
%     omega = tril(ones(p, p), -1);
%     indx = find(omega);
%     omega(indx) = v.TV(1 : 0.5 * p * (p - 1));
%     omega = omega' - omega;
%     K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
%     output = x1 * omega + x2 * K;
% end
