function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_Grassmann_RQ(C, method, fns, params)
% driver for Rayleigh quotient on the Grassmann manifold.
% X^* = arg min_{X in Gr(p, n)} trace(X^T C X), where C is an n by n matrix.
%
% INPUT:
% C: the symmetric matrix in the cost function
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang
    if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 14)
        fns.f = @(x)f_intr(x, C);
        fns.gf = @(x)gf_intr(x, C);
        fns.hessianv = @(x, v)hessianv_newton_intr(x, v, C);
    else
        fns.f = @(x)f(x, C);
        fns.gf = @(x)gf(x, C);
        fns.hessianv = @(x, v)hessianv_newton(x, v, C);
    end
    fns.gf_exists = @(x)gf_exists(x, C);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, T, nR] = check_grad_Hess(fns, params);
    end
end

function [output, x] = f(x, C)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    temp = x.CU;
    output = sum(sum(x.U .* temp));
end

function [output, x] = gf(x, C)  % gradient of f(x)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    Q = 2 * x.CU;
    output.TV = Q - x.U * (x.U' * Q);
end

function [output, x] = hessianv_newton(x, v, C)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    temp = C * v.TV;
    output.TV = 2 * (temp - x.U * (x.U' * temp) - v.TV * (x.U' * x.CU));
end

function [output, x] = f_intr(x, C)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    output = sum(sum(x.U .* x.CU));
end

function [output, x] = gf_intr(x, C)  % gradient of f(x)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    Q = 2 * x.CU;
    output.TV = reshape(x.Up' * Q, [], 1);
end

function output = hessianv_newton_intr(x, v, C)
    if(~isfield(x, 'CU'))
        x.CU = C * x.U;
    end
    v = full_c(x, v);
    temp = C * v.TV;
    output.TV = 2 * (temp - x.U * (x.U' * temp) - v.TV * (x.U' * x.CU));
    output = intr_c(x, output);
end

function output = gf_exists(x, C)
    output = 1;
end

function output = intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    K = x2' * eta.TV; % K
    output.TV = reshape(K, [], 1);
end

function output = full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    output.TV = x2 * reshape(v.TV, (n - p), p);
end
