function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_OBB(A, method, params)
% driver for Oriented Bounding Box problem.
% See https://www.elen.ucl.ac.be/Proceedings/esann/esannpdf/es2010-117.pdf
% 
% INPUT:
% A: a n by N matrix. A column is a point in n dimensional space. The number of points is N.
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang

    fns.f = @(x)f(x, A);
    if(params.vector_transport == 7 || params.vector_transport == 13)
        fns.gf = @(x)gf_intr(x, A);
    else
        fns.gf = @(x)gf(x, A);
    end
    fns.gf_exists = @(x)gf_exists(x, A);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    end
end

function [output, x] = f(x, A)
    xx = x.U * A;
    max_xx = max(xx');
    min_xx = min(xx');
    diff = max_xx - min_xx;
    output = 1;
    for i = 1 : length(diff)
        output = diff(i) * output;
    end
end

function [output, x] = gf(x, A)  % gradient of f(x)
    nn = size(A, 1);
    xx = x.U * A;
    [max_xx, max_index] = max(xx');
    [min_xx, min_index] = min(xx');
    diff = max_xx - min_xx;
    output = zeros(size(A));
    total = 1;
    for i = 1 : nn
        total = total * diff(i);
    end
    
    for i = 1 : nn
        output(i, max_index(i)) = total / diff(i);
    end
    
    for i = 1 : nn
        output(i, min_index(i)) = - total / diff(i);
    end
    output = output * A';
    output.TV = output - x.U * (x.U' * output + output' * x.U) / 2;
end

function [output, x] = gf_intr(x, A)  % gradient of f(x)
    nn = size(A, 1);
    xx = x.U * A;
    [max_xx, max_index] = max(xx');
    [min_xx, min_index] = min(xx');
    diff = max_xx - min_xx;
    Temp = zeros(size(A));
    total = 1;
    for i = 1 : nn
        total = total * diff(i);
    end
    
    for i = 1 : nn
        Temp(i, max_index(i)) = total / diff(i);
    end
    
    for i = 1 : nn
        Temp(i, min_index(i)) = - total / diff(i);
    end
    Q = Temp * A';
    [n, p] = size(x.U);
    omega = 0.5 * x.U' * Q;
    omega = omega - omega';
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    output.TV(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
end

function output = gf_exists(x, A)
    output = 1;
end
