function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_SBNLDR_partly_smooth(Sigma, method, fns, params)
% driver for Broomhead and Kirby's NLDR with partly smooth cost function
% See paper: "Dimensionality reduction using secant-based projection 
% methods : the induced dynamics in projected systems"
%
% INPUT:
% Sigma: the set of unit secants
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang

    if(params.vector_transport == 7 || params.vector_transport == 13)
        fns.f = @(x)f_intr(x, Sigma);
        fns.gf = @(x)gf_intr(x, Sigma);
    else
        fns.f = @(x)f(x, Sigma);
        fns.gf = @(x)gf(x, Sigma);
    end
    fns.gf_exists = @(x)gf_exists(x, Sigma);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    end
end

function [output, x] = f(x, Sigma)
    min_f = inf;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        fv = norm(x.U' * k)^2;
        if(fv < min_f)
            min_f = fv;
            min_indx = i;
        end
    end
    output = - min_f;
end

function [output, x] = gf(x, Sigma)  % gradient of f(x)
    min_f = inf;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        fv = norm(x.U' * k)^2;
        if(fv < min_f)
            min_f = fv;
            min_indx = i;
        end
    end
    k_min = Sigma(min_indx, :)';
    temp = 2 * k_min * (k_min' * x.U);
    output.TV = - (temp - x.U * (x.U' * temp));
end

function [output, x] = f_intr(x, Sigma)
    x1 = x.U;
    min_f = inf;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        fv = norm(x1' * k)^2;
        if(fv < min_f)
            min_f = fv;
            min_indx = i;
        end
    end
    output = - min_f;
end

function [output, x] = gf_intr(x, Sigma)  % gradient of f(x)
    x1 = x.U;
    x2 = x.Up;
    min_f = inf;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        fv = norm(x1' * k)^2;
        if(fv < min_f)
            min_f = fv;
            min_indx = i;
        end
    end
    k_min = Sigma(min_indx, :)';
    temp = 2 * k_min * (k_min' * x1);
    output.TV = - reshape(x2' * temp, [], 1);
end

function output = gf_exists(x, Sigma)
    output = 1;
end

function output = intr_c(x, eta)
    x1 = x{1};
    x2 = x{2};
    [n, p] = size(x1);
    K = x2' * eta; % K
    output = reshape(K, [], 1);
end

function output = full_c(x, v)
    x1 = x{1};
    x2 = x{2};
    [n, p] = size(x1);
    output = x2 * reshape(v, (n - p), p);
end
