function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_SBNLDR_smooth(Sigma, method, fns, params)
% driver for Broomhead and Kirby's NLDR with smooth cost function
% See paper: "Dimensionality reduction using secant-based projection 
% methods : the induced dynamics in projected systems"
%
% INPUT:
% Sigma: the set of unit secants
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang
    if(params.vector_transport == 7 || params.vector_transport == 13)
        fns.f = @(x)f_intr(x, Sigma);
        fns.gf = @(x)gf_intr(x, Sigma);
        fns.hessianv = @(x, v)hessianv_newton_intr(x, v, Sigma);
    else
        fns.f = @(x)f(x, Sigma);
        fns.gf = @(x)gf(x, Sigma);
        fns.hessianv = @(x, v)hessianv_newton(x, v, Sigma);
    end
    fns.gf_exists = @(x)gf_exists(x, Sigma);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, T, nR] = check_grad_Hess(fns, params);
    end
end

function [output, x] = f(x, Sigma)
    output = 0;
    for i = 1 : size(Sigma, 1)
        output = output + 1 / norm(x.U' * Sigma(i, :)');
    end
    output = output / size(Sigma, 1);
end

function [output, x] = gf(x, Sigma)  % gradient of f(x)
    output.TV = 0;
    for i = 1 : size(Sigma, 1)
        output.TV = output.TV - 1 / norm(x.U' * Sigma(i, :)')^3 * Sigma(i, :)' * (Sigma(i, :) * x.U);
    end
    output.TV = (output.TV - x.U * (x.U' * output.TV)) / size(Sigma, 1);
end

function output = hessianv_newton(x, v, Sigma)
    output.TV = 0;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        n = norm(x.U' * k);
        temp1 = k * (k' * x.U);
        temp2 = k * (k' * v.TV);
        temp3 = - v.TV * (x.U' * temp1);
        temp4 = (temp3 + temp2) / n^3;
        temp5 = -3 * temp1 / n^5 * ((k' * x.U) * (v.TV' * k));
        output.TV = output.TV + temp4 + temp5;
    end
    output.TV = - output.TV / size(Sigma, 1);
    output.TV = output.TV - x.U * (x.U' * output.TV);
end

function [output, x] = f_intr(x, Sigma)
    x1 = x.U;
    output = 0;
    for i = 1 : size(Sigma, 1)
        output = output + 1 / norm(x1' * Sigma(i, :)');
    end
    output = output / size(Sigma, 1);
end

function [output, x] = gf_intr(x, Sigma)  % gradient of f(x)
    x1 = x.U;
    x2 = x.Up;
    Temp = 0;
    for i = 1 : size(Sigma, 1)
        Temp = Temp - 1 / norm(x1 * x1' * Sigma(i, :)')^3 * Sigma(i, :)' * (Sigma(i, :) * x1);
    end
    Temp = Temp / size(Sigma, 1);
    output.TV = reshape(x2' * Temp, [], 1);
end

function output = hessianv_newton_intr(x, v, Sigma)
    v = full_c(x, v);
    x1 = x.U;
    x2 = x.Up;
    output.TV = 0;
    for i = 1 : size(Sigma, 1)
        k = Sigma(i, :)';
        n = norm(x1' * k);
        temp1 = k * (k' * x1);
        temp2 = k * (k' * v.TV);
        temp3 = - v.TV * (x1' * temp1);
        temp4 = (temp3 + temp2) / n^3;
        temp5 = -3 * temp1 / n^5 * ((k' * x1) * (v.TV' * k));
        output.TV = output.TV + temp4 + temp5;
    end
    output.TV = - output.TV / size(Sigma, 1);
    output.TV = output.TV - x1 * (x1' * output.TV);
    output = intr_c(x, output);
end

function output = gf_exists(x, Sigma)
    output = 1;
end

function output = intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    K = x2' * eta.TV; % K
    output.TV = reshape(K, [], 1);
end

function output = full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    output.TV = x2 * reshape(v.TV, (n - p), p);
end
