function [X, F, G, timecost] = driver_manton_lowrank_Qinv(A, Q_inv, fns, params, method)

    if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 14)
        fns.f = @(x)f_intr(x, A, Q_inv);
        fns.gf = @(x)gf_intr(x, A, Q_inv);
        fns.hessianv = @(x, eta)hessianv_newton_intr(x, eta, A, Q_inv);
        fns.proj = @(x, eta)proj_intr(x, eta);
    else
        fns.f = @(x)f(x, A, Q_inv);
        fns.gf = @(x)gf(x, A, Q_inv);
        fns.hessianv = @(x, eta)hessianv_newton(x, eta, A, Q_inv);
        fns.proj = @(x, eta)proj(x, eta);
    end
    
    
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, T, nR] = check_grad_Hess(fns, params);
    end

end

function [output, x] = f(x, A, Q_inv)
    [m, n] = size(A);
    p = size(x.U, 2);
      
    % QN =  kron(N, I) * Q^{-1}
    x.QN = zeros(m*p, m*n);
    for i = 1 : (m*n)
        q = reshape(Q_inv(:,i), m, n);
        x.QN(:,i) = reshape(q * x.U, [], 1);
    end
    
    % h = kron(N, I)^T Q^{-1} kron(N,I)
    x.h = zeros(m*p, m*p);
    for i = 1 : (m*p)
        q = reshape(x.QN(i, :), m, n);
        x.h(i, :) = reshape(q * x.U, [], 1);
    end

    % kron(x, I)*vec{A}=vec{Ax}
    x.AX = reshape(A * x.U, [], 1);
    x.hAX = x.h \ x.AX;
%     x.hAX = linsolve(x.h, x.AX);
    output = x.AX' * x.hAX;
end

function [output, x] = f_intr(x, A, Q_inv)
    x1 = x.U;
    [m, n] = size(A);
    p = size(x1, 2);
    
    x.QN = zeros(m*p, m*n);
    for i = 1 : (m*n)
        q = reshape(Q_inv(:,i), m, n);
        x.QN(:,i) = reshape(q * x1, [], 1);
    end
    
    x.h = zeros(m*p, m*p);
    for i = 1 : (m*p)
        q = reshape(x.QN(i, :), m, n);
        x.h(i, :) = reshape(q * x1, [], 1);
    end
    
    % kron(x, I)*vec{A}=vec{Ax}
    x.AX = reshape(A*x1, [], 1);
    x.hAX = x.h \ x.AX;
%     x.hAX = linsolve(x.h, x.AX);
    output = x.AX' * x.hAX;
    
end

function [output, x] = gf(x, A, Q_inv)
    [m, n] = size(A);
    p = size(x.U, 2);
    
    if(~isfield(x, 'hAX'))
        if(~isfield(x, 'h'))
            if(~isfield(x,'QN'))
                x.QN = zeros(m*p, m*n);
                for i = 1 : (m*n)
                    q = reshape(Q_inv(:,i), m, n);
                    x.QN(:,i) = reshape(q * x.U, [], 1);
                end
            end
            x.h = zeros(m*p, m*p);
            for i = 1 : (m*p)
                q = reshape(x.QN(i, :), m, n);
                x.h(i, :) = reshape(q * x.U, [], 1);
            end 
        end
        if(~isfield(x,'AX'))
            x.AX = reshape(A * x.U, [], 1);
        end
        x.hAX = x.h \ x.AX;
    end
    x.matrixA = reshape(x.hAX, m, p);
    
    x.vectorC = Q_inv * reshape(x.matrixA * x.U', [], 1);
    x.matrixC = reshape(x.vectorC, m, n);
    
    x.grad = 2 * (A - x.matrixC)' * x.matrixA;
    output.TV = x.grad - x.U * (x.U' * x.grad);
    
end

function [output, x] = gf_intr(x, A, Q_inv)
    x1 = x.U;
    x2 = x.Up;
    [m, n] = size(A);
    p = size(x1, 2);
    
    if(~isfield(x, 'hAX'))
        if(~isfield(x, 'h'))
            if(~isfield(x,'QN'))
                x.QN = zeros(m*p, m*n);
                for i = 1 : (m*n)
                    q = reshape(Q_inv(:,i), m, n);
                    x.QN(:,i) = reshape(q * x1, [], 1);
                end
            end
            x.h = zeros(m*p, m*p);
            for i = 1 : (m*p)
                q = reshape(x.QN(i, :), m, n);
                x.h(i, :) = reshape(q * x1, [], 1);
            end 
        end
        if(~isfield(x,'AX'))
            x.AX = reshape(A * x1, [], 1);
        end
        x.hAX = x.h \ x.AX;
    end
    x.matrixA = reshape(x.hAX, m, p);
    
    x.vectorC = Q_inv * reshape(x.matrixA * x1', [], 1);
    x.matrixC = reshape(x.vectorC, m, n);
    
    x.grad = 2 * (A - x.matrixC)' * x.matrixA;
    output.TV = x.grad - x1 * (x1' * x.grad);
    output.TV = reshape(x2' * output.TV, [], 1);
end

function output = hessianv_newton(x, eta, A, Q_inv)
    [m, n] = size(A);
    p = size(x.U, 2);
    
    % compute matrix A and C
    if(~isfield(x, 'matrixC'))
        if(~isfield(x, 'matrixA'))
            if(~isfield(x, 'hAX'))
                if(~isfield(x, 'h'))
                    if(~isfield(x,'QN'))
                        x.QN = zeros(m*p, m*n);
                        for i = 1 : (m*n)
                            q = reshape(Q_inv(:,i), m, n);
                            x.QN(:,i) = reshape(q * x.U, [], 1);
                        end
                    end
                    x.h = zeros(m*p, m*p);
                    for i = 1 : (m*p)
                        q = reshape(x.QN(i, :), m, n);
                        x.h(i, :) = reshape(q * x.U, [], 1);
                    end 
                end
                if(1)%~isfield(x,'AX'))
                    x.AX = reshape(A * x.U, [], 1);
                end
                x.hAX = x.h \ x.AX;
            end
            x.matrixA = reshape(x.hAX, m, p); 
        end
        x.vectorC = Q_inv * reshape(x.matrixA * x.U', [], 1);
        x.matrixC = reshape(x.vectorC, m, n);
    end
     
    % compute dAm
    tempa = reshape(x.matrixA * eta.TV', [], 1);
    tempb = x.QN * tempa;    
    tempd = reshape((A - x.matrixC) * eta.TV, [], 1);
    dA = x.h \ (tempd - tempb);
    dAm = reshape(dA, m, p);
    
    % compute dC
    dCv = Q_inv * (tempa + reshape(dAm * x.U', [], 1));
    dC = reshape(dCv, m, n);
   
    % (eta*N^T+N*eta^T)*(X-C)^T*A
    N_eta = x.U * eta.TV';
    temp = N_eta + N_eta';
    a = temp * (A - x.matrixC)' * x.matrixA;
    
    % (I-NN^T){dC/dN}^TA
    b = dC' * x.matrixA - x.U * ((dC * x.U)' * x.matrixA);
    
    % (I-NN^T)*(X-C)^TdA
    c = (A - x.matrixC)' * dAm - x.U * ((A - x.matrixC) * x.U)' * dAm;
     
    output_temp = 2 * (- a - b + c);
    output.TV = output_temp - x.U * (x.U' * output_temp);
end

function output = hessianv_newton_intr(x, eta, A, Q_inv)
    x1 = x.U;
    eta = full_c(x, eta);
    [m, n] = size(A);
    p = size(x1, 2);
    
    % compute matrix A and C
    if(~isfield(x, 'matrixC'))
        if(~isfield(x, 'matrixA'))
            if(~isfield(x, 'hAX'))
                if(~isfield(x, 'h'))
                    if(~isfield(x,'QN'))
                        x.QN = zeros(m*p, m*n);
                        for i = 1 : (m*n)
                            q = reshape(Q_inv(:,i), m, n);
                            x.QN(:,i) = reshape(q * x1, [], 1);
                        end
                    end
                    x.h = zeros(m*p, m*p);
                    for i = 1 : (m*p)
                        q = reshape(x.QN(i, :), m, n);
                        x.h(i, :) = reshape(q * x1, [], 1);
                    end 
                end
                if(~isfield(x,'AX'))
                    x.AX = reshape(A*x1, [], 1);
                end
                x.hAX = x.h \ x.AX;
            end
            x.matrixA = reshape(x.hAX, m, p); 
        end
        x.vectorC = Q_inv * reshape(x.matrixA * x1', [], 1);
        x.matrixC = reshape(x.vectorC, m, n);
    end

   
    % compute dAm
    tempa = reshape(x.matrixA * eta.TV', [], 1);
    tempb = x.QN * tempa;    
    tempd = reshape((A - x.matrixC) * eta.TV, [], 1);
    dA = x.h \ (tempd - tempb);
    dAm = reshape(dA, m, p);
    
    % compute dC
    dCv = Q_inv * (tempa + reshape(dAm * x1', [], 1));
    dC = reshape(dCv, m, n);
    
    % (eta*N^T+N*eta^T)*(X-C)^T*A
    N_eta = x1 * eta.TV';
    temp = N_eta + N_eta';
    a = temp * (A - x.matrixC)' * x.matrixA;
    
    % (I-NN^T){dC/dN}^TA
    b = dC' * x.matrixA - x1 * ((dC * x1)' * x.matrixA);
    
    % (I-NN^T)*(X-C)^TdA
    c = (A - x.matrixC)'*dAm - x1 * ((A - x.matrixC) * x1)' * dAm;
     
    outputt = 2 * (- a - b + c);
    output.TV = outputt - x1 * (x1' * outputt);
    output = intr_c(x, output);
end

function output = proj(x, eta)
    output.TV = eta.TV - x.U * (x.U' * eta.TV);
end

function output = proj_intr(x, eta)
    output.TV = eta.TV;
end


function output = full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    output.TV = x2 * reshape(v.TV, (n - p), p);
end

function output = intr_c(x, eta)
    x1 = x.U;
    if(~isfield(x, 'Up'))
        x.Up = null(x.U');
    end
    x2 = x.Up;
    [n, p] = size(x1);
    K = x2' * eta.TV; % K
    output.TV = reshape(K, [], 1);
end




