function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_partly_sphere(method, fns, params)
% driver for Lipschitz partly smooth function on sphere.
% x^* = arg min_{x in S^{n - 1}} max(|x_i|).
% 
% INPUT:
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang
    fns.f = @(x)f(x);
    fns.gf = @(x)gf(x);
    fns.gf_exists = @(x)gf_exists(x);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    end
end

function [output, x] = f(x)
    output = max(abs(x.U));
end

function [output, x] = gf(x)
    output.TV = zeros(size(x.U));
    [m, i] = max(abs(x.U));
    output.TV(i) = sign(x.U(i));
%     output = (eye(size(x, 1)) - x * x') * output; %% small numerical error, but slow
    output.TV = (output.TV - x.U * (x.U' * output.TV)); %% big numerical error, but fast
end

function output = gf_exists(x)
    output = 1;
%     if(min(diff(sort(abs(x)))) < 1e-10);
%         output = 0;
%     end
end
