function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_soft_ICA(C, method, params)
% driver for soft ICA problem.
% X^* = arg min_{x in St(p, n)} - \sum_{i=1}^N ||diag(Y^T C_i Y)||_F^2, where C_i are positive-semidefinite symmetric matrix.
% 
% INPUT:
% C: the set of all samples. an n by n by N matrix where n is the size of the sample and N is the number of samples
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang

    fns.f = @(x)soft_ICA_f(x, C);
    if(params.retraction == 2 || params.retraction == 4 ||  params.retraction == 6)
        fprintf('gradient of test function: stiefel manifold is considered as embedded space\n')
        fns.gf = @(x)soft_ICA_gf_embedded(x, C);
        fns.hessianv = @(x, v)hessianv_embedded(x, v, C);
        if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
            fns.f = @(x)soft_ICA_f_intr(x, C);
            fns.gf = @(x)soft_ICA_gf_embedded_intr(x, C);
            fns.hessianv = @(x, v)hessianv_embedded_intr(x, v, C);
        end
    else
        fprintf('gradient of quotient function: stiefel manifold is considered as quotient space\n')
        fns.gf = @(x)soft_ICA_gf_quotient(x, C);
        fns.hessianv = @(x, v)hessianv_quotient(x, v, C);
        if(params.vector_transport == 7 || params.vector_transport == 13 || params.vector_transport == 19 || params.vector_transport == 25)
            fns.f = @(x)soft_ICA_f_intr(x, C);
            fns.gf = @(x)soft_ICA_gf_quotient_intr(x, C);
            fns.hessianv = @(x, v)hessianv_quotient_intr(x, v, C);
        end
    end
    fns.gf_exists = @(x)gf_exists(x, B, N);
    
    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RWRBFGS(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 12)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    end
end

function [output, Y] = soft_ICA_f(Y, C)
    N = size(C, 3);
    output = 0;
    for i = 1 : N
        output = output - norm(diag(Y.U' * C(:, :, i) * Y.U))^2;
    end
end

function [output, Y] = soft_ICA_f_intr(Y, C)
    N = size(C, 3);
    output = 0;
    for i = 1 : N
        output = output - norm(diag(Y.U' * C(:, :, i) * Y.U))^2;
    end
end

function [output, Y] = soft_ICA_gf_embedded(Y, C)
    N = size(C, 3);
    [n, p] = size(Y.U);
    Q = zeros(n, p);
    for i = 1 : N
        temp = C(:, :, i) * Y.U;
        Q = Q - 4 * temp * diag(diag(Y.U' * temp));
    end
    temp = Y.U' * Q;
    output.TV = Q - Y.U * (temp + temp') / 2;
end

function [output, Y] = soft_ICA_gf_embedded_intr(Y, C)
    x1 = Y.U;
    x2 = Y.Up;
    N = size(C, 3);
    [n, p] = size(x1);
    Q = zeros(n, p);
    for i = 1 : N
        temp = C(:, :, i) * x1;
        Q = Q - 4 * temp * diag(diag(x1' * temp));
    end

    K = x2' * Q;
    omega = 0.5 * x1' * Q;
    omega = omega - omega';
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(p, p), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);%nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function [output, Y] = soft_ICA_gf_quotient(Y, C)
    N = size(C, 3);
    [n, p] = size(Y.U);
    Q = zeros(n, p);
    for i = 1 : N
        Q = Q - 4 * C(:, :, i) * Y.U * diag(diag(Y.U' * C(:, :, i) * Y.U));
    end
    output.TV = Q - Y.U * Q' * Y.U;
end

function output = gf_exists(x, C)
    output = 1;
end

function [output, Y] = soft_ICA_gf_quotient_intr(Y, C)
    x1 = Y.U;
    x2 = Y.Up;
    N = size(C, 3);
    [n, p] = size(x1);
    Q = zeros(n, p);
    for i = 1 : N
        temp = C(:, :, i) * x1;
        Q = Q - 4 * temp * diag(diag(x1' * temp));
    end
    
    omega = x1' * Q;
    omega = omega' - omega;
    K = x2' * Q;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(p, p), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);%%nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = hessianv_embedded(Y, v, C)
    N = size(C, 3);
    [n, p] = size(Y.U);
    Q1 = zeros(n, p);
    Q2 = zeros(n, p);
    for i = 1 : N
        temp1 = C(:, :, i) * Y.U;
        temp2 = Y.U' * temp1;
        
%         Q1 = Q1 - 4 * C(:, :, i) * (v .* repmat(diag(temp2)', n, 1) + 2 * Y .* repmat(diag(v' * temp1)', n, 1));
%         Q2 = Q2 - 4 * temp1 .* repmat(diag(temp2)', n, 1);
        Q1 = Q1 - 4 * C(:, :, i) * (v.TV * diag(diag(temp2)) + 2 * Y.U * diag(diag(v.TV' * temp1)));
        Q2 = Q2 - 4 * temp1 * diag(diag(temp2));
        
%         Q1 = Q1 - 4 * C(:, :, i) * (v * diag(diag(temp2)) + Y * diag(diag(v' * temp1)) + Y * diag(diag(temp1' * v)));
%         Q2 = Q2 - 4 * C(:, :, i) * Y * diag(diag(temp2));
%         Q1 = Q1 - 4 * C(:, :, i) * (v * diag(diag(Y' * C(:, :, i) * Y)) + Y * diag(diag(v' * C(:, :, i) * Y)) + Y * diag(diag(Y' * C(:, :, i) * v)));
%         Q2 = Q2 - 4 * C(:, :, i) * Y * diag(diag(Y' * C(:, :, i) * Y));
    end
    temp1 = Y.U' * Q2;
    temp2 = v.TV * (temp1 + temp1') / 2;
    temp3 = Q1 - temp2;
    temp4 = Y.U' * temp3;
    output.TV = temp3 - Y.U * (temp4 + temp4') / 2;
end

function output = hessianv_embedded_intr(Y, v, C)
    v = embedded_full_c(Y, v);
    Y1 = Y.U;
    
    N = size(C, 3);
    [n, p] = size(Y1);
    Q1 = zeros(n, p);
    Q2 = zeros(n, p);
    for i = 1 : N
        temp1 = C(:, :, i) * Y1;
        temp2 = Y1' * temp1;
        
%         Q1 = Q1 - 4 * C(:, :, i) * (v .* repmat(diag(temp2)', n, 1) + 2 * Y .* repmat(diag(v' * temp1)', n, 1));
%         Q2 = Q2 - 4 * temp1 .* repmat(diag(temp2)', n, 1);
        Q1 = Q1 - 4 * C(:, :, i) * (v.TV * diag(diag(temp2)) + 2 * Y1 * diag(diag(v.TV' * temp1)));
        Q2 = Q2 - 4 * temp1 * diag(diag(temp2));
        
%         Q1 = Q1 - 4 * C(:, :, i) * (v * diag(diag(temp2)) + Y * diag(diag(v' * temp1)) + Y * diag(diag(temp1' * v)));
%         Q2 = Q2 - 4 * C(:, :, i) * Y * diag(diag(temp2));
%         Q1 = Q1 - 4 * C(:, :, i) * (v * diag(diag(Y' * C(:, :, i) * Y)) + Y * diag(diag(v' * C(:, :, i) * Y)) + Y * diag(diag(Y' * C(:, :, i) * v)));
%         Q2 = Q2 - 4 * C(:, :, i) * Y * diag(diag(Y' * C(:, :, i) * Y));
    end
    temp1 = Y1' * Q2;
    temp2 = v.TV * (temp1 + temp1') / 2;
    temp3 = Q1 - temp2;
    temp4 = Y1' * temp3;
    output.TV = temp3 - Y1 * (temp4 + temp4') / 2;

    output = embedded_intr_c(Y, output);
end

function output = hessianv_quotient(Y, v, C)
    N = size(C, 3);
    [n, p] = size(Y.U);
    Q1 = zeros(n, p);
    Q2 = zeros(n, p);
    for i = 1 : N
        temp1 = C(:, :, i) * Y.U;
        temp2 = Y.U' * temp1;
        
        Q1 = Q1 - 4 * C(:, :, i) * (v.TV * diag(diag(temp2)) + 2 * Y.U * diag(diag(v.TV' * temp1)));
        Q2 = Q2 - 4 * temp1 * diag(diag(temp2));
    end
    % Q1 D gradf(Y) [xi]
    % Q2 gradf(Y)
    temp1 = Q2' * v.TV;
    temp2 = v.TV * Q2';
    temp3 = v.TV * (Y.U' * Q2);
    output.TV = Q1 - Y.U * (Q1)' * Y.U - Y.U * (temp1 - temp1') / 2 - (temp2 - temp2') / 2 * Y.U - 0.5 * (temp3 - Y.U * (Y.U' * temp3));
    
%     temp1 = Y' * Q1;
%     temp2 = Y' * Q2;
%     output = Q1 - Y * (temp1 + temp1') / 2 - v * (temp2 + temp2') / 2;
end

function output = hessianv_quotient_intr(Y, v, C)
    v = quotient_full_c(Y, v);
    Y1 = Y.U;
    
    N = size(C, 3);
    [n, p] = size(Y1);
    Q1 = zeros(n, p);
    Q2 = zeros(n, p);
    for i = 1 : N
        temp1 = C(:, :, i) * Y1;
        temp2 = Y1' * temp1;
        
        Q1 = Q1 - 4 * C(:, :, i) * (v.TV * diag(diag(temp2)) + 2 * Y1 * diag(diag(v.TV' * temp1)));
        Q2 = Q2 - 4 * temp1 * diag(diag(temp2));
    end
    % Q1 D gradf(Y) [xi]
    % Q2 gradf(Y)
    temp1 = Q2' * v.TV;
    temp2 = v.TV * Q2';
    temp3 = v.TV * (Y1' * Q2);
    output.TV = Q1 - Y1 * (Q1)' * Y1 - Y1 * (temp1 - temp1') / 2 - (temp2 - temp2') / 2 * Y1 - 0.5 * (temp3 - Y1 * (Y1' * temp3));
    
    output = quotient_intr_c(Y, output);
end

function output = M_diag(Y, D)
    [n, p] = size(Y);
    output = zeros(n, p);
    for i = 1 : p
        output(:, i) = Y(:, i) * D(i, i);
    end
end

% function output = build_bases_embedded(x)
%     [n, p] = size(x);
%     x_perp = null(x');
%     x = x / sqrt(2);
%     d = n * p - p * (p + 1) / 2;
%     output = zeros(n * p, d);
%     col = 1;
%     for i = 2 : p
%         for j = i : p
%             output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
%             output((j - 1) * n + 1 : j * n, col) = x(:, i - 1);
%             col = col + 1;
%         end
%     end
%     for i = 1 : p
%         output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp(1 : n, 1 : n - p);
%         col = col + n - p;
%     end
% end

function output = embedded_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - r2 * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;
end

function output = quotient_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = quotient_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1));
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;

%     omega = zeros(p, p);
%     ind = 1;
%     for i = 1 : p - 1
%         omega(i, i + 1 : end) = v(ind : ind + p - 1 - i)';
%         ind = ind + p - i;
%     end
%     omega = omega - omega';
%     K = reshape(v(ind : end), (n - p), p);
%     output = x1 * omega + x2 * K;
end
