function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_sphere_RQ(A, method, fns, params)
% driver for Rayleigh Quotient minimization on the sphere.
% x^* = arg min_{x in S^{n - 1}} x^T A x, where A is a n by n symmetric matrix.
% 
% INPUT:
% A: the matrix in the cost function
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang
    fns.f = @(x)f(x, A);
    fns.gf = @(x)gf(x, A);
    fns.gf_exists = @(x)gf_exists(x, A);
    fns.hessianv = @(x, v)hessianv_newton(x, v, A);

    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBroydenFamily(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RSD(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 6)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    elseif(method == 7)
        [X, F, G, timecost, nf, ng, T, nR] = RTR_SD(fns, params);
    elseif(method == 8)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RGS(fns, params);
    elseif(method == 9)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 10)
        [X, F, G, timecost, nf, ng, T, nV, nR] = RCG(fns, params);
    elseif(method == 11)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RNewton(fns, params);
    end
end

function [output, x] = f(x, A)
    x.Ax = A * x.U;
    output = x.U' * x.Ax;
end

function [output, x] = gf(x, A)
    if(~isfield(x, 'Ax'))
        x.Ax = A * x.U;
    end
    temp = x.Ax;
%     output = 2 * (eye(size(x, 1)) - x * x') * temp; %% small numerical error but slow
    output.TV = 2 * (temp - x.U * (x.U' * temp)); %% big numerical error, but fast
end

function output = gf_exists(x, A)
    output = 1;
end

function [output, x] = hessianv_newton(x, w, A)
    % The hessian
    if(~isfield(x, 'Ax'))
        x.Ax = A * x.U;
    end
    xtax = x.U'* x.Ax;
    temp = (A * w.TV-w.TV * xtax);
    output.TV = 2 * (temp - x.U * (x.U' * temp));
end
