function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp] = driver_synchronization(H, E, kappas, p, method, params)
% driver for synchronization of rotation problem.
% See paper: {C}ramer-{R}ao bounds for synchronization of rotations
% 
% INPUT:
% H: estimations of R_i R_j^T for some (i, j). It is an n by n by N by N matrix
% E: knwon edges in the graph
% kappa: argument in the cost function
% p: argument in the cost function
% method: indication of method
% fns: a struct that contains required function handles 
% params: a struct that contains parameters
% 
% OUTPUT:
% X : all iterates
% F : function values of all iterates
% G : norm of gradient of all iterates
% T : accumulated computational time for iterations
% timecost : total computational time
% nf : number of function evaluations
% ng : number of gradient evaluations
% nR : number of retraction actions
% nH : number of Hessian actions
% nV : number of vector transport actions without parameters
% nVp : number of vector transport action with parameters
% 
% By Wen Huang
    N = size(E, 1);
    tempc3 = c3(1/2);
    for i = 1 : N
        for j = 1 : N
            c3kappas(i, j) = tempc3;%%-- c3(kappas(i, j));
        end
    end
    c3_0 = c3(0);
    M = 2;

    fns.f = @(R)f(R, H, E, kappas, p, c3kappas, c3_0, M);
    fns.gf = @(R)gf(R, H, E, kappas, p, c3kappas, c3_0, M);
    
    if(method == 1)
        fns.gf = @(R)gf_full(R, H, E, kappas, p, c3kappas, c3_0, M);
        fns.hessianv = @(x, v)fhessianv_full(x, v,  H, E, kappas, p, c3kappas, c3_0, M);
    end
    
    fns.proj = @proj;
    fns.Hv = @Hv;
    fns.STV = @STV;
    fns.TVaddTV = @TVaddTV;
    fns.phi = @phi;
    if(params.Retraction == 1)
        fns.R = @R_Exp;
    elseif(params.Retraction == 2)
        fns.R = @R;
    end
    fns.inpro = @inpro;
    
    if(method == 4 || method == 5)
        fns.Tranv = @Tranv;
        fns.Tranv_Params = @Tranv_Params;
        fns.invTranv = @invTranv;
        fns.TranH = @TranH;
    elseif(method == 2 || method == 3)
        if(params.Retraction == 2)
            fns.beta = @beta_qf;
            fns.Tranv = @Tranv_qf;
            fns.Tranv_Params = @Tranv_Params_qf;
            fns.invTranv = @invTranv_qf;
            fns.TranH = @TranH_qf;
        else
            fns.beta = @beta;
            fns.Tranv = @Tranv;
            fns.Tranv_Params = @Tranv_Params;
            fns.invTranv = @invTranv;
            fns.TranH = @TranH;
        end
    end
    fns.rank1operator = @rank1operator;
    fns.operadd = @operadd;
    if(method == 1)
        fns.inpro = @inpro_full;
        fns.g = @inpro_full;
        fns.proj = @proj_full;
        if(params.Retraction == 1)
            fns.R = @R_Exp_full;
        elseif(params.Retraction == 2)
            fns.R = @R_full;
        end
    end
    
    nH = 0;
    nV = 0;
    nVp = 0;
    if(method == 1)
        [X, F, G, timecost, nf, ng, nH, T, nR] = RTR_Newton(fns, params);
    elseif(method == 2)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RBFGS(fns, params);
    elseif(method == 3)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRBFGS(fns, params);
    elseif(method == 4)
        [X, F, G, timecost, nf, ng, nH, T, nV, nR] = RTR_SR1(fns, params);
    elseif(method == 5)
        [X, F, G, timecost, nf, ng, T, nV, nVp, nR] = LRTR_SR1(fns, params);
    end
end

function output = STV(x, a, eta)
    output.TV = a * eta.TV;
end

function output = TVaddTV(x, eta1, eta2)
    output.TV = eta1.TV + eta2.TV;
end

function output = proj_full(x, eta)
    [n, n, N] = size(x.U);
    output.TV = zeros(n, n, N);
    for i = 1 : N
        output.TV(:, :, i) = eta.TV(:, :, i) - x.U(:, :, i) * (x.U(:, :, i)' * eta.TV(:, :, i) + eta.TV(:, :, i)' * x.U(:, :, i)) / 2;
    end
end

function output = proj(x, eta)
    output = eta;
end

function output = c3(kappa)
    output = exp(kappa) * (I(0, 2 * kappa) - I(1, 2 * kappa));
end

function output = I(v, x)
    fhandle = @(theta)intI(theta, v, x);
    output = quad(fhandle,0,2 * pi) / 2 / pi;
end

function output = intI(theta, v, x)
    output = exp(x * cos(theta)) .* cos(v * theta);
end

function [output, R] = f(R, H, E, kappas, p, c3kappas, c3_0, M)
    output = 0;
    N = size(E, 1);
    for i = 1 : N
        for j = i + 1 : N
            if(E(i, j) == 1)
                temp = H(:, :, i, j) * R.U(:, :, j);
                temp2 = sum(sum(R.U(:, :, i) .* temp));
                output = output + log(p / c3kappas(i, j) * exp(kappas(i, j) * temp2) + (1 - p) / c3_0);
            end
        end
    end
    output = -output / M;
end

function [output, R] = gf(R, H, E, kappas, p, c3kappas, c3_0, M)
    n = size(R.U, 1);
    N = size(E, 1);
    l = zeros(N, N);
    w = zeros(N, N);
    for i = 1 : N
        for j = i + 1 : N
            temp = H(:, :, i, j) * R.U(:, :, j);
            temp2 = sum(sum(R.U(:, :, i) .* temp));
            l(i, j) = p / c3kappas(i, j) * exp(kappas(i, j) * temp2);
            w(i, j) = kappas(i, j) * l(i, j) / (l(i, j) + (1 - p) / c3_0);
        end
    end
    output.TV = zeros(n, n, N);
    for k = 1 : N
        for i = 1 : k - 1
            if(E(i, k) == 1)
                output.TV(:, :, k) = output.TV(:, :, k) + w(i, k) * H(:, :, i, k)' * R.U(:, :, i);
            end
        end
        for j = k + 1 : N
            if(E(k, j) == 1)
                output.TV(:, :, k) = output.TV(:, :, k) + w(k, j) * H(:, :, k, j) * R.U(:, :, j);
            end
        end
        output.TV(:, :, k) = output.TV(:, :, k) - R.U(:, :, k) * (R.U(:, :, k)' * output.TV(:, :, k) + output.TV(:, :, k)' * R.U(:, :, k)) / 2;
    end
    output.TV = -output.TV / M;
    output = embedded_intr_c(R, output);
end

function [output, R] = gf_full(R, H, E, kappas, p, c3kappas, c3_0, M)
    n = size(R.U, 1);
    N = size(E, 1);
    l = zeros(N, N);
    w = zeros(N, N);
    for i = 1 : N
        for j = i + 1 : N
            l(i, j) = p / c3kappas(i, j) * exp(kappas(i, j) * trace(R.U(:, :, i)' * H(:, :, i, j) * R.U(:, :, j)));
            w(i, j) = kappas(i, j) * l(i, j) / (l(i, j) + (1 - p) / c3_0);
        end
    end
    output.TV = zeros(n, n, N);
    for k = 1 : N
        for i = 1 : k - 1
            if(E(i, k) == 1)
                output.TV(:, :, k) = output.TV(:, :, k) + w(i, k) * H(:, :, i, k)' * R.U(:, :, i);
            end
        end
        for j = k + 1 : N
            if(E(k, j) == 1)
                output.TV(:, :, k) = output.TV(:, :, k) + w(k, j) * H(:, :, k, j) * R.U(:, :, j);
            end
        end
        output.TV(:, :, k) = output.TV(:, :, k) - R.U(:, :, k) * (R.U(:, :, k)' * output.TV(:, :, k) + output.TV(:, :, k)' * R.U(:, :, k)) / 2;
    end
    
    output.TV = -output.TV / M;
end

function output = fhessianv_full(x, v,  H, E, kappas, p, c3kappas, c3_0, M)
    n = size(x.U, 1);
    N = size(E, 1);
    l = zeros(N, N);
    w = zeros(N, N);
    dw = zeros(N, N);
    for i = 1 : N
        for j = i + 1 : N
            l(i, j) = p / c3kappas(i, j) * exp(kappas(i, j) * trace(x.U(:, :, i)' * H(:, :, i, j) * x.U(:, :, j)));
            w(i, j) = kappas(i, j) * l(i, j) / (l(i, j) + (1 - p) / c3_0);
            dw(i, j) = kappas(i, j) * (1 - p) / c3_0 / (l(i, j) + (1 - p) / c3_0)^2 * l(i, j) * kappas(i, j) * (sum(sum(v.TV(:, :, i) .* (H(:, :, i, j) * x.U(:, :, j)))) + sum(sum(x.U(:, :, i) .* (H(:, :, i, j) * v.TV(:, :, j)))));
        end
    end
    output.TV = zeros(n, n, N);
    for k = 1 : N
        Vk = zeros(n, n);
        DVketa = zeros(n, n);
        for i = 1 : k - 1
            if(E(i, k) == 1)
                Vk = Vk + w(i, k) * H(:, :, i, k)' * x.U(:, :, i);
                DVketa = DVketa + w(i, k) * H(:, :, i, k)' * v.TV(:, :, i) + dw(i, k) * H(:, :, i, k)' * x.U(:, :, i);
            end
        end
        for j = k + 1 : N
            if(E(k, j) == 1)
                Vk = Vk + w(k, j) * H(:, :, k, j) * x.U(:, :, j);
                DVketa = DVketa + w(k, j) * H(:, :, k, j) * v.TV(:, :, j) + dw(k, j) * H(:, :, k, j) * x.U(:, :, j);
            end
        end
        output.TV(:, :, k) = (DVketa - v.TV(:, :, k) * Vk' * x.U(:, :, k) - x.U(:, :, k) * Vk' * v.TV(:, :, k) - x.U(:, :, k) * DVketa' * x.U(:, :, k)) * 0.5;
        output.TV(:, :, k) = output.TV(:, :, k) - x.U(:, :, k) * (x.U(:, :, k)' * output.TV(:, :, k) + output.TV(:, :, k)' * x.U(:, :, k)) / 2;
    end
    output.TV = -output.TV / M;
end

% function output = fhessianv_full(x, v,  H, E, kappas, p, c3kappas, c3_0, M)
%     [n, n, N] = size(x);
%     gfx = gf_full(x, H, E, kappas, p, c3kappas, c3_0, M);
%     vv = reshape(v, n * n * N, 1);
%     for i = 1 : n * n * N
%         if(sign(vv(i)) ~= 0)
%             s = sign(vv(i));
%             break;
%         end
%     end
%     c = 1e-4 * s / sqrt(inpro_full(x, v, v));
%     gfdx = gf_full(R_Exp_full(x, c * v), H, E, kappas, p, c3kappas, c3_0, M);
% %     gfdx = gf_full(R_full(x, c * v), H, E, kappas, p, c3kappas, c3_0, M);
%     for i = 1 : N
%         gfdx(:, :, i) = gfdx(:, :, i) - x(:, :, i) * (x(:, :, i)' * gfdx(:, :, i) + gfdx(:, :, i)' * x(:, :, i)) / 2;
%     end
% 
%     output = (gfdx - gfx) / c;
% end

function output = R_Exp(x, eta)
    eta = embedded_full_c(x, eta);
    [n, n, N] = size(x.U);
    output.U = zeros(n, n, N);
    for i = 1 : N
        output.U(:, :, i) = x.U(:, :, i) * expm(x.U(:, :, i)' * eta.TV(:, :, i));
    end
    
    for i = 1 : N
        output.U(:, :, i) = qf(output.U(:, :, i));
    end
end

function output = R_Exp_full(x, eta)
    [n, n, N] = size(x.U);
    output.U = zeros(n, n, N);
    for i = 1 : N
        output.U(:, :, i) = x.U(:, :, i) * expm(x.U(:, :, i)' * eta.TV(:, :, i));
    end
    
    for i = 1 : N
        output.U(:, :, i) = qf(output.U(:, :, i));
    end
end

function output = Hv(x, H, v)
    output.TV = H * v.TV;
end

function output = phi()
    output = 1;
end

function output = R(x, eta)
    eta = embedded_full_c(x, eta);
    [n, n, N] = size(x.U);
    output.U = zeros(n, n, N);
    for i = 1 : N
        output.U(:, :, i) = qf(x.U(:, :, i) + eta.TV(:, :, i));
    end
end

function output = R_full(x, eta)
    [n, n, N] = size(x.U);
    output.TV = zeros(n, n, N);
    for i = 1 : N
        output.U(:, :, i) = qf(x.U(:, :, i) + eta.TV(:, :, i));
    end
end

function output = qf(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    diagR(diagR == 0) = 1;
    output = output * spdiags(sign(diagR), 0, k, k);
end

function output = beta_qf(x1, eta, x2)
    output = eta.beta;
%     Tee = Tranv_R(x1, eta, x2, eta);
%     output = sqrt(inpro(x1, eta, eta) / inpro(x2, Tee, Tee));
end

function output = beta(x1, eta, x2)
    output = 1;
end

function output = Tranv_R(x, d, y, v)
    d = embedded_full_c(x, d);
    v = embedded_full_c(x, v);
    [n, n, N] = size(x.U);
    output.TV = zeros(n, n, N);
    for i = 1 : N
        output.TV(:, :, i) = Tranv_R_qf(x.U(:, :, i), d.TV(:, :, i), y.U(:, :, i), v.TV(:, :, i));
    end
    output = embedded_intr_c(y, output);
end

function output = Tranv_R_qf(x, d, y, v)
    temp = inv(y' * (x + d));
    temp2 = v * temp;
    temp3 = y' * temp2;
    output = y * rho_skew(temp3) + temp2 - y * temp3;
end

function output = rho_skew(A)
    output = tril(A, -1) - tril(A, -1)';
end

function output = inpro(x, v1, v2)
    output = v1.TV' * v2.TV;
end

function output = inpro_full(x, v1, v2)
    [n, n, N] = size(x.U);
    output = 0;
    for i = 1 : N
        output = output + sum(sum(v1.TV(:, :, i) .* v2.TV(:, :, i)));%%-- trace(v1(:, :, i)' * v2(:, :, i));
    end
end

function [output, d, x, y] = Tranv(x, d, y, v)
    output = v;
end

function [output, d, x, y] = invTranv(x, d, y, v)
    output = v;
end

function output = TranH(x, d, y, H)
    output = H;
end

function [output, d, x, y] = Tranv_qf(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        Tee = Tranv_R(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        bc2.TV = d.beta * Tee.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * vv1.TV * (vv1.TV' * v.TV) / (vv1.TV' * vv1.TV);
    output.TV = output.TV - 2 * vv2.TV * (vv2.TV' * output.TV) / (vv2.TV' * vv2.TV);
end

function [output, d, x, y] = invTranv_qf(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        Tee = Tranv_R(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        bc2.TV = d.beta * Tee.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * vv2.TV * (vv2.TV' * v.TV) / (vv2.TV' * vv2.TV);
    output.TV = output.TV - 2 * vv1.TV * (vv1.TV' * output.TV) / (vv1.TV' * vv1.TV);
end

function output = TranH_qf(x, d, y, H)
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output = H - 2 * (H * vv1.TV) * (vv1.TV' / (vv1.TV' * vv1.TV));
    output = output - 2 * (output * vv2.TV) * (vv2.TV' / (vv2.TV' * vv2.TV));
    output = output - 2 * vv1.TV * ((vv1.TV' * output) / (vv1.TV' * vv1.TV));
    output = output - 2 * vv2.TV * ((vv2.TV' * output) / (vv2.TV' * vv2.TV));
end

function output = rank1operator(x, v1, v2)
    output = v1.TV * v2.TV';
end

function output = operadd(x, H1, H2)
    output = H1 + H2;
end

function output = build_bases(x)
    [n, n, N] = size(x.U);
    output = zeros(n * n, n * (n - 1) / 2 * N);
    bc = n * (n - 1) / 2;
    for i = 1 : N
        output(:, (i - 1) * bc + 1 : i * bc) = build_bases_embedded(x.U(:, :, i));
    end
end

function output = build_bases_embedded(x)
    [n, p] = size(x);
    x_perp = null(x');
    x = x / sqrt(2);
    d = n * p - p * (p + 1) / 2;
    output = zeros(n * p, d);
    col = 1;
    for i = 2 : p
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
            output((j - 1) * n + 1 : j * n, col) = x(:, i - 1);
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp(1 : n, 1 : n - p);
        col = col + n - p;
    end
end

function output = embedded_intr_c(x, eta)
    r2 = sqrt(2);
    [n, n, N] = size(x.U);
    output.TV = zeros(0.5 * n * (n - 1) * N, 1);
    blocksize = 0.5 * n * (n - 1);
    for i = 1 : N
        omega = x.U(:, :, i)' * eta.TV(:, :, i);
        omega = - sqrt(2) * omega;
        indx = find(tril(ones(n, n), -1));
        output.TV(blocksize * (i - 1) + 1 : blocksize * i) = omega(indx);
    end
end

function output = embedded_full_c(x, v)
    r2 = sqrt(2);
    [n, n, N] = size(x.U);
    omega = tril(ones(n, n), -1);
    indx = find(omega);
    blocksize = 0.5 * n * (n - 1);
    output.TV = zeros(n, n, N);
    for i = 1 : N
        omega = zeros(n, n);
        omega(indx) = v.TV(blocksize * (i - 1) + 1 : blocksize * i) / r2;
        omega = omega' - omega;
        output.TV(:, :, i) = x.U(:, :, i) * omega;
    end
end
