function fs = grassmann(retraction, vector_transport)
% This function return some necessary function handles of the Grassmann manifold
% 
% INPUT:
% Grassmann manifold : Gr(p, n)
%     params.retraction : 
%         '1', exponential mapping
%         '2', qf factorization : qf(X + eta)
%         '3', constructed retraction for intrinsic method of Iso vector transport 
%     params.vector_transport : the isometric vector transport
%         '1', noniso: vector transport by projection
%         '2', Iso: parallel transport
%         '3', Iso: direct-rotation with basis on tangent space
%         '4', Iso: direct-rotation with basis on normal space
%         '5', Iso: smooth basis on tangent space
%         '6', Iso: smooth basis on normal space
%         '7', Iso: intrinsic method
%         '8',Locking-Iso(qf): direct-rotation with basis on tangent space
%         '9',Locking-Iso(qf): direct-rotation with basis on normal space
%         '10',Locking-Iso(qf): smooth basis on tangent space
%         '11',Locking-Iso(qf): smooth basis on normal space
%         '12',Locking-Iso(qf): construct vector transport
%         '13',Locking-Iso(qf): intrinsic method
% 
% OUTPUT:
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.invBv(x, B, v) : return a tangent vector vv on T_x M which satisfies B vv = v.
%     fns.phi() : return the coefficient of RBroyden Family update formula.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.SRTV(x, m, seed) : return m random tangent vectors in T_x M, seed is the random seed.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
%     fns.Tranv_Params(x1, d, x2, v, TranvParams) : return the tangent vector which is the same as fns.Tranv(x1, d, x2, v), 
%                                                   TranvParams is some parameters such that the cost becomes smaller and 
%                                                   are given by fns.Tranv or fns.invTranv.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
% 
% By Wen Huang
%     fprintf('Grassmann manifold\n');
    fs.Hv = @Hv;
    fs.phi = @phi;
    fs.operadd = @operadd;
    fs.proj = @proj;
    fs.STV = @STV;
    fs.TVaddTV = @TVaddTV;

    if(vector_transport == 7 || vector_transport == 13 || vector_transport == 14)
        fs.Hv = @Hv_intr;
        fs.invBv = @invBv_intr;
        fs.proj = @proj_intr;
        fs.SRTV = @SRTV_intr;
        fs.rank1operator = @rank1operator_intr;
        fs.inpro = @inpro_intr;
        if(retraction == 1)
%             fprintf('exponential mapping\n')
            fs.R = @R_exp_intr;
            fs.beta = @beta_exp_intr;
        elseif(retraction == 2)
%             fprintf('retraction: qf\n')
            fs.R = @R_qf_intr;
            fs.beta = @beta_qf_intr;
        elseif(retraction == 3)
%             fprintf('retraction: construct retraction for intrinsic method of vector transport\n')
            fs.R = @R_construct_intr;
            fs.beta = @beta_construct_intr;
        end
    else
        fs.SRTV = @SRTV;
        fs.rank1operator = @rank1operator;
        fs.inpro = @inpro;

        if(retraction == 1)
%             fprintf('exponential mapping\n')
            fs.R = @R_exp;
            fs.beta = @beta_exp;
        elseif(retraction == 2)
%             fprintf('retraction: qf\n')
            fs.R = @R_qf;
            fs.beta = @beta_qf;
        end
    end
    
    if(vector_transport == 1)
%         fprintf('nonIso: vector transport by projection \n')
        fs.Tranv = @Tranv_proj;
        fs.Tranv_Params = @Tranv_proj;
        fs.invTranv = @invTranv_proj;
        fs.TranH = @TranH_proj;
    elseif(vector_transport == 2)
%         fprintf('Iso: parallel transport\n')
        fs.Tranv = @Tranv_parallel;
        fs.Tranv_Params = @Tranv_parallel;
        fs.invTranv = @invTranv_parallel;
        fs.TranH = @TranH_parallel;
    elseif(vector_transport == 3)
%         fprintf('Iso: direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano;
        fs.Tranv_Params = @Tranv_cano_Params;
        fs.invTranv = @invTranv_cano;
        fs.TranH = @TranH_cano;
    elseif(vector_transport == 4)
%         fprintf('Iso: direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn;
        fs.Tranv_Params = @Tranv_cano_NonIn_Params;
        fs.invTranv = @invTranv_cano_NonIn;
        fs.TranH = @TranH_cano_NonIn;
    elseif(vector_transport == 5)
%         fprintf('Iso: smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth;
        fs.Tranv_Params = @Tranv_smooth_Params;
        fs.invTranv = @invTranv_smooth;
        fs.TranH = @TranH_smooth;
    elseif(vector_transport == 6)
%         fprintf('Iso: smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn;
        fs.Tranv_Params = @Tranv_smooth_NonIn_Params;
        fs.invTranv = @invTranv_smooth_NonIn;
        fs.TranH = @TranH_smooth_NonIn;
    elseif(vector_transport == 7)
%         fprintf('Iso: intrinsic method \n')
        fs.Tranv = @Tranv_smooth_intr;
        fs.Tranv_Params = @Tranv_smooth_intr_Params;
        fs.invTranv = @invTranv_smooth_intr;
        fs.TranH = @TranH_smooth_intr;
    elseif(vector_transport == 8)
%         fprintf('Locking-Iso(qf): direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano_qf;
        fs.Tranv_Params = @Tranv_cano_qf_Params;
        fs.invTranv = @invTranv_cano_qf;
        fs.TranH = @TranH_cano_qf;
    elseif(vector_transport == 9)
%         fprintf('Locking-Iso(qf): direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn_qf;
        fs.Tranv_Params = @Tranv_cano_NonIn_qf_Params;
        fs.invTranv = @invTranv_cano_NonIn_qf;
        fs.TranH = @TranH_cano_NonIn_qf;
    elseif(vector_transport == 10)
%         fprintf('Locking-Iso(qf): smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth_qf;
        fs.Tranv_Params = @Tranv_smooth_qf_Params;
        fs.invTranv = @invTranv_smooth_qf;
        fs.TranH = @TranH_smooth_qf;
    elseif(vector_transport == 11)
%         fprintf('Locking-Iso(qf): smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn_qf;
        fs.Tranv_Params = @Tranv_smooth_NonIn_qf_Params;
        fs.invTranv = @invTranv_smooth_NonIn_qf;
        fs.TranH = @TranH_smooth_NonIn_qf;
    elseif(vector_transport == 12)
%         fprintf('Locking-Iso(qf): construct vector transport \n')
        fs.Tranv = @Tranv_qf;
        fs.Tranv_Params = @Tranv_qf_Params;
        fs.invTranv = @invTranv_qf;
        fs.TranH = @TranH_qf;
    elseif(vector_transport == 13)
%         fprintf('Locking-Iso(qf): intrinsic method \n')
        fs.Tranv = @Tranv_qf_intr;
        fs.Tranv_Params = @Tranv_qf_intr_Params;
        fs.invTranv = @invTranv_qf_intr;
        fs.TranH = @TranH_qf_intr;
    end
end

function output = STV(x, a, eta)
    output.TV = a * eta.TV;
end

function output = TVaddTV(x, eta1, eta2)
    output.TV = eta1.TV + eta2.TV;
end

function output = Hv(x, H, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(H * v.TV, n, p);
end

function output = Hv_intr(x, H, v)
    output.TV = H * v.TV;
end

function output = phi(x, y, s, B, inp_s_y, sBs, fns, v, times)
    output = 1;
end

function output = SRTV(x, m, seed)
    rand('state', seed);
    basis_x = build_bases(x);
    [nn, d] = size(basis_x);
    [n, p] = size(x.U);
    u = rand(d, m) - 0.5;
    TV = basis_x * u;
    for i = 1 : m
        output{i}.TV = reshape(TV(:, i), n, p);
    end
end

function output = SRTV_intr(x, m, seed)
    rand('state', seed);
    [n, p] = size(x.U);
    d = (n - p) * p;
    u = rand(d, m) - 0.5;
    for i = 1 : m
        output{i}.TV = u(:, i);
    end
end

function output = proj(x, eta)
    output.TV = eta.TV - x.U * (x.U' * eta.TV);
end

function output = proj_intr(x, eta)
    output = eta;
end

function output = rank1operator(x, v1, v2)
    [n, p] = size(x.U);
    vv1 = reshape(v1.TV, n * p, 1);
    vv2 = reshape(v2.TV, n * p, 1);
    output = vv1 * vv2';
end

function output = rank1operator_intr(x, v1, v2)
    output = v1.TV * v2.TV';
end

function output = operadd(x, H1, H2)
    output = H1 + H2;
end

function output = inpro(x, v1, v2)
    output = sum(sum(v1.TV .* v2.TV));
end

function output = inpro_intr(x, v1, v2)
    output = v1.TV' * v2.TV;
end

%% exponential mapping when it is considered as an embedded manifold
function output = R_exp(x, eta)
    [U, S, V] = svd(eta.TV, 0);
    S = diag(S);
    output.U = (x.U * (V * diag(cos(S))) + U * diag(sin(S))) * V';
end

function output = R_exp_intr(x, eta)
    eta = full_c(x, eta);
    [U, S, V] = svd(eta.TV, 0);
    S = diag(S);
    output.U = (x.U * (V * diag(cos(S))) + U * diag(sin(S))) * V';
    output.Up = null(output.U');
end

function output = beta_exp(x1, eta, x2)
    output = 1;
end

function output = beta_exp_intr(x1, eta, x2)
    output = 1;
end

%% qf and keep in an orthogonal section
function output = R_qf(x, eta)
    output.U = qf(x.U + eta.TV);
%     [U, S, V] = svd(x' * output);
%     Ot = V * U';
%     output = output * Ot;
end

function output = beta_qf(x1, eta, x2)
    output = eta.beta;
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro(x1, eta, eta) / inpro(x2, Tee, Tee));
end

function output = R_qf_intr(x, eta)
    eta = full_c(x, eta);
    output.U = qf(x.U + eta.TV);
    output.Up = null(output.U');
    
%     out = qf(x{1} + eta);
%     [U, S, V] = svd(x{1}' * out);
%     Ot = V * U';
%     output{1} = out * Ot;
%     output{2} = null(output{1}');
end

function output = beta_qf_intr(x1, eta, x2)
    if(isfield(eta, 'beta'))
        output = eta.beta;
    else
        output = 1;
    end
%     eta = full_c(x1, eta);
%     Tee = Tranv_R_qf(x1{1}, eta, x2{1}, eta);
%     output = sqrt(inpro(x1{1}, eta, eta) / inpro(x2{1}, Tee, Tee));
end

function output = Tranv_R_qf(x, d, y, v)
    temp = inv(y.U' * (x.U + d.TV));
    temp2 = v.TV * temp;
    output.TV = temp2 - y.U * (y.U' * temp2);
%     temp = inv(y' * (x + d));
%     temp2 = v * temp;
%     temp3 = y' * temp2;
%     output = y * (rho_skew(temp3) - temp3) + temp2;
end

function output = rho_skew(A)
    output = tril(A, -1) - tril(A, -1)';
end

% %% polar
% function output = R_polar(x, eta)
%     [n, p] = size(x);
%     output = (x + eta) * (eye(p) + eta' * eta)^(-0.5);
% end

function output = R_construct_intr(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
%     omega = tril(ones(p, p), -1);
%     indx = find(omega);
%     omega(indx) = eta(1 : 0.5 * p * (p - 1));
%     omega = omega' - omega;
    K = reshape(eta.TV, (n - p), p);
    M = ([[zeros(p, p);K], [-K'; zeros(n - p, n - p)]]);
    v = [x1, x2];
    
    v = v * expm(M);
    
%     v = expv_Stiefel(1, M, v);
    v1 = v(1 : n * p);
    v2 = v(n * p + 1 : end);
    output.U = reshape(v1, n, p);
    output.Up = reshape(v2, n, n - p);
% error('asf');
end

function output = beta_construct_intr(x1, eta, x2)
    output = 1;
end

% function output = R_section_construct_intr(x, eta)
%     x1 = x{1};
%     x2 = x{2};
%     O = x{3};
%     xs = x{4};
% 
%     [n, p] = size(x1);
% %     omega = tril(ones(p, p), -1);
% %     indx = find(omega);
% %     omega(indx) = eta(1 : 0.5 * p * (p - 1));
% %     omega = omega' - omega;
%     K = reshape(eta, (n - p), p);
%     M = ([[zeros(p, p);K], [-K'; zeros(n - p, n - p)]]);
%     v = [x1, x2];
%     
%     v = v * expm(M);
%     
% %     v = expv_Stiefel(1, M, v);
%     v1 = v(1 : n * p);
%     v2 = v(n * p + 1 : end);
%     output{1} = reshape(v1, n, p);
%     output{2} = reshape(v2, n, n - p);
%     
%     [U, S, V] = svd(output{1}' * xs);
%     if(S(end, end) > 1e-10)
%         output{3} = U * V';
%         output{4} = xs;
%     else
%         error('R_section_construct_intr.');
%     end
%     
% % error('asf');
% end

%% nonIso: vector transport by projection
function [output, d, x, y] = Tranv_proj(x, d, y, v)
    output.TV = v.TV - y.U * (y.U' * v.TV);
end

function [output, d, x, y] = invTranv_proj(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases(x);
    end
    perp_basis_x = x.perp_basis;
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases(y);
    end
    perp_basis_y = y.perp_basis;
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(perp_basis_x' * perp_basis_y));
    end
    
    output.TV = reshape(v.TV, n * p, 1) - perp_basis_y * (d.pinvPxPy * (perp_basis_x' * reshape(v.TV, n * p, 1)));
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_proj(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases(x);
    end
    perp_basis_x = x.perp_basis;
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases(y);
    end
    perp_basis_y = y.perp_basis;
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(perp_basis_x' * perp_basis_y));
    end
    output = H - H * perp_basis_y * d.pinvPxPy * perp_basis_x';
    output = output - perp_basis_y * (perp_basis_y' * output);
%     output = H - H * perp_basis_x * perp_basis_x';
%     output = output - perp_basis_x * pinv(perp_basis_y' * perp_basis_x) * (perp_basis_y' * output);
end

%% parallel vector transport when it is considered as an embedded manifold
function [output, d, x, y] = Tranv_parallel(x, d, y, v)
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'S'))
        [d.U, S, d.TV] = svd(d.TV, 0);
        d.S = diag(S);
    end
    temp = d.U' * v.TV;
    output.TV = (- x.U * (d.TV * diag(sin(d.S))) * temp + d.U * diag(cos(d.S)) * temp) + v.TV - d.U * temp;
end

function [output, d, x, y] = invTranv_parallel(x, d, y, v)
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'S'))
        [d.U, S, d.TV] = svd(d.TV, 0);
        d.S = diag(S);
    end
    U = d.U;
    V = d.TV;
    S = d.S;
    temp = diag(S) * V';
    invd = x.U * V * diag(sin(S)) * temp - U * diag(cos(S)) * temp;
    [U, S, V] = svd(invd, 0);
    temp = U' * v.TV;
    S = diag(S);
    output.TV = (- y.U * (V * diag(sin(S))) * temp + U * diag(cos(S)) * temp) + v.TV - U * temp;
end

function output = TranH_parallel(x, d, y, H)
    [n, p] = size(x.U);
    dim = p * (n - p);
    bases = build_bases(x);
    P_bases = zeros(n * p, dim);
    P_Hb = zeros(n * p, dim);
    
    U = d.U;
    V = d.TV;
    S = d.S;
    C = (- x.U * (V * diag(sin(S))) * U' + U * diag(cos(S)) * U') + eye(size(U, 1)) - U * U';
    
    for i = 1 : dim
        P_bases(:, i) = reshape(C * reshape(bases(:, i), n, p), n * p, 1);
    end
    Hb = H * bases;
    for i = 1 : dim
        P_Hb(:, i) = reshape(C * reshape(Hb(:, i), n, p), n * p, 1);
    end
    output = P_Hb * P_bases';
end

%% Emb Iso: direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [d.U, S, d.TV] = svd(full(ybases' * xbases));
    end
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (d.U * (d.TV' * (xbases' * v.TV))), n, p);
end

function [output, d, x, y] = invTranv_cano(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [d.U, S, d.TV] = svd(full(ybases' * xbases));
    end
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (d.TV * (d.U' * (ybases' * v.TV))), n, p);
end

function output = TranH_cano(x, d, y, H)
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    temp = U * V';
    output = ybases * (temp * (xbases' * (H * xbases)) * temp') * ybases';
end

%% Emb Iso: direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases(y);
    end
    By_perp = y.B_perp;
    
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    output.TV = reshape((v.TV + (d.tilde_Q * (d.U * d.TV') - d.Q) * (d.Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_cano_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases(y);
    end
    By_perp = y.B_perp;
    
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    output.TV = reshape((v.TV + (d.Q * (d.TV * d.U') - d.tilde_Q) * (d.tilde_Q' * v.TV)), n, p);
end

function output = TranH_cano_NonIn(x, d, y, H)
    output = H + H * (d.Q * (d.TV * d.U') - d.tilde_Q) * d.tilde_Q';
    output = output + (d.tilde_Q * (d.U * d.TV') - d.Q) * (d.Q' * output);
end

%% Emb Iso: smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (xbases' * v.TV), n, p);
end

function [output, d, x, y] = invTranv_smooth(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (ybases' * v.TV), n, p);
end

function output = TranH_smooth(x, d, y, H)
    xbases = x.bases;
    ybases = y.bases;
    output = ybases * (xbases' * (H * xbases)) * ybases';
end

%% Emb Iso: smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases(y);
    end
    By_perp = y.B_perp;
    
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    output.TV = reshape((v.TV - (d.tilde_Q + d.Q) * (d.Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_smooth_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    output.TV = reshape((v.TV - (d.Q + d.tilde_Q) * (d.tilde_Q' * v.TV)), n, p);
end

function output = TranH_smooth_NonIn(x, d, y, H)
    output = H - H * (d.Q + d.tilde_Q) * d.tilde_Q';
    output = output - (d.tilde_Q + d.Q) * (d.Q' * output);
end

%% Emb Iso: intrinsic method
function [output, d, x, y] = Tranv_smooth_intr(x, d, y, v)
    output = v;
end

function [output, d, x, y] = invTranv_smooth_intr(x, d, y, v)
    output = v;
end

function output = TranH_smooth_intr(x, d, y, H)
    output = H;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano_qf(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro(y, nu1, TIv) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro(y, nu2, output) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_qf(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(x, 'bases') || ~isfield(y, 'bases') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        [eps1, d, x, y] = Tranv_cano(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro(y, nu2, v) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro(y, nu1, output) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(x.bases * (d.TV * (d.U' * (y.bases' * output.TV))), size(x.U));
end

function output = TranH_cano_qf(x, d, y, H)
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    temp = U * V';
    temp = (temp * (xbases' * (H * xbases)) * temp') * ybases';
    temp = temp - 2 / inpro(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybases - nu1.TV * (nu1.TV' * ybases * (2 / inpro(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn_qf(x, d, y, v)
    [TIv, TranvParams] = Tranv_cano_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro(y, nu1, TIv) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro(y, nu2, output) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_NonIn_qf(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro(y, nu2, v) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro(y, nu1, output) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV + (d.Q * (d.TV * d.U') - d.tilde_Q) * (d.tilde_Q' * output.TV)), size(x.U));
end

function output = TranH_cano_NonIn_qf(x, d, y, H)
    nu1 = d.nu1;
    nu2 = d.nu2;
    output = H + H * (d.Q * (d.TV * d.U') - d.tilde_Q) * d.tilde_Q';
    output = output + (d.tilde_Q * (d.U * d.TV') - d.Q) * (d.Q' * output);
    output = output - 2 / inpro(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth_qf(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro(y, nu1, TIv) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro(y, nu2, output) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_qf(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro(y, nu2, v) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro(y, nu1, output) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(x.bases * (y.bases' * output.TV), size(x.U));
end

function output = TranH_smooth_qf(x, d, y, H)
    nu1 = d.nu1;
    nu2 = d.nu2;
    temp = x.bases' * (H * x.bases) * y.bases';
    temp = temp - 2 / inpro(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = y.bases - nu1.TV * (nu1.TV' * y.bases * (2 / inpro(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn_qf(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro(y, nu1, TIv) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro(y, nu2, output) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_NonIn_qf(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Q') || ~isfield(d, 'tilde_Q'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro(y, nu2, v) / inpro(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro(y, nu1, output) / inpro(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV - (d.Q + d.tilde_Q) * (d.tilde_Q' * output.TV)), size(d.TV));
end

function output = TranH_smooth_NonIn_qf(x, d, y, H)
    nu1 = d.nu1;
    nu2 = d.nu2;
    output = H - H * (d.Q + d.tilde_Q) * d.tilde_Q';
    output = output - (d.tilde_Q + d.Q) * (d.Q' * output);
    output = output - 2 / inpro(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): construct vector transport
function [output, d, x, y] = Tranv_qf(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = x.bases' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        bc2.TV = d.beta * y.bases' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    tempv = x.bases' * reshape(v.TV, n * p, 1);
    tempv = tempv - 2 * vv1.TV * (vv1.TV' * tempv) / (vv1.TV' * vv1.TV);
    tempv = tempv - 2 * vv2.TV * (vv2.TV' * tempv) / (vv2.TV' * vv2.TV);
    output.TV = y.bases * tempv;
    output.TV = reshape(output.TV, n, p);
end

function [output, d, x, y] = invTranv_qf(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases(x);
    end
    if(~isfield(y, 'bases'))
        y.bases = build_bases(y);
    end
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro(x, d, d) / inpro(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    tempv = bases_y' * reshape(v.TV, n * p, 1);
    tempv = tempv - 2 * vv2.TV * (vv2.TV' * tempv) / (vv2.TV' * vv2.TV);
    tempv = tempv - 2 * vv1.TV * (vv1.TV' * tempv) / (vv1.TV' * vv1.TV);
    output.TV = bases_x * tempv;
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_qf(x, d, y, H)
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output = x.bases' * (H * x.bases);
    output = output - 2 * (output * vv1.TV) * vv1.TV' / (vv1.TV' * vv1.TV);
    output = output - 2 * (output * vv2.TV) * vv2.TV' / (vv2.TV' * vv2.TV);
    output = output - 2 * vv1.TV * (vv1.TV' * output) / (vv1.TV' * vv1.TV);
    output = output - 2 * vv2.TV * (vv2.TV' * output) / (vv2.TV' * vv2.TV);
    output = output * y.bases';
    output = y.bases * output;
end

%% Emb Locking-Iso(qf): intrinsic approach
function [output, d, x, y] = Tranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro(x, full_d, full_d) / inpro(y, Tee, Tee));
        Tee_intr = intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * vv1.TV * (vv1.TV' * v.TV) / (vv1.TV' * vv1.TV);
    output.TV = output.TV - 2 * vv2.TV * (vv2.TV' * output.TV) / (vv2.TV' * vv2.TV);
end

function [output, d, x, y] = invTranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro(x, full_d, full_d) / inpro(y, Tee, Tee));
        Tee_intr = intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * vv2.TV * (vv2.TV' * v.TV) / (vv2.TV' * vv2.TV);
    output.TV = output.TV - 2 * vv1.TV * (vv1.TV' * output.TV) / (vv1.TV' * vv1.TV);
end

function output = TranH_qf_intr(x, d, y, H)
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output = H - 2 * (H * vv1.TV) * (vv1.TV' / (vv1.TV' * vv1.TV));
    output = output - 2 * (output * vv2.TV) * (vv2.TV' / (vv2.TV' * vv2.TV));
    output = output - 2 * vv1.TV * ((vv1.TV' * output) / (vv1.TV' * vv1.TV));
    output = output - 2 * vv2.TV * ((vv2.TV' * output) / (vv2.TV' * vv2.TV));
end

%% assistant function------------------------------------------------------
function [output, R] = qf(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    diagR(diagR == 0) = 1;
    output = output * spdiags(sign(diagR), 0, k, k);
end

function [output, R] = qf_thin(B)
%     output = orth(B);
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    output = output(:, abs(diagR) > sqrt(eps));
    output = full(output * spdiags(sign(diagR), 0, size(output, 2), size(output, 2)));
end

function output = build_perp_bases(x)
    [n, p] = size(x.U);
    xs = mat2cell(repmat(x.U, 1, p), [n], p * ones(1, p));
    xs{1} = sparse(xs{1});
    output = blkdiag(xs{:});
end

function output = build_bases(x)
%     [n, p] = size(x);
%     x_perp = null(x');
%     output = sparse(n * p, p * (n - p));
%     for i = 1 : p
%         output((i - 1) * n + 1 : i * n, (i - 1) * (n - p) + 1 : i * (n - p)) = x_perp;
%     end

    [n, p] = size(x.U);
    if(~isfield(x, 'Up'))
        x.Up = null(x.U');
    end
    x_perp = x.Up;
    x_perps = mat2cell(repmat(x_perp, 1, p), [n], (n - p) * ones(1, p));
    x_perps{1} = sparse(x_perps{1});
    output = blkdiag(x_perps{:});
end

function output = intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    K = x2' * eta.TV; % K
    output.TV = reshape(K, [], 1);
end

function output = full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    output.TV = x2 * reshape(v.TV, (n - p), p);
end
