function fs = orthogroup(retraction, vector_transport)
% This function return some necessary function handles of the orthogonal group
% 
% INPUT:
% Orthogonal group : O(n)
%     params.retraction : 
%         '1', exponential mapping
%         '2', qf factorization : qf(X + eta)
%     params.vector_transport : the isometric vector transport
%         '1', noniso: vector transport by projection
%         '2', Iso: parallel transport
%         '3', Iso: direct-rotation with basis on tangent space
%         '4', Iso: direct-rotation with basis on normal space
%         '5', Iso: smooth basis on tangent space
%         '6', Iso: smooth basis on normal space
%         '7', Iso: intrinsic method
%         '8',Locking-Iso(qf): direct-rotation with basis on tangent space
%         '9',Locking-Iso(qf): direct-rotation with basis on normal space
%         '10',Locking-Iso(qf): smooth basis on tangent space
%         '11',Locking-Iso(qf): smooth basis on normal space
%         '12',Locking-Iso(qf): construct vector transport
%         '13',Locking-Iso(qf): intrinsic method
% 
% OUTPUT:
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1
%     adds v2.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.invBv(x, B, v) : return a tangent vector vv on T_x M which satisfies B vv = v.
%     fns.phi(x, y, s, B, inp_s_y, sBs, fns, v, times) : return the coefficient of RBroyden Family update formula.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.SRTV(x, m, seed) : return m random tangent vectors in T_x M, seed is the random seed.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
%     fns.Tranv_Params(x1, d, x2, v, TranvParams) : return the tangent vector which is the same as fns.Tranv(x1, d, x2, v), 
%                                                   TranvParams is some parameters such that the cost becomes smaller and 
%                                                   are given by fns.Tranv or fns.invTranv.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
% 
% By Wen Huang

    fprintf('orthogonal group\n');
    fs.Hv = @Hv;
    fs.phi = @phi;
    fs.operadd = @operadd;
    fs.proj = @proj;
    fs.STV = @STV;
    fs.TVaddTV = @TVaddTV;
    fs.SRTV = @SRTV_embedded;
    fs.rank1operator = @rank1operator_embedded;
    fs.inpro = @inpro_embedded;

    if(vector_transport == 7 || vector_transport == 13)
        fs.Hv = @Hv_intr;
        fs.invBv = @invBv_intr;
        fs.proj = @proj_intr;
        fs.SRTV = @SRTV_embedded_intr;
        fs.rank1operator = @rank1operator_embedded_intr;
        fs.inpro = @inpro_embedded_intr;
        if(retraction == 1)
            fprintf('exponential mapping\n')
            fs.R = @R_exp_embedded_intr;
            fs.beta = @beta_exp_intr;
        elseif(retraction == 2)
            fprintf('retraction: qf\n')
            fs.R = @R_qf_embedded_intr;
            fs.beta = @beta_qf_embedded_intr;
        end
    else
        if(retraction == 1)
            fprintf('exponential mapping\n')
            fs.R = @R_exp_embedded;
            fs.beta = @beta_exp;
        elseif(retraction == 2)
            fprintf('retraction: qf\n')
            fs.R = @R_qf_embedded;
            fs.beta = @beta_qf_embedded;
        end
    end
    
    if(vector_transport == 1)
        fprintf('nonIso: vector transport by projection \n')
        fs.Tranv = @Tranv_proj;
%         fs.Tranv_Params = @Tranv_proj;
        fs.invTranv = @invTranv_proj;
        fs.TranH = @TranH_proj;
    elseif(vector_transport == 2)
        fprintf('Iso: parallel transport\n')
        fs.Tranv = @Tranv_parallel_embedded;
%         fs.Tranv_Params = @Tranv_parallel_embedded;
        fs.invTranv = @invTranv_parallel_embedded;
        fs.TranH = @TranH_parallel_embedded;
    elseif(vector_transport == 3)
        fprintf('Iso: direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano;
%         fs.Tranv_Params = @Tranv_cano_Params;
        fs.invTranv = @invTranv_cano;
        fs.TranH = @TranH_cano;
    elseif(vector_transport == 4)
        fprintf('Iso: direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn;
%         fs.Tranv_Params = @Tranv_cano_NonIn_Params;
        fs.invTranv = @invTranv_cano_NonIn;
        fs.TranH = @TranH_cano_NonIn;
    elseif(vector_transport == 5)
        fprintf('Iso: smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth;
%         fs.Tranv_Params = @Tranv_smooth_Params;
        fs.invTranv = @invTranv_smooth;
        fs.TranH = @TranH_smooth;
    elseif(vector_transport == 6)
        fprintf('Iso: smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn;
%         fs.Tranv_Params = @Tranv_smooth_NonIn_Params;
        fs.invTranv = @invTranv_smooth_NonIn;
        fs.TranH = @TranH_smooth_NonIn;
    elseif(vector_transport == 7)
        fprintf('Iso: intrinsic method \n')
        fs.Tranv = @Tranv_smooth_intr;
%         fs.Tranv_Params = @Tranv_smooth_intr_Params;
        fs.invTranv = @invTranv_smooth_intr;
        fs.TranH = @TranH_smooth_intr;
    elseif(vector_transport == 8)
        fprintf('Locking-Iso(qf): direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano_qf_embedded;
%         fs.Tranv_Params = @Tranv_cano_qf_embedded_Params;
        fs.invTranv = @invTranv_cano_qf_embedded;
        fs.TranH = @TranH_cano_qf_embedded;
    elseif(vector_transport == 9)
        fprintf('Locking-Iso(qf): direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn_qf_embedded;
%         fs.Tranv_Params = @Tranv_cano_NonIn_qf_embedded_Params;
        fs.invTranv = @invTranv_cano_NonIn_qf_embedded;
        fs.TranH = @TranH_cano_NonIn_qf_embedded;
    elseif(vector_transport == 10)
        fprintf('Locking-Iso(qf): smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth_qf_embedded;
%         fs.Tranv_Params = @Tranv_smooth_qf_embedded_Params;
        fs.invTranv = @invTranv_smooth_qf_embedded;
        fs.TranH = @TranH_smooth_qf_embedded;
    elseif(vector_transport == 11)
        fprintf('Locking-Iso(qf): smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn_qf_embedded;
%         fs.Tranv_Params = @Tranv_smooth_NonIn_qf_embedded_Params;
        fs.invTranv = @invTranv_smooth_NonIn_qf_embedded;
        fs.TranH = @TranH_smooth_NonIn_qf_embedded;
    elseif(vector_transport == 12)
        fprintf('Locking-Iso(qf): construct vector transport \n')
        fs.Tranv = @Tranv_qf_embedded;
%         fs.Tranv_Params = @Tranv_qf_embedded_Params;
        fs.invTranv = @invTranv_qf_embedded;
        fs.TranH = @TranH_qf_embedded;
    elseif(vector_transport == 13)
        fprintf('Locking-Iso(qf): intrinsic method \n')
        fs.Tranv = @Tranv_qf_intr;
%         fs.Tranv_Params = @Tranv_qf_intr_Params;
        fs.invTranv = @invTranv_qf_intr;
        fs.TranH = @TranH_qf_intr;
    end
end

function output = STV(x, a, eta)
    output.TV = a * eta.TV;
end

function output = TVaddTV(x, eta1, eta2)
    output.TV = eta1.TV + eta2.TV;
end

function output = Hv(x, H, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(H * v.TV, n, p);
end

function output = Hv_intr(x, H, v)
    output.TV = H * v.TV;
end

function output = phi(x, y, s, B, inp_s_y, sBs, fns, v, times)
    output = 1;
end

function output = SRTV_embedded(x, m, seed)
    rand('state', seed);
    basis_x = build_bases_embedded(x);
    [nn, d] = size(basis_x);
    [n, p] = size(x.U);
    u = rand(d, m) - 0.5;
    TV = basis_x * u;
    for i = 1 : m
        output{i}.TV = reshape(TV(:, i), n, p);
    end
end

function output = SRTV_embedded_intr(x, m, seed)
    rand('state', seed);
    [n, p] = size(x.U);
    d = 0.5 * n * (n - 1);
    u = rand(d, m) - 0.5;
    for i = 1 : m
        output{i}.TV = u(:, i);
    end
end

function output = proj(x, eta)
    temp = x.U' * eta.TV;
    output.TV = eta.TV - x.U * ((temp + temp') / 2);
end

function output = proj_intr(x, eta)
    output = eta;
end

function output = rank1operator_embedded(x, v1, v2)
    [n, p] = size(x.U);
    vv1 = reshape(v1.TV, n * p, 1);
    vv2 = reshape(v2.TV, n * p, 1);
    output = vv1 * vv2';
end

function output = rank1operator_embedded_intr(x, v1, v2)
    output = v1.TV * v2.TV';
end

function output = operadd(x, H1, H2)
    output = H1 + H2;
end

function output = inpro_embedded(x, v1, v2)
    output = sum(sum(v1.TV .* v2.TV));
end

function output = inpro_embedded_intr(x, v1, v2)
    output = v1.TV' * v2.TV;
end

%% exponential mapping when it is considered as an embedded manifold
function output = R_exp_embedded(x, eta)
    output.U = x.U * expm(x.U' * eta.TV);
end

function output = beta_exp(x1, eta, x2)
    output = 1;
end

function output = R_exp_embedded_intr(x, eta)
    eta = embedded_full_c(x, eta);
    output.U = x.U * expm(x.U' * eta.TV);
    output.Up = null(output.U');
end

function output = beta_exp_intr(x1, eta, x2)
    output = 1;
end

%% qf
function output = R_qf_embedded(x, eta)
    output.U = qf(x.U + eta.TV);
end

function output = beta_qf_embedded(x1, eta, x2)
    output = eta.beta;
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro_embedded(x1, eta, eta) / inpro_embedded(x2, Tee, Tee));
end

function output = R_qf_embedded_intr(x, eta)
    eta = embedded_full_c(x, eta);
    output.U = qf(x.U + eta.TV);
    output.Up = null(output.U');
end

function output = beta_qf_embedded_intr(x1, eta, x2)
    if(isfield(eta, 'beta'))
        output = eta.beta;
    else
        output = 1;
    end
%     eta = embedded_full_c(x1, eta);
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro_embedded(x1, eta, eta) / inpro_embedded(x2, Tee, Tee));
end

function output = Tranv_R_qf(x, d, y, v)
    temp = inv(y.U' * (x.U + d.TV));
    temp2 = v.TV * temp;
    temp3 = y.U' * temp2;
    output.TV = y.U * (rho_skew(temp3) - temp3) + temp2;
%     output = y * rho_skew(temp3) + temp2 - y * temp3;%%---------------
end

function output = rho_skew(A)
    output = tril(A, -1) - tril(A, -1)';
end

%% polar
function output = R_polar(x, eta)
    [n, p] = size(x.U);
    output.U = (x.U + eta.TV) * (eye(p) + eta.TV' * eta.TV)^(-0.5);
end

%% Emb&Quo nonIso: vector transport by projection
function [output, d, x, y] = Tranv_proj(x, d, y, v)
    temp = y.U' * v.TV;
    output.TV = v.TV - y.U * ((temp + temp') / 2);
%     output = v - y * (y' * v + v' * y) / 2;%%-------
%     [n, p] = size(x);
%     perp_basis_x = build_perp_bases_embedded(x);
%     perp_basis_y = build_perp_bases_embedded(y);
%     output = reshape(v, n * p, 1) - perp_basis_x * (pinv(perp_basis_y' * perp_basis_x) * (perp_basis_y' * reshape(v, n * p, 1)));
%     output = reshape(output, n, p);
%     TranvParams{1} = perp_basis_y;
%     TranvParams{2} = perp_basis_x;
end

function [output, d, x, y] = invTranv_proj(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases_embedded(x);
    end
    perp_basis_x = x.perp_basis;
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases_embedded(y);
    end
    perp_basis_y = y.perp_basis;
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(perp_basis_x' * perp_basis_y));
    end
    
    output.TV = reshape(v.TV, n * p, 1) - perp_basis_y * (d.pinvPxPy * (perp_basis_x' * reshape(v.TV, n * p, 1)));
    output.TV = reshape(output.TV, n, p);
%     TranvParams{1} = perp_basis_y;
%     TranvParams{2} = perp_basis_x;
%     output = v - x * (x' * v + v' * x) / 2;
end

function output = TranH_proj(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases_embedded(x);
    end
    perp_basis_x = x.perp_basis;
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases_embedded(y);
    end
    perp_basis_y = y.perp_basis;
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(perp_basis_x' * perp_basis_y));
    end
    output = H - H * perp_basis_y * d.pinvPxPy * perp_basis_x';
    output = output - perp_basis_y * (perp_basis_y' * output);
%     output = H - H * perp_basis_x * perp_basis_x';
%     output = output - perp_basis_x * pinv(perp_basis_y' * perp_basis_x) * (perp_basis_y' * output);
end

%% parallel vector transport when it is considered as an embedded manifold
function [output, d, x, y] = Tranv_parallel_embedded(x, d, y, v)
    B = expm(x.U' * d.TV / 2);
    output.TV = x.U * B * (x.U' * v.TV) * B;
end

function [output, d, x, y] = invTranv_parallel_embedded(x, d, y, v, TranvParams)
    B = expm(logm(y.U' * x.U) / 2);
    output.TV = y.U * B * (y.U' * v.TV) * B;
end

function output = TranH_parallel_embedded(x, d, y, H)
    [n, p] = size(x.U);
    dim = n * p - p * (p + 1) / 2;
    bases = build_bases_embedded(x);
    P_bases = zeros(n * p, dim);
    P_Hb = zeros(n * p, dim);
    B = expm(x.U' * d.TV / 2);
    C = x.U * B * x.U';
    for i = 1 : dim
        P_bases(:, i) = reshape(C * reshape(bases(:, i), n, p) * B, n * p, 1);
    end
    Hb = H * bases;
    for i = 1 : dim
        P_Hb(:, i) = reshape(C * reshape(Hb(:, i), n, p) * B, n * p, 1);
    end
    output = P_Hb * P_bases';
end

%% Emb Iso: direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (d.U * (d.TV' * (xbases' * v.TV))), n, p);
end

function [output, d, x, y] = invTranv_cano(x, d, y, v, TranvParams)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (d.TV * (d.U' * (ybases' * v.TV))), n, p);
end

function output = TranH_cano(x, d, y, H)
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    temp = d.U * d.TV';
    output = ybases * (temp * (xbases' * (H * xbases)) * temp') * ybases';
end

%% Emb Iso: direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    output.TV = reshape((v.TV + (d.tilde_Q * (d.U * d.TV') - d.Q) * (d.Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_cano_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    output.TV = reshape((v.TV + (d.Q * (d.TV * d.U') - d.tilde_Q) * (d.tilde_Q' * v.TV)), n, p);
end

function output = TranH_cano_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(tilde_Q' * Q);
    end
    output = H + H * (d.Q * (d.TV * d.U') - d.tilde_Q) * d.tilde_Q';
    output = output + (d.tilde_Q * (d.U * d.TV') - d.Q) * (d.Q' * output);
end

%% Emb Iso: smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (xbases' * v.TV), n, p);
end

function [output, d, x, y] = invTranv_smooth(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (ybases' * v.TV), n, p);
end

function output = TranH_smooth(x, d, y, H)
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    output = ybases * (xbases' * (H * xbases)) * ybases';
end

%% Emb Iso: smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    output.TV = reshape((v.TV - (d.tilde_Q + d.Q) * (d.Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_smooth_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    output.TV = reshape((v.TV - (d.Q + d.tilde_Q) * (d.tilde_Q' * v.TV)), n, p);
end

function output = TranH_smooth_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    
    output = H - H * (d.Q + d.tilde_Q) * d.tilde_Q';
    output = output - (d.tilde_Q + d.Q) * (d.Q' * output);
end

%% Emb Iso: intrinsic method
function [output, d, x, y] = Tranv_smooth_intr(x, d, y, v)
    output = v;
end

function [output, d, x, y] = invTranv_smooth_intr(x, d, y, v)
    output = v;
end

function output = TranH_smooth_intr(x, d, y, H)
    output = H;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_qf_embedded(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_cano(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(x.bases * (d.TV * (d.U' * (y.bases' * output.TV))), size(x.U));
end

function output = TranH_cano_qf_embedded(x, d, y, H)
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    temp = U * V';
    temp = (temp * (xbases' * (H * xbases)) * temp') * ybases';
    temp = temp - 2 / inpro_embedded(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro_embedded(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybases - nu1.TV * (nu1.TV' * ybases * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn_qf_embedded(x, d, y, v)
    [TIv, d, x. y] = Tranv_cano_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_NonIn_qf_embedded(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV + (d.Q * (d.TV * d.U') - d.tilde_Q) * (d.tilde_Q' * output.TV)), size(x.U));
end

function output = TranH_cano_NonIn_qf_embedded(x, d, y, H)
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    output = H + H * (Q * (V * U') - tilde_Q) * tilde_Q';
    output = output + (tilde_Q * (U * V') - Q) * (Q' * output);
    output = output - 2 / inpro_embedded(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro_embedded(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_qf_embedded(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    xbases = x.bases;
    ybases = y.bases;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(xbases * (ybases' * output.TV), size(x.U));
end

function output = TranH_smooth_qf_embedded(x, d, y, H)
    xbases = x.bases;
    ybases = y.bases;
    nu1 = d.nu1;
    nu2 = d.nu2;
    temp = xbases' * (H * xbases) * ybases';
    temp = temp - 2 / inpro_embedded(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro_embedded(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybases - nu1.TV * (nu1.TV' * ybases * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn_qf_embedded(x, d, y, v)
    [TIv, d, x. y] = Tranv_smooth_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_NonIn_qf_embedded(x, d, y, v)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Q') || ~isfield(d, 'tilde_Q'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV - (d.Q + d.tilde_Q) * (d.tilde_Q' * output.TV)), size(x.U));
end

function output = TranH_smooth_NonIn_qf_embedded(x, d, y, H)
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    nu1 = d.nu1;
    nu2 = d.nu2;

    output = H - H * (Q + tilde_Q) * tilde_Q';
    output = output - (tilde_Q + Q) * (Q' * output);
    output = output - 2 / inpro_embedded(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro_embedded(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): construct vector transport
function [output, d, x, y] = Tranv_qf_embedded(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    tempv = bases_x' * reshape(v.TV, n * p, 1);
    tempv = tempv - 2 * vv1.TV * (vv1.TV' * tempv) / (vv1.TV' * vv1.TV);
    tempv = tempv - 2 * vv2.TV * (vv2.TV' * tempv) / (vv2.TV' * vv2.TV);
    output.TV = bases_y * tempv;
    output.TV = reshape(output.TV, n, p);
end

function [output, d, x, y] = invTranv_qf_embedded(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    tempv = bases_y' * reshape(v.TV, n * p, 1);
    tempv = tempv - 2 * vv2.TV * (vv2.TV' * tempv) / (vv2.TV' * vv2.TV);
    tempv = tempv - 2 * vv1.TV * (vv1.TV' * tempv) / (vv1.TV' * vv1.TV);
    output.TV = bases_x * tempv;
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_qf_embedded(x, d, y, H)
    bases_x = x.bases;
    bases_y = y.bases;
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output = bases_x' * (H * bases_x);
    output = output - 2 * (output * vv1.TV) * vv1.TV' / (vv1.TV' * vv1.TV);
    output = output - 2 * (output * vv2.TV) * vv2.TV' / (vv2.TV' * vv2.TV);
    output = output - 2 * vv1.TV * (vv1.TV' * output) / (vv1.TV' * vv1.TV);
    output = output - 2 * vv2.TV * (vv2.TV' * output) / (vv2.TV' * vv2.TV);
    output = output * bases_y';
    output = bases_y * output;
end

%% Emb Locking-Iso(qf): intrinsic approach
function [output, d, x, y] = Tranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = embedded_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_embedded(x, full_d, full_d) / inpro_embedded(y, Tee, Tee));
        Tee_intr = embedded_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * vv1.TV * (vv1.TV' * v.TV) / (vv1.TV' * vv1.TV);
    output.TV = output.TV - 2 * vv2.TV * (vv2.TV' * output.TV) / (vv2.TV' * vv2.TV);
end

function [output, d, x, y] = invTranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = embedded_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_embedded(x, full_d, full_d) / inpro_embedded(y, Tee, Tee));
        Tee_intr = embedded_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output.TV = v.TV - 2 * vv2.TV * (vv2.TV' * v.TV) / (vv2.TV' * vv2.TV);
    output.TV = output.TV - 2 * vv1.TV * (vv1.TV' * output.TV) / (vv1.TV' * vv1.TV);
end

function output = TranH_qf_intr(x, d, y, H)
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output = H - 2 * (H * vv1.TV) * (vv1.TV' / (vv1.TV' * vv1.TV));
    output = output - 2 * (output * vv2.TV) * (vv2.TV' / (vv2.TV' * vv2.TV));
    output = output - 2 * vv1.TV * ((vv1.TV' * output) / (vv1.TV' * vv1.TV));
    output = output - 2 * vv2.TV * ((vv2.TV' * output) / (vv2.TV' * vv2.TV));
end

%%assistant function------------------------------------------------------
function [output, R] = qf(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    diagR(diagR == 0) = 1;
    output = output * spdiags(sign(diagR), 0, k, k);
end

function [output, R] = qf_thin(B)
%     output = orth(B);
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    output = output(:, abs(diagR) > sqrt(eps));
    output = full(output * spdiags(sign(diagR), 0, size(output, 2), size(output, 2)));
end

function output = build_perp_bases_embedded(x)
    r2 = sqrt(2);
    [n, p] = size(x.U);
    output = sparse(n * p, p * (p + 1) / 2);
%     output = zeros(n * p, p * (p + 1) / 2);
    times = 1;
    for j = 1 : p
        st = (j - 1) * n + 1;
        en = j * n;
        output(st : en, times) = x.U(:, j);
        times = times + 1;
    end

    for i = 2 : p
        for j = i : p
            st = (j - i) * n + 1;
            en = (j - i + 1) * n;
            output(st : en, times) = x.U(:, j) / r2;
            st = (j - 1) * n + 1;
            en = j * n;
            output(st : en, times) = x.U(:, j - i + 1) / r2;
            times = times + 1;
        end
    end
end

function output = build_bases_embedded(x)
    [n, p] = size(x.U);
    x.U = x.U / sqrt(2);
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
%     output = zeros(n * p, d);
    col = 1;
    for i = 2 : p
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x.U(:, j);
            output((j - 1) * n + 1 : j * n, col) = x.U(:, i - 1);
            col = col + 1;
        end
    end
end

function output = embedded_intr_c(x, eta)
    r2 = sqrt(2);
    [n, p] = size(x.U);
    omega = x.U' * eta.TV; % omega
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
end

function output = embedded_full_c(x, v)
    r2 = sqrt(2);
    [n, p] = size(x.U);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega' - omega;
    output.TV = x.U * omega;
end
