function fs = sphere(retraction, vector_transport)
% This function return some necessary function handles of unit sphere
%
% INPUT:
% sphere : S^{n - 1}
%     params.retraction : 
%         '1', exponential mapping
%         '2', R_x(eta) = (x + eta) / norm(x + eta)
%     params.vector_transport : the isometric vector transport
%         '1', parallel transport
%         '2', isometric vector transport (canonical angles)
%         '3', differentiated the '2' retraction
%         '4', vector transport by projection
% 
% OUTPUT:
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.phi(x, y, s, B, inp_s_y, sBs, fns, v, times) : return the coefficient of RBroyden Family update formula.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.SRTV(x, m, seed) : return m random tangent vectors in T_x M, seed is the random seed.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
%     fns.Tranv_Params(x1, d, x2, v, TranvParams) : return the tangent vector which is the same as fns.Tranv(x1, d, x2, v), 
%                                                   TranvParams is some parameters such that the cost becomes smaller and 
%                                                   are given by fns.Tranv or fns.invTranv.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
%
% By Wen Huang

    fprintf('unit sphere\n');
    fs.Hv = @Hv;
    fs.STV = @STV;
    fs.TVaddTV = @TVaddTV;
    fs.rank1operator = @rank1operator;
    fs.operadd = @operadd;
    fs.inpro = @inpro;
    fs.phi = @phi;
    fs.proj = @proj;
    fs.SRTV = @SRTV;
    if(retraction == 1)
        fprintf('exponential mapping\n')
        fs.R = @R_exp;
        fs.beta = @beta_exp;
    elseif(retraction == 2)
        fprintf('retraction: (x + eta) / norm(x + eta)\n')
        fs.R = @R_normalized;
        fs.beta = @beta_normalized;
    end
    if(vector_transport == 1)
        fprintf('parallel vector transport\n')
        fs.Tranv = @Tranv_parallel;
%         fs.Tranv_Params = @Tranv_parallel;
        fs.invTranv = @invTranv_parallel;
        fs.TranH = @TranH_parallel;
    elseif(vector_transport == 2)
        fprintf('vector transport : reflection based on L space\n')
        fs.Tranv = @Tranv_iso;
        fs.invTranv = @invTranv_iso;
%         fs.Tranv_Params = @Tranv_iso;
        fs.TranH = @TranH_iso;
    elseif(vector_transport == 3)
        fprintf('vector transport : differentiated retraction\n')
        fs.Tranv = @Tranv_diff;
        fs.invTranv = @invTranv_diff;
%         fs.Tranv_Params = @Tranv_diff;
        fs.TranH = @TranH_diff;
    elseif(vector_transport == 4)
        fprintf('vector transport : projection\n')
        fs.Tranv = @Tranv_proj;
        fs.invTranv = @invTranv_proj;
%         fs.Tranv_Params = @Tranv_proj;
        fs.TranH = @TranH_proj;
    end
end

function output = STV(x, s, eta)
    output.TV = s * eta.TV;
end

function output = TVaddTV(x, v1, v2)
    output.TV = v1.TV + v2.TV;
end

function output = Hv(x, H, v)
    output.TV = H * v.TV;
end

function output = phi(x, y, s, B, inp_s_y, sBs, fns, v, times)
    output = 1;
end

function output = proj(x, eta)
    output.TV = eta.TV - x.U * (x.U' * eta.TV);
end

function output = SRTV(x, m, seed)
    rand('state', seed);
    n = length(x.U);
    u = rand(n - 1, m) - 0.5;
    basis_x = null(x.U');
    TV = basis_x * u;
    for i = 1 : m
        output{i}.TV = TV(:, i);
    end
end

function output = rank1operator(x, v1, v2)
    output = v1.TV * v2.TV';
end

function output = operadd(x, H1, H2)
    output = H1 + H2;
end

function output = inpro(x, v1, v2)
    output = v1.TV' * v2.TV;
end

%% exponential mapping
function output = R_exp(x, eta) % retraction
    neta = norm(eta.TV);
    output.U = x.U * cos(neta) + (sin(neta) / neta) * eta.TV;
end

function output = beta_exp(x1, eta, x2)
    output = 1;
end

%% (x + eta) / norm(x + eta)
function output = R_normalized(x, eta) % retraction
    output.U = (x.U + eta.TV) / norm(x.U + eta.TV);
end

function output = beta_normalized(x1, eta, x2)
    if(~isfield(eta, 'beta'))
        if(norm(eta.TV) < eps)
            output = 1;
            return;
        end
        nn = norm(x1.U + eta.TV);
        vv.TV = (eta.TV - x2.U * (x2.U' * eta.TV)) / nn;
        eta.beta = norm(eta.TV) / norm(vv.TV);
    end
    output = eta.beta;    
end

%% parallel vector transport
function [output, d, x, x2] = Tranv_parallel(x, d, x2, v)
    nn = norm(x.U + x2.U);
    output.TV = v.TV - 2 * v.TV' * x2.U / nn / nn * (x.U + x2.U);
end

function [output, d, x, x2] = invTranv_parallel(x, d, x2, v)
    nn = norm(x.U + x2.U);
    output.TV = v.TV - 2 * v.TV' * x.U / nn / nn * (x.U + x2.U);
end

function output = TranH_parallel(x1, d, x2, H)
    n = size(x1.U, 1);
    x12.U = x1.U + x2.U;
    nn = norm(x12.U);
    temp1 = x2.U' * H;
    temp2 = H * x12.U;
    output = H - 2 / nn / nn * x12.U * (temp1) - 2 / nn / nn * (temp2 - 2 / nn / nn * (temp1 * x12.U) * x12.U) * x1.U';
end

%% isometric vector transport
function [output, d, x, x2] = Tranv_iso(x, d, x2, v)
    r.TV = x2.U - x.U * (x.U' * x2.U);
    q.TV = r.TV / norm(r.TV);
    tilde_r.TV = x.U - x2.U * (x2.U' * x.U);
    tilde_q.TV = tilde_r.TV / norm(tilde_r.TV);
    output.TV = (v.TV + (- tilde_q.TV - q.TV) * (q.TV' * v.TV)); %% the sign of tilde_q is negative, because sign(q' * tilde_q) is negative.
end

function [output, d, x, x2] = invTranv_iso(x, d, x2, v)
    r.TV = x.U - x2.U * (x2.U' * x.U);
    q.TV = r.TV / norm(r.TV);
    tilde_r.TV = x2.U - x.U * (x.U' * x2.U);
    tilde_q.TV = tilde_r.TV / norm(tilde_r.TV);
    output.TV = (v.TV + (- tilde_q.TV - q.TV) * (q.TV' * v.TV)); %% the sign of tilde_q is negative, because sign(q' * tilde_q) is negative.
end

function output = TranH_iso(x1, d, x2, H)
    r.TV = x2.U - x1.U * (x1.U' * x2.U);
    q.TV = r.TV / norm(r.TV);
    tilde_r.TV = x1.U - x2.U * (x2.U' * x1.U);
    tilde_q.TV = tilde_r.TV / norm(tilde_r.TV);
    output = H + (H * (- q.TV - tilde_q.TV)) * tilde_q.TV' - (tilde_q.TV + q.TV) * (q.TV' * H + (q.TV' * (H * (- q.TV - tilde_q.TV))) * tilde_q.TV');
end

%% differentiated retraction
function [output, d, x, x2] = Tranv_diff(x, d, x2, v)
    nxd = norm(x.U + d.TV);
    output.TV = (v.TV - (x.U + d.TV) * ((x.U + d.TV)' * v.TV) / (nxd * nxd)) / nxd;
end

function [output, d, x, x2] = invTranv_diff(x, d, x2, v)
    nxd = norm(x.U + d.TV);
    output.TV = nxd * (v.TV - (x.U + d.TV) * ((x.U' * v.TV) / (x.U' * (x.U + d.TV))));
end

function output = TranH_diff(x1, d, x2, H)
    nxd = norm(x1.U + d.TV);
    TempH = (H - (H * (x1.U + d.TV)) * (x1.U' / (x1.U' * (x1.U + d.TV))));
    output = (TempH - (x1.U + d.TV) * (((x1.U + d.TV)' * TempH) / (nxd * nxd)));
end

%% vector transport by projection
function [output, d, x, x2] = Tranv_proj(x, d, x2, v)
    nxd = norm(x.U + d.TV);
    output.TV = (v.TV - (x.U + d.TV) * ((x.U + d.TV)' * v.TV) / (nxd * nxd));
end

function [output, d, x, x2] = invTranv_proj(x, d, x2, v)
    nxd = norm(x.U + d.TV);
    output.TV = (v.TV - (x.U + d.TV) * ((x.U' * v.TV) / (x.U' * (x.U + d.TV))));
end

function output = TranH_proj(x1, d, x2, H)
    nxd = norm(x1.U + d.TV);
    TempH = (H - (H * (x1.U + d.TV)) * (x1.U' / (x1.U' * (x1.U + d.TV))));
    output = (TempH - (x1.U + d.TV) * (((x1.U + d.TV)' * TempH) / (nxd * nxd)));
end
