function fs = stiefel(retraction, vector_transport)
% This function return some necessary function handles of the Stiefel manifold
% 
% INPUT:
% Stiefel manifold : St(p, n)
%     params.retraction : 
%         '1', exponential mapping when it is considered as Quotient space
%              beta is 1
%         '2', exponential mapping when it is considered as Embedding space
%              beta is 1
%         '3', qf factorization : qf(X + eta)
%              beta is computed by using quotient metric
%         '4', qf factorization : qf(X + eta)
%              beta is computed by using embedded metric
%         '5', construct retraction for intrinsic method of Quo Iso vector transport 
%              beta is 1
%         '6', construct retraction for intrinsic method of Emb Iso vector transport 
%              beta is 1
%     params.vector_transport : the isometric vector transport
%         '1', noniso: vector transport by projection
%         '2', Emb Iso: parallel transport
%         '3', Emb Iso: direct-rotation with basis on tangent space
%         '4', Emb Iso: direct-rotation with basis on normal space
%         '5', Emb Iso: smooth basis on tangent space
%         '6', Emb Iso: smooth basis on normal space
%         '7', Emb Iso: intrinsic method
%         '8', Quo Iso: parallel transport
%         '9', Quo Iso: direct-rotation with basis on horizontal space
%         '10', Quo Iso: direct-rotation with basis on vertical space
%         '11',Quo Iso: smooth basis on horizontal space
%         '12',Quo Iso: smooth basis on vertical space
%         '13',Quo Iso: intrinsic method
%         '14',Emb Locking-Iso(qf): direct-rotation with basis on tangent space
%         '15',Emb Locking-Iso(qf): direct-rotation with basis on normal space
%         '16',Emb Locking-Iso(qf): smooth basis on tangent space
%         '17',Emb Locking-Iso(qf): smooth basis on normal space
%         '18',Emb Locking-Iso(qf): construct vector transport
%         '19',Emb Locking-Iso(qf): intrinsic method
%         '20',Quo Locking-Iso(qf): direct-rotation with basis on horizontal space
%         '21',Quo Locking-Iso(qf): direct-rotation with basis on vertical space
%         '22',Quo Locking-Iso(qf): smooth basis on horizontal space
%         '23',Quo Locking-Iso(qf): smooth basis on vertical space
%         '24',Quo Locking-Iso(qf): construct vector transport
%         '25',Quo Locking-Iso(qf): intrinsic method
%         '26',Emb Locking-nonIso(qf): vector transport by rojection
% 
% OUTPUT:
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.Hv(x, H, v) : return tangent vector which is given by that operator H affect on tangent vector v on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.phi(x, y, s, B, inp_s_y, sBs, fns, v, times) : return the coefficient of RBroyden Family update formula.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.SRTV(x, m, seed) : return m random tangent vectors in T_x M, seed is the random seed.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.beta(x, eta) : return \|eta\| / \|TR_eta (eta)\|, where TR is the differentiated retraction.
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
%     fns.Tranv_Params(x1, d, x2, v, TranvParams) : return the tangent vector which is the same as fns.Tranv(x1, d, x2, v), 
%                                                   TranvParams is some parameters such that the cost becomes smaller and 
%                                                   are given by fns.Tranv or fns.invTranv.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
%     fns.TranH(x1, d, x2, H1) : return a operator, H2, that affects on T_x2 M. H2 satisfies that for any v \in T_x2(M), H2(v) = Tran (H1 (Tran^{-1}(v)))
%                             where Tran is the vector transport fns.Tranv(x1, d, x2, v).
%     fns.rank1operator(x, v1, v2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = g_x(v2, v) * v1.
%     fns.operadd(x, H1, H2) : return a operator, H, on T_x M that satisfies that for any v \in T_x M, H(v) = H1(v) + H2(v).
%     fns.coTVtimesTV(x, coeta, eta) : return a number which is the result of the cotangent vector, coeta, acts on the tangent vector, eta.
%     fns.coTV(x1, d, x2, v) : return the desired cotangent vector required by Ring and Wirth's RBFGS
% 
% By Wen Huang

    fprintf('Stiefel manifold\n');
    fs.Hv = @Hv;
    fs.phi = @phi;
    fs.operadd = @operadd;
    fs.proj = @proj;
    fs.STV = @STV;
    fs.TVaddTV = @TVaddTV;
    
    % using intrinsic approach
    if(vector_transport == 7 || vector_transport == 13 || vector_transport == 19 || vector_transport == 25)
        fprintf('intrinsic dimension approach\n')
        fs.Hv = @Hv_intr;
        fs.invBv = @invBv_intr;
        fs.proj = @proj_intr;
        fs.SRTV = @SRTV_intr;
        fs.rank1operator = @rank1operator_intr;
        fs.inpro = @inpro_intr;
        fs.coTVtimesTV = @inpro_intr;
        if(retraction == 1)
            fprintf('exponential mapping: stiefel manifold is considered as quotient space\n')
            fs.R = @R_exp_quotient_intr;
            fs.beta = @beta_exp_intr;
        elseif(retraction == 2)
            fprintf('exponential mapping: stiefel manifold is considered as embedded space\n')
            fs.R = @R_exp_embedded_intr;
            fs.beta = @beta_exp_intr;
        elseif(retraction == 3)
            fprintf('retraction: qf for stiefel manifold is considered as quotient space\n')
            fs.R = @R_qf_quotient_intr;
            fs.beta = @beta_qf_quotient_intr;
            fs.coTV = @coTV_quotient_intr;
        elseif(retraction == 4)
            fprintf('retraction: qf for stiefel manifold is considered as embedded space\n')
            fs.R = @R_qf_embedded_intr;
            fs.beta = @beta_qf_embedded_intr;
            fs.coTV = @coTV_embedded_intr;
        elseif(retraction == 5)
            fprintf('retraction: construct retraction for intrinsic method of vector transport as quotient space\n')
            fs.R = @R_construct_quotient_intr;
            fs.beta = @beta_construct_quotient_intr;
            fs.coTV = @coTV_construct_quotient_intr;
        elseif(retraction == 6)
            fprintf('retraction: construct retraction for intrinsic method of vector transport as embedded space\n')
            fs.R = @R_construct_embedded_intr;
            fs.beta = @beta_construct_embedded_intr;
            fs.coTV = @coTV_construct_embedded_intr;
        end
    else
        if(retraction == 1 || retraction == 3)
            fs.SRTV = @SRTV_quotient;
            fprintf('rank 1 operator: stiefel manifold is considered as quotient space\n')
            fs.rank1operator = @rank1operator_quotient;
            fprintf('inner product: stiefel manifold is considered as quotient space\n')
            fs.inpro = @inpro_quotient;
            fs.coTV = @coTV_quotient;
            fs.coTVtimesTV = @inpro_quotient;
        elseif(retraction == 2 || retraction == 4)
            fs.SRTV = @SRTV_embedded;
            fprintf('rank 1 operator: stiefel manifold is considered as embedded space\n')
            fs.rank1operator = @rank1operator_embedded;
            fprintf('inner product: stiefel manifold is considered as embedded space\n')
            fs.inpro = @inpro_embedded;
            fs.coTV = @coTV_embedded;
            fs.coTVtimesTV = @inpro_embedded;
        end

        if(retraction == 1)
            fprintf('exponential mapping: stiefel manifold is considered as quotient space\n')
            fs.R = @R_exp_quotient;
            fs.beta = @beta_exp;
        elseif(retraction == 2)
            fprintf('exponential mapping: stiefel manifold is considered as embedded space\n')
            fs.R = @R_exp_embedded;
            fs.beta = @beta_exp;
        elseif(retraction == 3)
            fprintf('retraction: qf for stiefel manifold is considered as quotient space\n')
            fs.R = @R_qf_quotient;
            fs.beta = @beta_qf_quotient;
        elseif(retraction == 4)
            fprintf('retraction: qf for stiefel manifold is considered as embedded space\n')
            fs.R = @R_qf_embedded;
            fs.beta = @beta_qf_embedded;
        end
    end
    
    if(vector_transport == 1)
        fprintf('Emb&Quo nonIso: vector transport by projection \n')
        fs.Tranv = @Tranv_proj;
        fs.Tranv_Params = @Tranv_proj;
        fs.invTranv = @invTranv_proj;
        fs.TranH = @TranH_proj;
    elseif(vector_transport == 2)
        fprintf('Emb Iso: parallel transport \n')
        fs.Tranv = @Tranv_parallel_embedded;
        fs.invTranv = @invTranv_parallel_embedded;
        fs.TranH = @TranH_parallel_embedded;
    elseif(vector_transport == 3)
        fprintf('Emb Iso: direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano;
%         fs.Tranv_Params = @Tranv_cano_Params;
        fs.invTranv = @invTranv_cano;
        fs.TranH = @TranH_cano;
    elseif(vector_transport == 4)
        fprintf('Emb Iso: direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn;
%         fs.Tranv_Params = @Tranv_cano_NonIn_Params;
        fs.invTranv = @invTranv_cano_NonIn;
        fs.TranH = @TranH_cano_NonIn;
    elseif(vector_transport == 5)
        fprintf('Emb Iso: smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth;
%         fs.Tranv_Params = @Tranv_smooth_Params;
        fs.invTranv = @invTranv_smooth;
        fs.TranH = @TranH_smooth;
    elseif(vector_transport == 6)
        fprintf('Emb Iso: smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn;
%         fs.Tranv_Params = @Tranv_smooth_NonIn_Params;
        fs.invTranv = @invTranv_smooth_NonIn;
        fs.TranH = @TranH_smooth_NonIn;
    elseif(vector_transport == 7)
        fprintf('Emb Iso: intrinsic method \n')
        fs.Tranv = @Tranv_smooth_intr;
%         fs.Tranv_Params = @Tranv_smooth_intr_Params;
        fs.invTranv = @invTranv_smooth_intr;
        fs.TranH = @TranH_smooth_intr;
    elseif(vector_transport == 8)
        fprintf('Quo Iso: parallel transport \n')
        fs.Tranv = @Tranv_parallel_quotient;
        fs.invTranv = @invTranv_parallel_quotient;
        fs.TranH = @TranH_parallel_quotient;
    elseif(vector_transport == 9)
        fprintf('Quo Iso: direct-rotation with basis on horizontal space \n')
        fs.Tranv = @Tranv_cano_Quo;
%         fs.Tranv_Params = @Tranv_cano_Quo_Params;
        fs.invTranv = @invTranv_cano_Quo;
        fs.TranH = @TranH_cano_Quo;
    elseif(vector_transport == 10)
        fprintf('Quo Iso: direct-rotation with basis on vertical space \n')
        fs.Tranv = @Tranv_cano_Quo_NonIn;
%         fs.Tranv_Params = @Tranv_cano_Quo_NonIn_Params;
        fs.invTranv = @invTranv_cano_Quo_NonIn;
        fs.TranH = @TranH_cano_Quo_NonIn;
    elseif(vector_transport == 11)
        fprintf('Quo Iso: smooth basis on horizontal space \n')
        fs.Tranv = @Tranv_smooth_Quo;
%         fs.Tranv_Params = @Tranv_smooth_Quo_Params;
        fs.invTranv = @invTranv_smooth_Quo;
        fs.TranH = @TranH_smooth_Quo;
    elseif(vector_transport == 12)
        fprintf('Quo Iso: smooth basis on vertical space \n')
        fs.Tranv = @Tranv_smooth_Quo_NonIn;
%         fs.Tranv_Params = @Tranv_smooth_Quo_NonIn_Params;
        fs.invTranv = @invTranv_smooth_Quo_NonIn;
        fs.TranH = @TranH_smooth_Quo_NonIn;
    elseif(vector_transport == 13)
        fprintf('Quo Iso: intrinsic method \n')
        fs.Tranv = @Tranv_smooth_Quo_intr;
%         fs.Tranv_Params = @Tranv_smooth_Quo_intr_Params;
        fs.invTranv = @invTranv_smooth_Quo_intr;
        fs.TranH = @TranH_smooth_Quo_intr;
    elseif(vector_transport == 14)
        fprintf('Emb Locking-Iso(qf): direct-rotation with basis on tangent space \n')
        fs.Tranv = @Tranv_cano_qf_embedded;
%         fs.Tranv_Params = @Tranv_cano_qf_embedded_Params;
        fs.invTranv = @invTranv_cano_qf_embedded;
        fs.TranH = @TranH_cano_qf_embedded;
    elseif(vector_transport == 15)
        fprintf('Emb Locking-Iso(qf): direct-rotation with basis on normal space \n')
        fs.Tranv = @Tranv_cano_NonIn_qf_embedded;
%         fs.Tranv_Params = @Tranv_cano_NonIn_qf_embedded_Params;
        fs.invTranv = @invTranv_cano_NonIn_qf_embedded;
        fs.TranH = @TranH_cano_NonIn_qf_embedded;
    elseif(vector_transport == 16)
        fprintf('Emb Locking-Iso(qf): smooth basis on tangent space \n')
        fs.Tranv = @Tranv_smooth_qf_embedded;
%         fs.Tranv_Params = @Tranv_smooth_qf_embedded_Params;
        fs.invTranv = @invTranv_smooth_qf_embedded;
        fs.TranH = @TranH_smooth_qf_embedded;
    elseif(vector_transport == 17)
        fprintf('Emb Locking-Iso(qf): smooth basis on normal space \n')
        fs.Tranv = @Tranv_smooth_NonIn_qf_embedded;
%         fs.Tranv_Params = @Tranv_smooth_NonIn_qf_embedded_Params;
        fs.invTranv = @invTranv_smooth_NonIn_qf_embedded;
        fs.TranH = @TranH_smooth_NonIn_qf_embedded;
    elseif(vector_transport == 18)
        fprintf('Emb Locking-Iso(qf): construct vector transport \n')
        fs.Tranv = @Tranv_qf_embedded;
%         fs.Tranv_Params = @Tranv_qf_embedded_Params;
        fs.invTranv = @invTranv_qf_embedded;
        fs.TranH = @TranH_qf_embedded;
    elseif(vector_transport == 19)
        fprintf('Emb Locking-Iso(qf): intrinsic method \n')
        fs.Tranv = @Tranv_qf_intr;
        fs.invTranv = @invTranv_qf_intr;
        fs.TranH = @TranH_qf_intr;
    elseif(vector_transport == 20)
        fprintf('Quo Locking-Iso(qf): direct-rotation with basis on horizontal space \n')
        fs.Tranv = @Tranv_cano_qf_quotient;
%         fs.Tranv_Params = @Tranv_cano_qf_quotient_Params;
        fs.invTranv = @invTranv_cano_qf_quotient;
        fs.TranH = @TranH_cano_qf_quotient;
    elseif(vector_transport == 21)
        fprintf('Quo Locking-Iso(qf): direct-rotation with basis on vertical space \n')
        fs.Tranv = @Tranv_cano_NonIn_qf_quotient;
%         fs.Tranv_Params = @Tranv_cano_NonIn_qf_quotient_Params;
        fs.invTranv = @invTranv_cano_NonIn_qf_quotient;
        fs.TranH = @TranH_cano_NonIn_qf_quotient;
    elseif(vector_transport == 22)
        fprintf('Quo Locking-Iso(qf): smooth basis on horizontal space \n')
        fs.Tranv = @Tranv_smooth_qf_quotient;
%         fs.Tranv_Params = @Tranv_smooth_qf_quotient_Params;
        fs.invTranv = @invTranv_smooth_qf_quotient;
        fs.TranH = @TranH_smooth_qf_quotient;
    elseif(vector_transport == 23)
        fprintf('Quo Locking-Iso(qf): smooth basis on vertical space \n')
        fs.Tranv = @Tranv_smooth_NonIn_qf_quotient;
%         fs.Tranv_Params = @Tranv_smooth_NonIn_qf_quotient_Params;
        fs.invTranv = @invTranv_smooth_NonIn_qf_quotient;
        fs.TranH = @TranH_smooth_NonIn_qf_quotient;
    elseif(vector_transport == 24)
        fprintf('Quo Locking-Iso(qf): construct vector transport \n')
        fs.Tranv = @Tranv_qf_quotient;
%         fs.Tranv_Params = @Tranv_qf_quotient_Params;
        fs.invTranv = @invTranv_qf_quotient;
        fs.TranH = @TranH_qf_quotient;
    elseif(vector_transport == 25)
        fprintf('Quo Locking-Iso(qf): intrinsic method \n')
        fs.Tranv = @Tranv_qf_quotient_intr;
%         fs.Tranv_Params = @Tranv_qf_quotient_intr_Params;
        fs.invTranv = @invTranv_qf_quotient_intr;
        fs.TranH = @TranH_qf_quotient_intr;
    elseif(vector_transport == 26)
        fprintf('Emb Locking-nonIso(qf): vector transport by rojection \n')
        fs.Tranv = @Tranv_qf_proj;
%         fs.Tranv_Params = @Tranv_qf_proj;
        fs.invTranv = @invTranv_qf_proj;
        fs.TranH = @TranH_qf_proj;
        fs.beta = @beta_qf_proj;
    end
end

function output = STV(x, a, eta)
    output.TV = a * eta.TV;
end

function output = TVaddTV(x, eta1, eta2)
    output.TV = eta1.TV + eta2.TV;
end

function output = Hv(x, H, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(H * v.TV, n, p);
end

function output = Hv_intr(x, H, v)
    output.TV = H * v.TV;
end

function output = phi(x, y, s, B, inp_s_y, sBs, fns, v, times)
%     invB = inv(B);
% %     vGv = fns.inpro(x, v, v);%% invB * v);
%     u = fns.inpro(x, y, invB * y) * sBs / inp_s_y / inp_s_y;
%     output = 0.1 / (1 - u);
%     if(output > 0)
%         output = 0;
%     end
% %     if(output < 0.95 / (1 - u))
% %         output = 0.95 / (1 - u);
% %     end

    output = 1; % RBFGS
end

function output = SRTV_quotient(x, m, seed)
    rand('state', seed);
    basis_x = build_bases_quotient(x);
    [nn, d] = size(basis_x);
    [n, p] = size(x.U);
    u = rand(d, m) - 0.5;
    TV = basis_x * u;
    for i = 1 : m
        output{i}.TV = reshape(TV(:, i), n, p);
    end
end

function output = SRTV_embedded(x, m, seed)
    rand('state', seed);
    basis_x = build_bases_embedded(x);
    [nn, d] = size(basis_x);
    [n, p] = size(x.U);
    u = rand(d, m) - 0.5;
    TV = basis_x * u;
    for i = 1 : m
        output{i}.TV = reshape(TV(:, i), n, p);
    end
end

function output = SRTV_intr(x, m, seed)
    rand('state', seed);
    [n, p] = size(x.U);
    d = 0.5 * p * (p - 1) + (n - p) * p;
    u = rand(d, m) - 0.5;
    for i = 1 : m
        output{i}.TV = u(:, i);
    end
end

function output = proj(x, eta)
    temp = x.U' * eta.TV;
    output.TV = eta.TV - x.U * ((temp + temp') / 2);
end

function output = proj_intr(x, eta)
    output = eta;
end

function output = rank1operator_embedded(x, v1, v2)
    [n, p] = size(x.U);
    vv1 = reshape(v1.TV, n * p, 1);
    vv2 = reshape(v2.TV, n * p, 1);
    output = vv1 * vv2';
end

function output = rank1operator_quotient(x, v1, v2)
    [n, p] = size(x.U);
    vv1 = reshape(v1.TV, n * p, 1);
    vv2 = reshape(v2.TV - 0.5 * x.U * (x.U' * v2.TV), n * p, 1);
    output = vv1 * vv2';
end

function output = rank1operator_intr(x, v1, v2)
    output = v1.TV * v2.TV';
end

function output = operadd(x, H1, H2)
    output = H1 + H2;
end

function output = inpro_embedded(x, v1, v2)
    output = sum(sum(v1.TV .* v2.TV));
end

function output = inpro_quotient(x, v1, v2)
    vv1 = reshape(v1.TV, size(x.U));
    vv2 = reshape(v2.TV, size(x.U));
    output = sum(sum(vv1 .* (vv2 - 0.5 * x.U * (x.U' * vv2))));
end

function output = inpro_intr(x, v1, v2)
    output = v1.TV' * v2.TV;
end

function output = embedded_co_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = x1' * eta; % omega
    omega = omega' - omega;
    K = x2' * eta; % K
    omega = omega / sqrt(2);
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = quotient_co_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
%     r2 = sqrt(2);
    [n, p] = size(x1);
    omega = x1' * eta; % omega
    omega = omega' - omega;
    K = x2' * eta; % K
    omega = omega / 2;%/ sqrt(2);
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = coTV_embedded_intr(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    y1 = y.U;
    y2 = y.Up;
    x1 = x.U;
    x2 = x.Up;
    d = embedded_full_c(x, d);
    v = embedded_full_c(y, v);
    [n, p] = size(x1);
    temp1 = inv(y1' * (x1 + d.TV));
    temp2 = y1' * v.TV;
    temp3 =  (v.TV - y1 * temp2);
    temp5 = 2 * y1 * tril(temp2, -1);
    
    temp7 = (temp3 + temp5) * temp1';
    output = embedded_co_intr_c(x, temp7);
end

function output = coTV_quotient_intr(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    y1 = y.U;
    y2 = y.Up;
    x1 = x.U;
    x2 = x.Up;
% dd = d;%%
    d = quotient_full_c(x, d);
    v = quotient_full_c(y, v);
    [n, p] = size(x1);
    temp1 = inv(y1' * (x1 + d.TV));
    temp2 = y1' * v.TV;
    temp3 =  (v.TV - y1 * temp2);
    temp5 = y1 * tril(temp2, -1);
    
    temp7 = (temp3 + temp5) * temp1';
    temp8 = temp7 + x1 * (x1' * temp7);
    output = quotient_co_intr_c(x, temp8);
end

function output = coTV_embedded(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    [n, p] = size(x.U);
    temp1 = inv(y.U' * (x.U + d.TV));
    temp2 = y.U' * v.TV;
    temp3 =  (v.TV - y.U * temp2);
    temp5 = 2 * y.U * tril(temp2, -1);
    temp6 = (temp3 + temp5) * temp1';
    temp7 = x.U' * temp6;
    output.TV = temp6 - 0.5 * x.U * temp7 - 0.5 * x.U * temp7';
end

function output = coTV_quotient(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    [n, p] = size(x.U);
    temp1 = inv(y.U' * (x.U + d.TV));
    temp2 = y.U' * v.TV;
    temp3 =  (v.TV - y.U * temp2);
    temp5 = y.U * tril(temp2, -1);
    
    temp7 = (temp3 + temp5) * temp1';
    temp8 = temp7 + x.U * (x.U' * temp7); %% this is coTV * G^{-1}. G^{-1} = I + x * x'.
    temp9 = x.U' * temp8;
    output.TV = temp8 - 0.5 * x.U * temp9 - 0.5 * x.U * temp9';
end

function output = coTV_construct_embedded_intr(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    y1 = y.U;
    y2 = y.Up;
    x1 = x.U;
    x2 = x.Up;
% dd = d;%%
    d = embedded_full_c(x, d);
    v = embedded_full_c(y, v);
    [n, p] = size(x1);
    
    omega = x1' * d.TV;
    K = x2' * d.TV;
    A = [omega, -K'; K, zeros(n - p, n - p)];
    [VV, DD] = eig(A);
    P = zeros(n, n);
    for i = 1 : n
        for j = i : n
            if(DD(i, i) == DD(j, j))
                P(i, j) = exp(DD(i, i));
                P(j, i) = P(i, j);
            else
                P(i, j) = (exp(DD(i, i)) - exp(DD(j, j))) / (DD(i, i) - DD(j, j));
                P(j, i) = P(i, j);
            end
        end
    end
    temp1 = P .* (VV' * speye(n, p) * (v.TV' * [x1, x2] * VV));
    temp2 = VV * temp1 * VV';
    temp3 = [temp2(1:p, 1:p), temp2(1:p, p+1 : end) - temp2(p+1:end, 1:p)'];
    temp4 = real([x1, x2] * temp3');
    output = embedded_co_intr_c(x, temp4);
end

function output = coTV_construct_quotient_intr(x, d, y, v)
    if(~isfield(d, 'v'))
        output = v;
        return;
    end
    y1 = y.U;
    y2 = y.Up;
    x1 = x.U;
    x2 = x.Up;
    d = quotient_full_c(x, d);
    v = quotient_full_c(y, v);
    [n, p] = size(x1);
    
    omega = x1' * d.TV;
    K = x2' * d.TV;
    A = [omega, -K'; K, zeros(n - p, n - p)];
    [VV, DD] = eig(A);
    P = zeros(n, n);
    for i = 1 : n
        for j = i : n
            if(DD(i, i) == DD(j, j))
                P(i, j) = exp(DD(i, i));
                P(j, i) = P(i, j);
            else
                P(i, j) = (exp(DD(i, i)) - exp(DD(j, j))) / (DD(i, i) - DD(j, j));
                P(j, i) = P(i, j);
            end
        end
    end
    temp1 = P .* (VV' * speye(n, p) * ((v.TV' - 0.5 * (v.TV' * y1) * y1') * [x1, x2] * VV));
    temp2 = VV * temp1 * VV';
    temp3 = [temp2(1:p, 1:p), temp2(1:p, p+1 : end) - temp2(p+1:end, 1:p)'];
    temp4 = real([x1, x2] * temp3');
    temp5 = temp4 + x1 * (x1' * temp4);
    output = quotient_co_intr_c(x, temp5);
end

function output = Tranv_R_construct(x, d, y, v)
    x1 = x.U;
    x2 = x.Up;
    y1 = y.U;
    y2 = y.Up;
    [n, p] = size(x1);
    omega1 = x1' * d.TV;
    K1 = x2' * d.TV;
    omega2 = x1' * v.TV;
    K2 = x2' * v.TV;
    A = [omega1, -K1'; K1, zeros(n - p, n - p)];
    VV = [omega2, -K2'; K2, zeros(n - p, n - p)];
    temp = [x1, x2] * deri_expm(A, VV);
    output.TV = real(temp(:, 1 : p));
end

function output = deri_expm(A, V)
    [n, n] = size(A);
    [U, D] =eig(A);
    P = zeros(n, n);
    for i = 1 : n
        for j = i : n
            if(D(i, i) == D(j, j))
                P(i, j) = exp(D(i, i));
                P(j, i) = P(i, j);
            else
                P(i, j) = (exp(D(i, i)) - exp(D(j, j))) / (D(i, i) - D(j, j));
                P(j, i) = P(i, j);
            end
        end
    end
    bV = U' * V * U;
    output = U * (bV .* P) * U';
end

function output = krontril(n)
    nn = n * n;
    output = spdiags(reshape(tril(ones(n, n), -1), [], 1), 0, nn, nn);
end

function output = krontranspose(n)
    nn = n * n;
    output = spdiags(nn, nn);
    for i = 1 : n
        for j = 1 : n
            output((i - 1) * n + j, i + (j - 1) * n) = 1;
        end
    end
end

%% exponential mapping when it is considered as an embedded manifold
function output = R_exp_embedded(x, eta)
    [n, p] = size(x.U);
    A = x.U' * eta.TV;
    S = eta.TV' * eta.TV;
    temp = expm([A -S; eye(p) A]) * eye(2 * p, p);
    output = [x.U eta.TV] * temp * expm(- A);
    %The following step is not needed analytically. However, it is needed
    %numerically.
    output.U = qf(output);
end

function output = R_exp_embedded_intr(x, eta)
    [n, p] = size(x.U);
    basisx = build_bases_embedded(x);
    eta.TV = basisx * eta.TV;
    eta.TV = reshape(eta.TV, n, p);
    A = x.U' * eta.TV;
    S = eta.TV' * eta.TV;
    temp = expm([A -S; eye(p) A]) * eye(2 * p, p);
    temp2 = [x.U eta.TV] * temp * expm(- A);
    output.U = basisx' * reshape(temp2, [], 1);
    output.Up = null(output.U');
end

%% exponential mapping when it is considered as an quotient manifold
function output = R_exp_quotient(x, eta)
    [n, p] = size(x.U);
    A = x.U' * eta.TV;
    [Q, B] = qr(eta.TV - x.U * (x.U' * eta.TV), 0);
    Temp = expm([A -B'; B zeros(p, p)]);
    output = [x.U Q] * Temp(:, 1 : p);
    %The following step is not needed analytically. However, it is needed
    %numerically.
    output.U = qf(output);
end

function output = R_exp_quotient_intr(x, eta)
    [n, p] = size(x.U);
    basisx = build_bases_quotient(x);
    basisx_flat = GB_quotient(x, basisx)';
    eta.TV = basisx * eta.TV;
    eta.TV = reshape(eta.TV, n, p);
    A = x.U' * eta.TV;
    [Q, B] = qr(eta.TV - x.U * (x.U' * eta.TV), 0);
    Temp = expm([A -B'; B zeros(p, p)]);
    temp2 = [x.U Q] * Temp(:, 1 : p);
    output = basisx_flat * reshape(temp2, [], 1);
    output.Up = null(output.U');
end

function output = beta_exp(x1, eta, x2)
    output = 1;
end

function output = beta_exp_intr(x1, eta, x2)
    output = 1;
end

%% qf
function output = R_qf_embedded(x, eta)
    output.U = qf(x.U + eta.TV);
end

function output = beta_qf_embedded(x1, eta, x2)
    output = eta.beta;
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro_embedded(x1, eta, eta) / inpro_embedded(x2, Tee, Tee));
end

function output = R_qf_embedded_intr(x, eta)
    fulleta = embedded_full_c(x, eta);
    output.U = qf(x.U + fulleta.TV);
    output.Up = null(output.U');
end

function output = beta_qf_embedded_intr(x1, eta, x2)
    output = eta.beta;
%     x11 = x1.U;
%     x21 = x2.U;
%     fulleta = embedded_full_c(x1, eta.TV);
% 
%     x12 = x1.Up;
%     x22 = x2.Up;
%     temp1 = fulleta * inv(x21' * (x11 + fulleta));
%     temp2 = x21' * temp1;
%     omegal = tril(temp2, -1);
%     K = x22' * temp1;
%     output = sqrt(inpro_embedded(x11, fulleta, fulleta) / (sum(sum(omegal .* omegal)) * 2 + sum(sum(K .* K))));
end

function output = R_qf_quotient(x, eta)
    output.U = qf(x.U + eta.TV);
end

function output = beta_qf_quotient(x1, eta, x2)
    output = eta.beta;
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro_quotient(x1, eta, eta) / inpro_quotient(x2, Tee, Tee));
end

function output = R_qf_quotient_intr(x, eta)
    eta = quotient_full_c(x, eta);
    output.U = qf(x.U + eta.TV);
    output.Up = null(output.U');
end

function output = beta_qf_quotient_intr(x1, eta, x2)
%     if(~isfield(eta, 'beta'))
%         x11 = x1.U;
%         x21 = x2.U;
%         eta = quotient_full_c(x1, eta);
%         x12 = x1.Up;
%         x22 = x2.Up;
%         temp1 = eta.TV * inv(x21' * (x11 + eta.TV));
%         temp2 = x21' * temp1;
%         omegal = tril(temp2, -1);
%         K = x22' * temp1;
%         eta.beta = sqrt(inpro_quotient(x1, eta, eta) / (sum(sum(omegal .* omegal)) + sum(sum(K .* K))));
%     end
    output = eta.beta;
end

function output = Tranv_R_qf(x, d, y, v)
    temp = inv(y.U' * (x.U + d.TV));
    temp2 = v.TV * temp;
    temp3 = y.U' * temp2;
    output.TV = y.U * (rho_skew(temp3) - temp3) + temp2;
end

function output = rho_skew(A)
    output = tril(A, -1) - tril(A, -1)';
end

%% construct retraction
function output = R_construct_embedded_intr(x, eta)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = eta.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega' - omega;
    K = reshape(eta.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    M = ([[omega;K], [-K'; zeros(n - p, n - p)]]);
    v = [x1, x2];
    
    v = v * expm(M);
    v1 = v(1 : n * p);
    v2 = v(n * p + 1 : end);
    output.U = reshape(v1, n, p);
    output.Up = reshape(v2, n, n - p);
end

function output = beta_construct_embedded_intr(x1, eta, x2)
    output = 1;
end

function output = R_construct_quotient_intr(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = eta.TV(1 : 0.5 * p * (p - 1));
    omega = omega' - omega;
    K = reshape(eta.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    M = ([[omega;K], [-K'; zeros(n - p, n - p)]]);
    v = [x1, x2];
    
    v = v * expm(M);
    v1 = v(1 : n * p);
    v2 = v(n * p + 1 : end);
    output.U = reshape(v1, n, p);
    output.Up = reshape(v2, n, n - p);
end

function output = beta_construct_quotient_intr(x1, eta, x2)
    output = 1;
end

%% polar
function output = R_polar(x, eta)
    [n, p] = size(x.U);
    output = (x.U + eta.TV) * (eye(p) + eta.TV' * eta.TV)^(-0.5);
end

%% Emb&Quo nonIso: vector transport by projection
function [output, d, x, y] = Tranv_proj(x, d, y, v)
    d.beta = 1;
    temp = y.U' * v.TV;
    output.TV = v.TV - y.U * ((temp + temp') / 2);
end

function [output, d, x, y] = invTranv_proj(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases_embedded(x);
    end
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases_embedded(y);
    end
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(x.perp_basis' * y.perp_basis));
    end
    output.TV = reshape(v.TV, n * p, 1) - y.perp_basis * (d.pinvPxPy * (x.perp_basis' * reshape(v.TV, n * p, 1)));
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_proj(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'perp_basis'))
        x.perp_basis = build_perp_bases_embedded(x);
    end
    if(~isfield(y, 'perp_basis'))
        y.perp_basis = build_perp_bases_embedded(y);
    end
    if(~isfield(d, 'pinvPxPy'))
        d.pinvPxPy = pinv(full(x.perp_basis' * y.perp_basis));
    end
    output = H - H * y.perp_basis * d.pinvPxPy * x.perp_basis';
    output = output - y.perp_basis * (y.perp_basis' * output);
end

%% Emb Iso: parallel transport
function [output, d, x, y] = Tranv_parallel_embedded(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    A = x.U' * d.TV;
    S = d.TV' * d.TV;
    M1 = [A -S; eye(p), A];
    M2 = [x.U d.TV];
    fhandle = @(t, w)ode_parallel_vt_embedded(t, w, M1, M2, A, n, p);
    options = odeset('RelTol',1e-5, 'AbsTol',1e-5);
    [T,V] = ode45(fhandle, [0 1], reshape(v.TV, n * p, 1), options);
    output.TV = reshape(V(end, :), n, p);
%     [T, V] = runge_kutta1(fhandle, reshape(v, n * p, 1), 0.1, 0, 1);
%     output = reshape(V(:, end), n, p);
    TranvParams = [];
end

function [output, d, x, y] = invTranv_parallel_embedded(x, d, y, v)
    invd = Tranv_parallel_embedded(x, d, y, d);
    invd.TV = -invd.TV;
    output = Tranv_parallel_embedded(y, invd, x, v);
end

function output = TranH_parallel_embedded(x, d, y, H)
    [n, p] = size(x.U);
    dim = n * p - p * (p + 1) / 2;
    bases = full(build_bases_embedded(x));
    P_bases = zeros(n * p, dim);
    P_Hb = zeros(n * p, dim);
    for i = 1 : dim
        Bi.TV = reshape(bases(:, i), n, p);
        Pi = Tranv_parallel_embedded(x, d, y, Bi);
        P_bases(:, i) = reshape(Pi.TV, n * p, 1);
    end
    Hb = H * bases;
    for i = 1 : dim
        Bi.TV = reshape(Hb(:, i), n, p);
        Pi = Tranv_parallel_embedded(x, d, y, Bi);
        P_Hb(:, i) = reshape(Pi.TV, n * p, 1);
    end
    output = P_Hb * P_bases';
end

%% Emb Iso: direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano(x, d, y, v)
    d.beta = 1;
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (U * (V' * (xbases' * v.TV))), n, p);
end

function [output, d, x, y] = invTranv_cano(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (V * (U' * (ybases' * v.TV))), n, p);
end

function output = TranH_cano(x, d, y, H)
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases' * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    temp = U * V';
    output = ybases * (temp * (xbases' * (H * xbases)) * temp') * ybases';
end

%% Emb Iso: direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    output.TV = reshape((v.TV + (tilde_Q * (U * V') - Q) * (Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_cano_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    output.TV = reshape((v.TV + (Q * (V * U') - tilde_Q) * (tilde_Q' * v.TV)), n, p);
end

function output = TranH_cano_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
        [d.U, S, d.TV] = svd(d.tilde_Q' * d.Q);
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    output = H + H * (Q * (V * U') - tilde_Q) * tilde_Q';
    output = output + (tilde_Q * (U * V') - Q) * (Q' * output);
end

%% Emb Iso: smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth(x, d, y, v)
    d.beta = 1;
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (xbases' * v.TV), n, p);
end

function [output, d, x, y] = invTranv_smooth(x, d, y, v)
    [n, p] = size(v.TV);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (ybases' * v.TV), n, p);
end

function output = TranH_smooth(x, d, y, H)
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    ybases = y.bases;
    output = ybases * (xbases' * (H * xbases)) * ybases';
end

%% Emb Iso: smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    output.TV = reshape((v.TV - (tilde_Q + Q) * (Q' * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_smooth_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    v.TV = reshape(v.TV, n * p, 1);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    output.TV = reshape((v.TV - (Q + tilde_Q) * (tilde_Q' * v.TV)), n, p);
end

function output = TranH_smooth_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perp'))
        x.B_perp = build_perp_bases_embedded(x);
    end
    Bx_perp = x.B_perp;
    if(~isfield(y, 'B_perp'))
        y.B_perp = build_perp_bases_embedded(y);
    end
    By_perp = y.B_perp;
    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        d.Q = qf_thin(full(By_perp - Bx_perp * (Bx_perp' * By_perp)));
        d.tilde_Q = qf_thin(full(Bx_perp - By_perp * (By_perp' * Bx_perp)));
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    
    output = H - H * (Q + tilde_Q) * tilde_Q';
    output = output - (tilde_Q + Q) * (Q' * output);
end

%% Emb Iso: intrinsic method
function [output, d, x, y] = Tranv_smooth_intr(x, d, y, v)
    d.beta = 1;
    output.TV = v.TV;
end

function [output, d, x, y] = invTranv_smooth_intr(x, d, y, v)
    output.TV = v.TV;
end

function output = TranH_smooth_intr(x, d, y, H)
    output = H;
end

%% Quo Iso: parallel transport
function [output, d, x, y] = Tranv_parallel_quotient(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    A = x.U' * d.TV;
    [Q, B] = qr(d.TV - x.U * (x.U' * d.TV), 0);
    M1 = [A -B'; B zeros(p, p)];
    M2 = [x.U Q];
    fhandle = @(t, w)ode_parallel_vt_quotient(t, w, M1, M2, A, n, p);
    options = odeset('RelTol',1e-5, 'AbsTol',1e-5);
    [T,V] = ode45(fhandle, [0 1], reshape(v.TV, n * p, 1), options);
    output.TV = reshape(V(end, :), n, p);
%     [T, V] = runge_kutta1(fhandle, reshape(v, n * p, 1), 0.1, 0, 1);
%     output = reshape(V(:, end), n, p);
end

function [output, d, x, y] = invTranv_parallel_quotient(x, d, y, v)
    invd = Tranv_parallel_quotient(x, d, y, d);
    invd.TV = -invd.TV;
    output = Tranv_parallel_quotient(y, invd, x, v);
end

function output = TranH_parallel_quotient(x, d, y, H)
    [n, p] = size(x.U);
    dim = n * p - p * (p + 1) / 2;
    bases = full(build_bases_quotient(x));
    P_bases = zeros(n * p, dim);
    P_Hb = zeros(n * p, dim);
    for i = 1 : dim
        Bi.TV = reshape(bases(:, i), n, p);
        [Pi, d] = Tranv_parallel_quotient(x, d, y, Bi);
        P_bases(:, i) = reshape(Pi.TV, n * p, 1);
    end
    Hb = H * bases;
    for i = 1 : dim
        Bi.TV = reshape(Hb(:, i), n, p);
        [Pi, d] = Tranv_parallel_quotient(x, d, y, Bi);
        P_Hb(:, i) = reshape(Pi.TV, n * p, 1);
    end
    for i = 1 : p
        P_bases((i - 1) * n + 1 : i * n, :) = P_bases((i - 1) * n + 1 : i * n, :) - 0.5 * y.U * y.U' * P_bases((i - 1) * n + 1 : i * n, :);
    end
    output = P_Hb * P_bases';
end

%% Quo Iso: direct-rotation with basis on horizontal space
function [output, d, x, y] = Tranv_cano_Quo(x, d, y, v)
    d.beta = 1;
    [n, p] = size(v.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(x, 'bases_flat'))
        x.bases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    xbases_flat = x.bases_flat;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases_flat * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(ybases * (U * (V' * (xbases_flat * v.TV))), n, p);
end

function [output, d, x, y] = invTranv_cano_Quo(x, d, y, v)
    [n, p] = size(v.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases_flat * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(xbases * (V * (U' * (ybases_flat * v.TV))), n, p);
end

function output = TranH_cano_Quo(x, d, y, H)
    [n, p] = size(d.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(x, 'bases_flat'))
        x.bases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    xbases_flat = x.bases_flat;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    if(~isfield(d, 'U') || ~isfield(d, 'V'))
        [U, S, V] = svd(full(ybases_flat * xbases));
        d.U = U;
        d.TV = V;
    end
    U = d.U;
    V = d.TV;
    temp = U * V';
    output = ybases * (temp * (xbases_flat * (H * xbases)) * temp') * ybases_flat;
end

%% Quo Iso: direct-rotation with basis on vertical space
function [output, d, x, y] = Tranv_cano_Quo_NonIn(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
        [U, S, V] = svd(d.Qy' * d.Qx);
        d.U = U;
        d.TV = V;
    end
    Qx = d.Qx;
    Qy = d.Qy;
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV - (1 - 1 / sqrt(2)) * x.U * (x.U' * v.TV), n * p, 1);
    output = reshape((v.TV + (Qy * (U * V') - Qx) * (Qx' * v.TV)), n, p);
    output.TV = output + (sqrt(2) - 1) * y.U * (y.U' * output);
end

function [output, d, x, y] = invTranv_cano_Quo_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
        [U, S, V] = svd(d.Qy' * d.Qx);
        d.U = U;
        d.TV = V;
    end
    Qx = d.Qx;
    Qy = d.Qy;
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV - (1 - 1 / sqrt(2)) * y.U * (y.U' * v.TV), n * p, 1);
    output = (v.TV + (Qx * (V * U') - Qy) * (Qy' * v.TV));
    output = reshape(output, n, p);
    output.TV = output + (sqrt(2) - 1) * x.U * (x.U' * output);
end

function output = TranH_cano_Quo_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
        [U, S, V] = svd(d.Qy' * d.Qx);
        d.U = U;
        d.TV = V;
    end
    Qx = d.Qx;
    Qy = d.Qy;
    U = d.U;
    V = d.TV;
    output = BGn05_quotient(x, H);
    output = G05B_quotient(x, output);
    output = output + output * (Qx * (V * U') - Qy) * Qy';
    output = output + (Qy * (U * V') - Qx) * (Qx' * output);
    output = BG05_quotient(y, output);
    output = Gn05B_quotient(y, output);
end

%% Quo Iso: smooth basis on horizontal space
function [output, d, x, y] = Tranv_smooth_Quo(x, d, y, v)
    d.beta = 1;
    [n, p] = size(v.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(x, 'bases_flat'))
        x.bases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    xbases_flat = x.bases_flat;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(full(ybases * (xbases_flat * v.TV)), n, p);
end

function [output, d, x, y] = invTranv_smooth_Quo(x, d, y, v)
    [n, p] = size(v.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    v.TV = reshape(v.TV, n * p, 1);
    output.TV = reshape(full(xbases * (ybases_flat * v.TV)), n, p);
end

function output = TranH_smooth_Quo(x, d, y, H)
    [n, p] = size(d.TV);
    dim = p * (n - (p + 1) / 2);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(x, 'bases_flat'))
        x.bases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    xbases_flat = x.bases_flat;
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    output = ybases * (xbases_flat * (H * xbases)) * ybases_flat;
end

%% Quo Iso: smooth basis on vertical space
function [output, d, x, y] = Tranv_smooth_Quo_NonIn(x, d, y, v)
    d.beta = 1;
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
    end
    Qx = d.Qx;
    Qy = d.Qy;

    v.TV = reshape(v.TV - (1 - 1 / sqrt(2)) * x.U * (x.U' * v.TV), n * p, 1);
    output = reshape((v.TV - (Qy + Qx) * (Qx' * v.TV)), n, p);
    output.TV = output + (sqrt(2) - 1) * y.U * (y.U' * output);
end

function [output, d, x, y] = invTranv_smooth_Quo_NonIn(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
    end
    Qx = d.Qx;
    Qy = d.Qy;
    v.TV = reshape(v.TV - (1 - 1 / sqrt(2)) * y.U * (y.U' * v.TV), n * p, 1);
    output = (v.TV - (Qx + Qy) * (Qy' * v.TV));
    output = reshape(output, n, p);
    output.TV = output + (sqrt(2) - 1) * x.U * (x.U' * output);
end

function output = TranH_smooth_Quo_NonIn(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'B_perpG05'))
        x.B_perpG05 = build_perp_bases_embedded(x);
    end
    Bx_perpG05 = x.B_perpG05;
    if(~isfield(y, 'B_perpG05'))
        y.B_perpG05 = build_perp_bases_embedded(y);
    end
    By_perpG05 = y.B_perpG05;
    if(~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        d.Qx = qf_thin(full(By_perpG05 - Bx_perpG05 * (Bx_perpG05' * By_perpG05)));
        d.Qy = qf_thin(full(Bx_perpG05 - By_perpG05 * (By_perpG05' * Bx_perpG05)));
    end
    Qx = d.Qx;
    Qy = d.Qy;
    
    output = BGn05_quotient(x, H);
    output = G05B_quotient(x, output);
    output = output - output * (Qx + Qy) * Qy';
    output = output - (Qy + Qx) * (Qx' * output);
    output = BG05_quotient(y, output);
    output = Gn05B_quotient(y, output);
end

%% Quo Iso: intrinsic approach
function [output, d, x, y] = Tranv_smooth_Quo_intr(x, d, y, v)
    d.beta = 1;
    output.TV = v.TV;
end

function [output, d, x, y] = invTranv_smooth_Quo_intr(x, d, y, v)
    output.TV = v.TV;
end

function output = TranH_smooth_Quo_intr(x, d, y, H)
    output = H;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on tangent space
function [output, d, x, y] = Tranv_cano_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_qf_embedded(x, d, y, v)
    [eps1, d, x, y] = Tranv_cano(x, d, y, d);
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(xbases * (V * (U' * (ybases' * output.TV))), size(x.U));
end

function output = TranH_cano_qf_embedded(x, d, y, H)
    [eps1, d, x, y] = Tranv_cano_qf_embedded(x, d, y, d);
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    temp = U * V';
    temp = (temp * (xbases' * (H * xbases)) * temp') * ybases';
    temp = temp - 2 / inpro_embedded(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro_embedded(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybases - nu1.TV * (nu1.TV' * ybases * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): direct-rotation with basis on normal space
function [output, d, x, y] = Tranv_cano_NonIn_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_NonIn_qf_embedded(x, d, y, v)

    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~isfield(d, 'U') || ~isfield(d, 'V') || ~~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_cano_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV + (Q * (V * U') - tilde_Q) * (tilde_Q' * output.TV)), size(x.U));
end

function output = TranH_cano_NonIn_qf_embedded(x, d, y, H)
    [eps1, d, x, y] = Tranv_cano_NonIn_qf_embedded(x, d, y, d);
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    U = d.U;
    V = d.TV;
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    output = H + H * (Q * (V * U') - tilde_Q) * tilde_Q';
    output = output + (tilde_Q * (U * V') - Q) * (Q' * output);
    output = output - 2 / inpro_embedded(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro_embedded(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): smooth basis on tangent space
function [output, d, x, y] = Tranv_smooth_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_qf_embedded(x, d, y, v)
    [eps1, d, x, y] = Tranv_smooth(x, d, y, d);
    xbases = x.bases;
    ybases = y.bases;
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape(xbases * (ybases' * output.TV), size(x.U));
end

function output = TranH_smooth_qf_embedded(x, d, y, H)
    [eps1, d, x, y] = Tranv_smooth_qf_embedded(x, d, y, d);
    xbases = x.bases;
    ybases = y.bases;
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    temp = xbases' * (H * xbases) * ybases';
    temp = temp - 2 / inpro_embedded(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro_embedded(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybases - nu1.TV * (nu1.TV' * ybases * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
    output = output * temp;
end

%% Emb Locking-Iso(qf): smooth basis on normal space
function [output, d, x, y] = Tranv_smooth_NonIn_qf_embedded(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth_NonIn(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_NonIn_qf_embedded(x, d, y, v)

    if(~isfield(d, 'Q') || ~isfield(d, 'tilde_Q') || ~~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        [eps1, d, x, y] = Tranv_smooth_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_embedded(y, nu2, v) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu1, output) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = reshape((output.TV - (Q + tilde_Q) * (tilde_Q' * output.TV)), size(x.U));
end

function output = TranH_smooth_NonIn_qf_embedded(x, d, y, H)
    [eps1, TranvParams] = Tranv_smooth_NonIn_qf_embedded(x, d, y, d);
    Q = d.Q;
    tilde_Q = d.tilde_Q;
    nu1 = d.nu1;
    nu2 = d.nu2;
    output = H - H * (Q + tilde_Q) * tilde_Q';
    output = output - (tilde_Q + Q) * (Q' * output);
    output = output - 2 / inpro_embedded(y, nu1, nu1) * (output * nu1.TV) * nu1.TV';
    output = output - 2 / inpro_embedded(y, nu2, nu2) * (output * nu2.TV) * nu2.TV';
    output = output - nu1.TV * (nu1.TV' * output * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
end

%% Emb Locking-Iso(qf): construct vector transport
function [output, d, x, y] = Tranv_qf_embedded(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    tempv.TV = bases_x' * reshape(v.TV, n * p, 1);
    tempv.TV = tempv.TV - 2 * vv1.TV * (vv1.TV' * tempv.TV) / (vv1.TV' * vv1.TV);
    tempv.TV = tempv.TV - 2 * vv2.TV * (vv2.TV' * tempv.TV) / (vv2.TV' * vv2.TV);
    output.TV = bases_y * tempv.TV;
    output.TV = reshape(output.TV, n, p);
end

function [output, d, x, y] = invTranv_qf_embedded(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    tempv.TV = bases_y' * reshape(v.TV, n * p, 1);
    tempv.TV = tempv.TV - 2 * vv2.TV * (vv2.TV' * tempv.TV) / (vv2.TV' * vv2.TV);
    tempv.TV = tempv.TV - 2 * vv1.TV * (vv1.TV' * tempv.TV) / (vv1.TV' * vv1.TV);
    output.TV = bases_x * tempv.TV;
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_qf_embedded(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_embedded(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_embedded(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        c1.TV = bases_x' * reshape(d.TV, n * p, 1);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, d, d) / inpro_embedded(y, Tee, Tee));
        bc2.TV = d.beta * bases_y' * reshape(Tee.TV, n * p, 1);
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output = bases_x' * (H * bases_x);
    output = output - 2 * (output * vv1.TV) * vv1.TV' / (vv1.TV' * vv1.TV);
    output = output - 2 * (output * vv2.TV) * vv2.TV' / (vv2.TV' * vv2.TV);
    output = output - 2 * vv1.TV * (vv1.TV' * output) / (vv1.TV' * vv1.TV);
    output = output - 2 * vv2.TV * (vv2.TV' * output) / (vv2.TV' * vv2.TV);
    output = output * bases_y';
    output = bases_y * output;
end

%% Emb Locking-Iso(qf): intrinsic approach
function [output, d, x, y] = Tranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = embedded_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_embedded(x, full_d, full_d) / inpro_embedded(y, Tee, Tee));
        Tee_intr = embedded_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;

    output.TV = v.TV - 2 * vv1.TV * (vv1.TV' * v.TV) / (vv1.TV' * vv1.TV);
    output.TV = output.TV - 2 * vv2.TV * (vv2.TV' * output.TV) / (vv2.TV' * vv2.TV);
end

function [output, d, x, y] = invTranv_qf_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = embedded_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_embedded(x, full_d, full_d) / inpro_embedded(y, Tee, Tee));
        Tee_intr = embedded_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output.TV = v.TV - 2 * vv2.TV * (vv2.TV' * v.TV) / (vv2.TV' * vv2.TV);
    output.TV = output.TV - 2 * vv1.TV * (vv1.TV' * output.TV) / (vv1.TV' * vv1.TV);
end

function output = TranH_qf_intr(x, d, y, H)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = embedded_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        beta = sqrt(inpro_embedded(x, full_d, full_d) / inpro_embedded(y, Tee, Tee));
        Tee_intr = embedded_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = -d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output = H - 2 * (H * vv1.TV) * (vv1.TV' / (vv1.TV' * vv1.TV));
    output = output - 2 * (output * vv2.TV) * (vv2.TV' / (vv2.TV' * vv2.TV));
    output = output - 2 * vv1.TV * ((vv1.TV' * output) / (vv1.TV' * vv1.TV));
    output = output - 2 * vv2.TV * ((vv2.TV' * output) / (vv2.TV' * vv2.TV));
end

%% Quo Locking-Iso(qf): direct-rotation with basis on horizontal space
function [output, d, x, y] = Tranv_cano_qf_quotient(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano_Quo(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'beta'))
        [eps1, d, x, y] = Tranv_cano_Quo(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_quotient(y, nu1, TIv) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu2, output) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_cano_Quo(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    xbases = x.bases;
    ybases = y.bases;
    U = d.U;
    V = d.TV;
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_quotient(y, nu2, v) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu1, output) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    dim = p * (n - (p + 1) / 2);
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    output.TV = reshape(xbases * (V * (U' * (ybases_flat * output.TV))), n, p);
end

function output = TranH_cano_qf_quotient(x, d, y, H)
    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'U') || ~isfield(d, 'V') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_cano_qf_quotient(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    xbases = x.bases;
    ybases = y.bases;
    xbases_flat = x.bases_flat;
    ybases_flat = y.bases_flat;
    U = d.U;
    V = d.TV;
    
    dim = p * (n - (p + 1) / 2);
    nu1_flat.TV = reshape(nu1.TV, n, p);
    nu1_flat.TV = (reshape(nu1_flat.TV - 0.5 * y.U * (y.U' * nu1_flat.TV), n * p, 1))';
    nu2_flat.TV = reshape(nu2.TV, n, p);
    nu2_flat.TV = (reshape(nu2_flat.TV - 0.5 * y.U * (y.U' * nu2_flat.TV), n * p, 1))';
    temp = U * V';
    temp = (temp * (xbases_flat * (H * xbases)) * temp') * ybases_flat;
    temp = temp - 2 / inpro_quotient(y, nu1, nu1) * (temp * nu1.TV) * nu1_flat.TV;
    temp = temp - 2 / inpro_quotient(y, nu2, nu2) * (temp * nu2.TV) * nu2_flat.TV;
    output = ybases - nu1.TV * (nu1_flat.TV * ybases * (2 / inpro_quotient(y, nu1, nu1)));
    output = output - nu2.TV * (nu2_flat.TV * output * (2 / inpro_quotient(y, nu2, nu2)));
    output = output * temp;
end

%% Quo Locking-Iso(qf): direct-rotation with basis on vertical space
function [output, d, x, y] = Tranv_cano_NonIn_qf_quotient(x, d, y, v)
    [TIv, d, x, y] = Tranv_cano_Quo_NonIn(x, d, y, v);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'beta'))
        [eps1, d, x, y] = Tranv_cano_Quo_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_quotient(y, nu1, TIv) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu2, output) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_cano_NonIn_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        [eps1, d, x, y] = Tranv_cano_Quo_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    Qx = d.Qx;
    Qy = d.Qy;
    U = d.U;
    V = d.TV;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_quotient(y, nu2, v) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV - 2 * inpro_quotient(y, nu1, output) / inpro_quotient(y, nu1, nu1) * nu1.TV, n, p);
    output.TV = reshape(output.TV - (1 - 1 / sqrt(2)) * y.U * (y.U' * output.TV), n * p, 1);
    output.TV = (output.TV + (Qx * (V * U') - Qy) * (Qy' * output.TV));
    output.TV = reshape(output.TV, n, p);
    output.TV = output.TV + (sqrt(2) - 1) * x.U * (x.U' * output.TV);
end

function output = TranH_cano_NonIn_qf_quotient(x, d, y, H)
    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy') || ~isfield(d, 'U') || ~isfield(d, 'V'))
        [eps1, d, x, y] = Tranv_cano_NonIn_qf_quotient(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    U = d.U;
    V = d.TV;
    Qx = d.Qx;
    Qy = d.Qy;
    
    nu1_flat.TV = reshape(nu1.TV, n, p);
    nu1_flat.TV = (reshape(nu1_flat.TV - 0.5 * y.U * (y.U' * nu1_flat.TV), n * p, 1))';
    nu2_flat.TV = reshape(nu2.TV, n, p);
    nu2_flat.TV = (reshape(nu2_flat.TV - 0.5 * y.U * (y.U' * nu2_flat.TV), n * p, 1))';
    output = BGn05_quotient(x, H);
    output = G05B_quotient(x, output);
    output = output + output * (Qx * (V * U') - Qy) * Qy';
    output = output + (Qy * (U * V') - Qx) * (Qx' * output);
    output = BG05_quotient(y, output);
    output = Gn05B_quotient(y, output);
    output = output - 2 / inpro_quotient(y, nu1, nu1) * (output * nu1.TV) * nu1_flat.TV;
    output = output - 2 / inpro_quotient(y, nu2, nu2) * (output * nu2.TV) * nu2_flat.TV;
    output = output - nu1.TV * (nu1_flat.TV * output * (2 / inpro_quotient(y, nu1, nu1)));
    output = output - nu2.TV * (nu2_flat.TV * output * (2 / inpro_quotient(y, nu2, nu2)));
end

%% Quo Locking-Iso(qf): smooth basis on horizontal space
function [output, d, x, y] = Tranv_smooth_qf_quotient(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth_Quo(x, d, y, v);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'beta'))
        [eps1, d, x, y] = Tranv_smooth_Quo(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_quotient(y, nu1, TIv) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu2, output) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_smooth_Quo(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    xbases = x.bases;
    ybases = y.bases;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_quotient(y, nu2, v) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu1, output) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    dim = p * (n - (p + 1) / 2);
    if(~isfield(y, 'bases_flat'))
        y.bases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    end
    ybases_flat = y.bases_flat;
    output.TV = reshape(xbases * (((ybases_flat * output.TV))), n, p);
end

function output = TranH_smooth_qf_quotient(x, d, y, H)

    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(x, 'bases') || ~isfield(y, 'bases'))
        [eps1, d, x, y] = Tranv_smooth_qf_quotient(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    xbases = x.bases;
    ybases = y.bases;
    xbases_flat = x.bases_flat;
    ybases_flat = y.bases_flat;
    
    dim = p * (n - (p + 1) / 2);
    nu1_flat.TV = reshape(nu1.TV, n, p);
    nu1_flat.TV = (reshape(nu1_flat.TV - 0.5 * y.U * (y.U' * nu1_flat.TV), n * p, 1))';
    nu2_flat.TV = reshape(nu2.TV, n, p);
    nu2_flat.TV = (reshape(nu2_flat.TV - 0.5 * y.U * (y.U' * nu2_flat.TV), n * p, 1))';
    temp = (xbases_flat * (H * xbases)) * ybases_flat;
    temp = temp - 2 / inpro_quotient(y, nu1, nu1) * (temp * nu1.TV) * nu1_flat.TV;
    temp = temp - 2 / inpro_quotient(y, nu2, nu2) * (temp * nu2.TV) * nu2_flat.TV;
    output = ybases - nu1.TV * (nu1_flat.TV * ybases * (2 / inpro_quotient(y, nu1, nu1)));
    output = output - nu2.TV * (nu2_flat.TV * output * (2 / inpro_quotient(y, nu2, nu2)));
    output = output * temp;
end

%% Quo Locking-Iso(qf): smooth basis on vertical space
function [output, d, x, y] = Tranv_smooth_NonIn_qf_quotient(x, d, y, v)
    [TIv, d, x, y] = Tranv_smooth_Quo_NonIn(x, d, y, v);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'beta'))
        [eps1, d, x, y] = Tranv_smooth_Quo_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_quotient(y, nu1, TIv) / inpro_quotient(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_quotient(y, nu2, output) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function [output, d, x, y] = invTranv_smooth_NonIn_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        [eps1, d, x, y] = Tranv_smooth_Quo_NonIn(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    Qx = d.Qx;
    Qy = d.Qy;
    
    v.TV = reshape(v.TV, [], 1);
    output.TV = v.TV - 2 * inpro_quotient(y, nu2, v) / inpro_quotient(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV - 2 * inpro_quotient(y, nu1, output) / inpro_quotient(y, nu1, nu1) * nu1.TV, n, p);
    output.TV = reshape(output.TV - (1 - 1 / sqrt(2)) * y.U * (y.U' * output.TV), n * p, 1);
    output.TV = (output.TV - (Qx + Qy) * (Qy' * output.TV));
    output.TV = reshape(output.TV, n, p);
    output.TV = output.TV + (sqrt(2) - 1) * x.U * (x.U' * output.TV);
end

function output = TranH_smooth_NonIn_qf_quotient(x, d, y, H)

    [n, p] = size(x.U);
    
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2') || ~isfield(d, 'Qx') || ~isfield(d, 'Qy'))
        [eps1, d, x, y] = Tranv_smooth_NonIn_qf_quotient(x, d, y, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        eps2.TV = beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    Qx = d.Qx;
    Qy = d.Qy;
    
    nu1_flat.TV = reshape(nu1.TV, n, p);
    nu1_flat.TV = (reshape(nu1_flat.TV - 0.5 * y.U * (y.U' * nu1_flat.TV), n * p, 1))';
    nu2_flat.TV = reshape(nu2.TV, n, p);
    nu2_flat.TV = (reshape(nu2_flat.TV - 0.5 * y.U * (y.U' * nu2_flat.TV), n * p, 1))';
    output = BGn05_quotient(x, H);
    output = G05B_quotient(x, output);
    output = output - output * (Qx + Qy) * Qy';
    output = output - (Qy + Qx) * (Qx' * output);
    output = BG05_quotient(y, output);
    output = Gn05B_quotient(y, output);
    output = output - 2 / inpro_quotient(y, nu1, nu1) * (output * nu1.TV) * nu1_flat.TV;
    output = output - 2 / inpro_quotient(y, nu2, nu2) * (output * nu2.TV) * nu2_flat.TV;
    output = output - nu1.TV * (nu1_flat.TV * output * (2 / inpro_quotient(y, nu1, nu1)));
    output = output - nu2.TV * (nu2_flat.TV * output * (2 / inpro_quotient(y, nu2, nu2)));
end

%% Quo Locking-Iso(qf): construct vector transport
function [output, d, x, y] = Tranv_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        tempd.TV = reshape(d.TV - 0.5 * x.U * (x.U' * d.TV), n * p, 1);
        c1.TV = bases_x' * tempd.TV;
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        tempTee.TV = reshape(Tee.TV - 0.5 * y.U * (y.U' * Tee.TV), n * p, 1);
        bc2.TV = d.beta * bases_y' * tempTee.TV;
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    tempv.TV = reshape(v.TV - 0.5 * x.U * (x.U' * v.TV), n * p, 1);
    tempv.TV = bases_x' * tempv.TV;
    tempv.TV = tempv.TV - 2 * vv1.TV * (vv1.TV' * tempv.TV) / (vv1.TV' * vv1.TV);
    tempv.TV = tempv.TV - 2 * vv2.TV * (vv2.TV' * tempv.TV) / (vv2.TV' * vv2.TV);
    output.TV = bases_y * tempv.TV;
    output.TV = reshape(output.TV, n, p);
end

function [output, d, x, y] = invTranv_qf_quotient(x, d, y, v)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    bases_x = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    bases_y = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        tempd.TV = reshape(d.TV - 0.5 * x.U * (x.U' * d.TV), n * p, 1);
        c1.TV = bases_x' * tempd.TV;
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        tempTee.TV = reshape(Tee.TV - 0.5 * y.U * (y.U' * Tee.TV), n * p, 1);
        bc2.TV = d.beta * bases_y' * tempTee.TV;
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
%     tempv = reshape(v - 0.5 * x * (x' * v), n * p, 1);
    tempv.TV = reshape(v.TV - 0.5 * y.U * (y.U' * v.TV), n * p, 1);
    tempv.TV = bases_y' * tempv.TV;
    tempv.TV = tempv.TV - 2 * vv2.TV * (vv2.TV' * tempv.TV) / (vv2.TV' * vv2.TV);
    tempv.TV = tempv.TV - 2 * vv1.TV * (vv1.TV' * tempv.TV) / (vv1.TV' * vv1.TV);
    output.TV = bases_x * tempv.TV;
    output.TV = reshape(output.TV, n, p);
end

function output = TranH_qf_quotient(x, d, y, H)
    [n, p] = size(x.U);
    if(~isfield(x, 'bases'))
        x.bases = build_bases_quotient(x);
    end
    xbases = x.bases;
    if(~isfield(y, 'bases'))
        y.bases = build_bases_quotient(y);
    end
    ybases = y.bases;
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        tempd.TV = reshape(d.TV - 0.5 * x.U * (x.U' * d.TV), n * p, 1);
        c1.TV = bases_x' * tempd.TV;
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, d, d) / inpro_quotient(y, Tee, Tee));
        tempTee.TV = reshape(Tee.TV - 0.5 * y.U * (y.U' * Tee.TV), n * p, 1);
        bc2.TV = d.beta * bases_y' * tempTee.TV;
        w.TV = -c1.TV;
        d.TVv1.TV = c1.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    dim = p * (n - (p + 1) / 2);
    xbases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    ybases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    output = xbases_flat * (H * xbases);
    output = output - 2 * (output * vv1.TV) * vv1.TV' / (vv1.TV' * vv1.TV);
    output = output - 2 * (output * vv2.TV) * vv2.TV' / (vv2.TV' * vv2.TV);
    output = output - 2 * vv1.TV * (vv1.TV' * output) / (vv1.TV' * vv1.TV);
    output = output - 2 * vv2.TV * (vv2.TV' * output) / (vv2.TV' * vv2.TV);
    output = ybases * (output * ybases_flat);
end

%% Quo Locking-Iso(qf): intrinsic approach
function [output, d, x, y] = Tranv_qf_quotient_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = quotient_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_quotient(x, full_d, full_d) / inpro_quotient(y, Tee, Tee));
        Tee_intr = quotient_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = - d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    output.TV = v.TV - 2 * (vv1.TV' * v.TV) / (vv1.TV' * vv1.TV) * vv1.TV;
    output.TV = output.TV - 2 * (vv2.TV' * output.TV) / (vv2.TV' * vv2.TV) * vv2.TV;
end

function [output, d, x, y] = invTranv_qf_quotient_intr(x, d, y, v)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = quotient_full_c(x, d);
        Tee = Tranv_R_qf(x, full_d, y, full_d);
        d.beta = sqrt(inpro_quotient(x, full_d, full_d) / inpro_quotient(y, Tee, Tee));
        Tee_intr = quotient_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = - d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output.TV = v.TV - 2 * (vv2.TV' * v.TV) / (vv2.TV' * vv2.TV) * vv2.TV;
    output.TV = output.TV - 2 * (vv1.TV' * output.TV) / (vv1.TV' * vv1.TV) * vv1.TV;
end

function output = TranH_qf_quotient_intr(x, d, y, H)
    if(~isfield(d, 'vv1') || ~isfield(d, 'vv2'))
        full_d = quotient_full_c(x, d);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_quotient(x, full_d, full_d) / inpro_quotient(y, Tee, Tee));
        Tee_intr = quotient_intr_c(y, Tee);
        bc2.TV = d.beta * Tee_intr.TV;
        w.TV = - d.TV;
        d.TVv1.TV = d.TV - w.TV;
        d.TVv2.TV = w.TV - bc2.TV;
    end
    vv1 = d.TVv1;
    vv2 = d.TVv2;
    
    output = H - 2 * (H * vv1.TV) * (vv1.TV' / (vv1.TV' * vv1.TV));
    output = output - 2 * (output * vv2.TV) * (vv2.TV' / (vv2.TV' * vv2.TV));
    output = output - 2 * vv1.TV * ((vv1.TV' * output) / (vv1.TV' * vv1.TV));
    output = output - 2 * vv2.TV * ((vv2.TV' * output) / (vv2.TV' * vv2.TV));
end

%% Emb Locking-nonIso(qf): vector transport by rojection
function [output, d, x, y] = Tranv_qf_proj(x, d, y, v)
    temp = y.U' * v.TV;
    TIv.TV = v.TV - y.U * ((temp + temp') / 2);
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        temp = y.U' * d.TV;
        eps1.TV = d.TV - y.U * ((temp + temp') / 2);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, eps1, eps1) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;
    
    TIv.TV = reshape(TIv.TV, [], 1);
    output.TV = TIv.TV - 2 * inpro_embedded(y, nu1, TIv) / inpro_embedded(y, nu1, nu1) * nu1.TV;
    output.TV = output.TV - 2 * inpro_embedded(y, nu2, output) / inpro_embedded(y, nu2, nu2) * nu2.TV;
    output.TV = reshape(output.TV, size(d.TV));
end

function output = TranH_qf_proj(x, d, y, H)
    if(~isfield(d, 'nu1') || ~isfield(d, 'nu2'))
        temp = y.U' * d.TV;
        eps1.TV = d.TV - y.U * ((temp + temp') / 2);
        Tee = Tranv_R_qf(x, d, y, d);
        d.beta = sqrt(inpro_embedded(x, eps1, eps1) / inpro_embedded(y, Tee, Tee));
        eps2.TV = d.beta * Tee.TV;
        d.nu1.TV = reshape(2 * eps1.TV, [], 1);
        d.nu2.TV = reshape(- eps1.TV - eps2.TV, [], 1);
    end
    nu1 = d.nu1;
    nu2 = d.nu2;

    xbasis = build_bases_embedded(x);
    ybasis = build_bases_embedded(y);

    temp = ybasis' * (H * xbasis) * (pinv(full(ybasis' * xbasis))) * ybasis';
    temp = temp - 2 / inpro_embedded(y, nu1, nu1) * (temp * nu1.TV) * nu1.TV';
    temp = temp - 2 / inpro_embedded(y, nu2, nu2) * (temp * nu2.TV) * nu2.TV';
    output = ybasis - nu1.TV * (nu1.TV' * ybasis * (2 / inpro_embedded(y, nu1, nu1)));
    output = output - nu2.TV * (nu2.TV' * output * (2 / inpro_embedded(y, nu2, nu2)));
    output = output * temp;
end

function output = beta_qf_proj(x1, eta, x2)
    output = eta.beta;
%     temp = x2' * eta;
%     eps1 = eta - x2 * ((temp + temp') / 2);
%     Tee = Tranv_R_qf(x1, eta, x2, eta);
%     output = sqrt(inpro_embedded(x1, eps1, eps1) / inpro_embedded(x2, Tee, Tee));
end

%%assistant function------------------------------------------------------
function [output, R] = qf(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    diagR(diagR == 0) = 1;
    output = output * spdiags(sign(diagR), 0, k, k);
end

function [output, R] = qf_thin(B)
    k = size(B, 2);
    [output, R] = qr(B, 0);
    diagR = diag(R);
    output = output(:, abs(diagR) > 1e-10);
    output = output * spdiags(sign(diagR), 0, size(output, 2), size(output, 2));
end

function output = GB_quotient(x, B)
    [n, p] = size(x.U);
    output = zeros(size(B));
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, :) = B((i - 1) * n + 1 : i * n, :) - (0.5 * x.U) * x.U'* B((i - 1) * n + 1 : i * n, :);
    end
end

function output = G05B_quotient(x, B)
    a = 1 - 1 / sqrt(2);
    [n, p] = size(x.U);
    output = zeros(size(B));
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, :) = B((i - 1) * n + 1 : i * n, :) - (a * x.U) * x.U'* B((i - 1) * n + 1 : i * n, :);
    end
end

function output = Gn05B_quotient(x, B)
    a = sqrt(2) - 1;
    [n, p] = size(x.U);
    output = zeros(size(B));
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, :) = B((i - 1) * n + 1 : i * n, :) + (a * x.U) * x.U'* B((i - 1) * n + 1 : i * n, :);
    end
end

function output = BG05_quotient(x, B)
    a = 1 - 1 / sqrt(2);
    [n, p] = size(x.U);
    output = zeros(size(B));
    for i = 1 : p
        output(:, (i - 1) * n + 1 : i * n) = B(:, (i - 1) * n + 1 : i * n) - a * (B(:, (i - 1) * n + 1 : i * n) * x.U) * x.U';
    end
end

function output = BGn05_quotient(x, B)
    a = sqrt(2) - 1;
    [n, p] = size(x.U);
    output = zeros(size(B));
    for i = 1 : p
        output(:, (i - 1) * n + 1 : i * n) = B(:, (i - 1) * n + 1 : i * n) + a * (B(:, (i - 1) * n + 1 : i * n) * x.U) * x.U';
    end
end

function output = build_perp_bases_embedded(x)
    r2 = sqrt(2);
    [n, p] = size(x.U);
    output = sparse(n * p, p * (p + 1) / 2);
%     output = zeros(n * p, p * (p + 1) / 2);
    times = 1;
    for j = 1 : p
        st = (j - 1) * n + 1;
        en = j * n;
        output(st : en, times) = x.U(:, j);
        times = times + 1;
    end

    for i = 2 : p
        for j = i : p
            st = (j - i) * n + 1;
            en = (j - i + 1) * n;
            output(st : en, times) = x.U(:, j) / r2;
            st = (j - 1) * n + 1;
            en = j * n;
            output(st : en, times) = x.U(:, j - i + 1) / r2;
            times = times + 1;
        end
    end
end

function output = build_perp_bases_quotient(x)
    r2 = sqrt(2);
    [n, p] = size(x.U);
    output = sparse(n * p, p * (p + 1) / 2);
%     output = zeros(n * p, p * (p + 1) / 2);
    times = 1;
    for j = 1 : p
        st = (j - 1) * n + 1;
        en = j * n;
        output(st : en, times) = x.U(:, j) * r2;
        times = times + 1;
    end

    for i = 2 : p
        for j = i : p
            st = (j - i) * n + 1;
            en = (j - i + 1) * n;
            output(st : en, times) = x,U(:, j);
            st = (j - 1) * n + 1;
            en = j * n;
            output(st : en, times) = x.U(:, j - i + 1);
            times = times + 1;
        end
    end
end

function output = build_bases_embedded(x)
    [n, p] = size(x.U);
    if(~isfield(x, 'Up'))
        x.Up = null(x.U');
    end
    x_perp = x.Up;
    x.U = x.U / sqrt(2);
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
%     output = zeros(n * p, d);
    col = 1;
    for i = 2 : p
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x.U(:, j);
            output((j - 1) * n + 1 : j * n, col) = x.U(:, i - 1);
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp(1 : n, 1 : n - p);
        col = col + n - p;
    end
end

function output = build_bases_quotient(x)
    [n, p] = size(x.U);
    if(~isfield(x, 'Up'))
        x.Up = null(x.U');
    end
    x_perp = x.Up;
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
%     output = zeros(n * p, d);
    col = 1;
    for i = 2 : p
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x.U(:, j);
            output((j - 1) * n + 1 : j * n, col) = x.U(:, i - 1);
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp(1 : n, 1 : n - p);
        col = col + n - p;
    end
end

function dw = ode_parallel_vt_embedded(t, w, M1, M2, A, n, p)
    w = reshape(w, n, p);
    Temp1 = expm(M1 * t);
    Temp2 = expm(- t * A);
    r = M2 * Temp1(:, 1 : p) * Temp2;
    dr = M2 * Temp1(:, p + 1 : 2 * p) * Temp2;
    dw = reshape(- 0.5 * r * (dr' * w + w' * dr), n * p, 1);
end

function dw = ode_parallel_vt_quotient(t, w, M1, M2, A, n, p)
    w = reshape(w, n, p);
    Temp = expm(M1 * t);
    r = M2 * Temp(:, 1 : p);
    dr = M2 * (Temp * M1(:, 1 : p));
    dw = reshape(- 0.5 * (dr * w' + w * dr') * r - 0.5 * r * (dr' * (w - r * r' * w) + w' * (dr - r * r' * dr)), n * p, 1);
end

function [t, y] = runge_kutta1(fhandle, y0, h, a, b)
    n = floor((b - a) / h);
    y = zeros(length(y0), n + 1);
    t = zeros(n + 1, 1);
    t(1) = a;
    y(:, 1) = y0;
    for i = 1 : n
        t(i + 1) = t(i) + h;
        k1 = fhandle(t(i), y(:, i));
        k2 = fhandle(t(i) + h / 2, y(:, i) + h * k1 / 2);
        k3 = fhandle(t(i) + h / 2, y(:, i) + h * k2 / 2);
        k4 = fhandle(t(i) + h, y(:, i) + h * k3);
        y(:, i + 1) = y(:, i) + h * (k1 + 2 * k2 + 2 * k3 + k4) / 6;
    end
end

function output = embedded_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - sqrt(2) * omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = quotient_intr_c(x, eta)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = x1' * eta.TV; % omega
    K = x2' * eta.TV; % K
    omega = - omega;
    output.TV = zeros(n * p - 0.5 * p * (p + 1), 1);
    indx = find(tril(ones(size(omega)), -1));
    output.TV(1 : 0.5 * p * (p - 1)) = omega(indx);
%     output(1 : 0.5 * p * (p - 1)) = nonzeros(tril(omega, -1));
    output.TV(0.5 * p * (p - 1) + 1 : end) = reshape(K, [], 1);
end

function output = embedded_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    r2 = sqrt(2);
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1)) / r2;
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;
end

function output = quotient_full_c(x, v)
    x1 = x.U;
    x2 = x.Up;
    [n, p] = size(x1);
    omega = tril(ones(p, p), -1);
    indx = find(omega);
    omega(indx) = v.TV(1 : 0.5 * p * (p - 1));
    omega = omega' - omega;
    K = reshape(v.TV(0.5 * p * (p - 1) + 1 : end), (n - p), p);
    output.TV = x1 * omega + x2 * K;
end

function output = mynull(A)
    [m, n] = size(A);
    randn('state', 1);
    T = randn(n, n);
    T(:, 1 : m) = A';
    V = qf(T);
    output = V(:,m+1:n);
end
