function varargout = LRTR_SR1(fns,params)
% Solver of limited-memory of Riemannian trust region with symmetric rank-1 update method
% 
% INPUT:
% fns : a struct that contains required function handles
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.STV(a, v) : return a tangent vector which is a times v.
%     fns.TVaddTV(v1, v2) : return a tangent vector which is equal to v1 adds v2.
%     fns.proj(x, eta) : return P_x(eta) by projecting v to tangent space of x.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by inverse vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x1(d).
% for partly smooth function
%     fns.Tranv(x1, d, x2, v) : return a tangent vector on T_x2 M, which is given by vector transport that transport v \in T_x M to T_{R_x(d)} M. Here x2 = R_x(d).
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
% not required:
%     params.error [1e-5]       : tolerance of stopping criterion.
%     params.StopCriterion [3]  : stopping criterion, 0 means call fns.IsStopped to check
%                                 1 means stop when relative error of objective function is less than tolerance,
%                                 2 means stop when norm of gradient is less than tolerance,
%                                 3 means stop when norm of gradient over initial norm of gradient is less than tolerance,
%                                 4 means stopping criterion for partly smooth function.
%     params.m [4]              : number of s and y for storage
%     params.max_t [300]        : the maximum number of iterations.
%     params.debug [1]          : '0' means almost silence, '1' means some output. '2' means a lot of output, '3' means more than need to know.
%     params.c [0.1]            : acceptance or rejection constant
%     params.tau1 [0.25]        : shrink radius constant
%     params.tau2 [2]           : expansion radius constant
%     params.max_Delta [200]    : max of Delta
%     params.min_Delta [1e-10]   : min of Delta
%     params.err_x [1e-5]       : tolerance for checking whether iterates are close to convergence. It is between [0, 1].
%     params.num_Grads [10]     : number of gradients for checking whether iterates are close to Clark stationary point. It is between [0, inf].
%     params.theta[0.1]         : A stopping criterion parameter for solving local model
%     params.kappa[0.9]         : A stopping criterion parameter for solving local model
%     params.min_innit[0]       : the minimum number of iterations for solving local model
%     params.max_innit[200]     : the maxmum number of iterations for solving local model
% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
%         '4', grassmann manifold : Gr(p, n)
%
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates
%     varargout{2} : function values of all iterates
%     varargout{3} : norm of gradient of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : number of gradient evaluations
%     varargout{7} : accumulated computational time for iterations
%     varargout{8} : number of vector transport actions without parameters
%     varargout{9} : number of vector transport action with parameters
%     varargout{10} : number of retraction actions
%
% By Wen Huang, the solver of local model is a modified version of that in http://www.math.fsu.edu/~cbaker/GenRTR/
%     fprintf('LRTR_SR1\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'StopCriterion', 3, 'int', 0, 4);
    params = set_default_para(params, 'm', 4, 'int', 0, inf);
    params = set_default_para(params, 'error', 1e-6, 'float', 0, 1);
    params = set_default_para(params, 'err_x', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'num_Grads', 10, 'int', 0, inf);
    params = set_default_para(params, 'max_t', 300, 'int', 1, inf);
    params = set_default_para(params, 'debug', 1, 'int', 0, 3);
    params = set_default_para(params, 'c', 0.1, 'float', 0, 0.5);
    params = set_default_para(params, 'tau1', 0.25, 'float', 0.1, 0.9);
    params = set_default_para(params, 'tau2', 2, 'float', 1.5, 5);
    params = set_default_para(params, 'max_Delta', 200, 'float', 1, inf);
    params = set_default_para(params, 'min_Delta', 1e-10, 'float', 0, 1);
    params = set_default_para(params, 'theta', 0.1, 'float', 0, 1);
    params = set_default_para(params, 'kappa', 0.9, 'float', 0, 1);
    params = set_default_para(params, 'min_innit', 0, 'int', 0, inf);
    params = set_default_para(params, 'max_innit', 200, 'int', 0, inf);
    params = set_default_para(params, 'keepnumX', 2100000000, 'int', 1, inf);
    params = set_default_para(params, 'useRand', 0, 'int', 0, 1);
    params = set_default_para(params, 'isplot', 0, 'int', 0, 1);
    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.R = fs.R;
        fns.proj = fs.proj;
        fns.inpro = fs.inpro;
        fns.STV = fs.STV;
        fns.TVaddTV = fs.TVaddTV;
        fns.invTranv = fs.invTranv;
        fns.Tranv = fs.Tranv;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'STV');
    check_function_handle(fns, 'TVaddTV');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'proj');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'invTranv');
    check_function_handle(fns, 'Tranv');
    if(params.StopCriterion == 0)
        check_function_handle(fns, 'IsStopped');
    end
    % Check whether required parameters are legal
    if(~isfield(params, 'x0'))
        error('Invalid arguments: missing initial x0');
    end
    fns.haveprec = 0;
    fns.g = fns.inpro;
    % Initialization
    c = params.c;
    tau1 = params.tau1;
    tau2 = params.tau2;
    m = params.m;
    err = inf;
    err_v = inf;
    err_x = inf;
    times = 0;
    Delta = 1;
    x = params.x0;
    err_g = 1;
    err_r = 1;
    [fx, x] = fns.f(x);
    [gradf, x] = fns.gf(x);
    X{1} = x;
    F(1) = fx;
    G(1) = sqrt(fns.inpro(x, gradf, gradf));
    gamma = 1;
    start_index = 1;
    S = [];
    Y = [];
    SS = [];
    SY = [];
    nf = 1;
    ng = 1;
    nV = 0;
    nVp = 0;
    nR = 0;
    ind = 1;
    Grads = [];
    updateinfo = '';
    timeinfo = '';
%     fprintf('times : %d, f(x) : %e\n', times, fx);
    tic
    T(1) = 0;
    while (err > params.error && times < params.max_t)
        if(params.debug > 1)
            tCGtimestart = toc;
        end
        EMPTYTV = fns.STV(x, 0, gradf);
        [eta, inner_it, stop_tCG, timeinfo] = tCG_LSR1(fns, x, gradf, S, Y, SS, SY, start_index, gamma, EMPTYTV, Delta, times, params, timeinfo);
        if(params.debug > 1)
            timeinfo = sprintf('time, tCG:%.2e,', toc - tCGtimestart);
            getsytimestart = toc;
        end
        xtemp = fns.R(x, eta);
        [fxtemp, xtemp] = fns.f(xtemp);
        [gradftemp, xtemp] = fns.gf(xtemp);
%         rho = (fx - fxtemp) / (- fns.inpro(x, gradf, eta) - 0.5 * fns.inpro(x, eta, Hv_LSR1(fns, x, S, Y, SS, SY, v, gamma, start_index)));
        Hvtemp = Hv_LSR1(fns, x, S, Y, SS, SY, eta, gamma, start_index);
        rho = (fx - fxtemp) / (- fns.inpro(x, fns.TVaddTV(x, gradf,  fns.STV(x, 0.5, Hvtemp)), eta));%% - fns.inpro(x, eta,)
        %update B
        s = eta;
        [y, eta, x, xtemp] = fns.invTranv(x, eta, xtemp, gradftemp);
        if(params.debug > 2)
            cprintf([1,0,1], sprintf('\t\tCheck isometry of inverse vector transport: |gf| / |inv T gf|: %e/%e \n', ...
                sqrt(fns.inpro(xtemp, gradftemp, gradftemp)), sqrt(fns.inpro(x, y, y))));
        end
        nV = nV + 1;
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('get s y f gf:%.2e,', toc - getsytimestart)];
            updatesytimestart = toc;
        end
        y = fns.TVaddTV(x, y, fns.STV(x, -1, gradf));
        Temp = fns.TVaddTV(x, y, fns.STV(x, -1, Hvtemp));
        denorm = fns.inpro(x, s, Temp);
        if(abs(denorm) > sqrt(eps) * sqrt(fns.inpro(x, s, s) * fns.inpro(x, Temp, Temp)))
            gamma = fns.inpro(x, y, y) / fns.inpro(x, s, y);
            if(length(S) < m)
                S{times + 1} = s;
                Y{times + 1} = y;
            else
                S{start_index} = s;
                Y{start_index} = y;
                start_index = start_index + 1;
                if(start_index > m)
                    start_index = 1;
                end
            end
            [SS, SY] = updateSS_SY(fns, x, S, Y, SS, SY, start_index);
        else
            fprintf('warnning! denormator is close to zero.\n');
        end
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('update s y:%.2e,', toc - updatesytimestart)];
        end
        if(rho > 0.75)
            if(stop_tCG == 2 || stop_tCG == 1)
                Delta = tau2 * Delta;
            end
            if(Delta > params.max_Delta)
                fprintf('reach max of Delta\n');
                break;
            end
        elseif(rho < 0.1)
            Delta = tau1 * Delta;
            if(Delta < params.min_Delta)
                fprintf('reach min of Delta\n');
                break;
            end
        end
        if(rho > c || (abs((fx - fxtemp)/(abs(fx) + 1)) < sqrt(eps) && fxtemp < fx))
            if(params.debug > 1)
                transportsytimestart = toc;
            end
            if(params.debug > 2)
                cprintf([1,0,1], sprintf('\t\tCheck isometry of vector transport: |s| / |T s|: %e/', sqrt(fns.inpro(x, S{1}, S{1}))));
            end
            if(length(S) < m)
                for i = 1 : times + 1
                    [S{i}, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, S{i});
                    [Y{i}, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, Y{i});
                end
            else
                for i = 1 : m
                    [S{i}, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, S{i});
                    [Y{i}, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, Y{i});
                end
            end
            if(params.debug > 2)
                cprintf([1,0,1], sprintf('%e\n', sqrt(fns.inpro(xtemp, S{1}, S{1}))));
            end
            nVp = nVp + 2 * length(S);
            err_r = abs((fx - fxtemp) / fxtemp);
            if(params.debug > 1)
                timeinfo = [timeinfo sprintf('transport s y:%.2e,', toc - transportsytimestart)];
            end
            if(params.StopCriterion == 4)
                if(params.debug > 1)
                    checkstoppartlysmooth = toc;
                end
                err_x = sqrt(fns.inpro(x, eta, eta));% norm(x - xtemp);%%---Whether err_x is OK is still a problem.
                                                      %% The ideal form should be dist(x1, x2) which may be expensive.
                for i = 1 : size(Grads, 2)
                    if(i ~= ind)
                        Gradsi.TV = reshape(Grads(:, i), size(gradf.TV));
                        [tempv, eta, x, xtemp] = fns.Tranv(x, eta, xtemp, Gradsi);
                        Grads(:, i) = reshape(tempv.TV, [], 1);
                        nV = nV + 1;
                    end
                end
                if(err_x < params.err_x)
                    Grads(:, ind) = reshape(gradftemp.TV, [], 1);
                    ind = ind + 1;
                    if(ind > params.num_Grads)
                        ind = ind - params.num_Grads;
                    end
                    [w, err_vectorv] = qpsubprob(Grads, 1);
                    err_vector.TV = reshape(err_vectorv, size(gradf.TV));
                    err_v = sqrt(fns.inpro(x, err_vector, err_vector));
                end
                if(params.debug > 1)
                    timeinfo = [timeinfo sprintf('check stop:%.2e,', toc - checkstoppartlysmooth)];
                end
            end
            x = xtemp;
            gradf = gradftemp;
            fx = fxtemp;
            err_g = sqrt(fns.inpro(x, gradf, gradf));
            if(params.debug > 0)
                updateinfo = '';
            end
        else
            if(params.debug > 0)
                updateinfo = ('without updating x_k');
            end
        end
        times = times + 1;
        nf = nf + 1;
        ng = ng + 1;
        nR = nR + 1;
        X{times + 1} = x;
        if(times + 1 - params.keepnumX > 0)
            X{times + 1 - params.keepnumX} = [];
        end
        F(times + 1) = fx;
        G(times + 1) = err_g;
        T(times + 1) = toc;
        if(params.StopCriterion == 0)
            [output_error, params] = fns.IsStopped(X, F, G, T, params, fns);
            err = 1 - output_error;
%             err = 1 - fns.IsStopped(X, F, G, T, params, fns);
        elseif(params.StopCriterion == 1)
            err = err_r;
        elseif(params.StopCriterion == 2)
            err = err_g;
        elseif(params.StopCriterion == 3)
            err = err_g / G(1);
        elseif(params.StopCriterion == 4)
            err = err_v;
        end
        if(params.debug > 1)
            cprintf('k', ['\t' timeinfo '\n']);
        end
        if(params.debug > 0)
            if(params.StopCriterion ~= 4)
                if(isempty(updateinfo))
                    fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,radius:%.3e,time:%.2e,stoptCG:%d,in_it:%d,rho:%.2e\n', ...
                        times, fx, err_r, err_g, Delta, T(end), stop_tCG, inner_it, rho);
                else
                    fprintf(['i:%d, ' updateinfo ', radius reduces to %.3e\n'], times, Delta);
                end
            else
                if(isempty(updateinfo))
                    fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,dx:%.3e,dg:%.3e,radius:%.3e,time:%.2e,stoptCG:%d,in_it:%d,rho:%.2e\n', ...
                        times, fx, err_r, err_g, err_x, err_v, Delta, T(end), stop_tCG, inner_it, rho);
                else
                    fprintf(['i:%d, ' updateinfo ', radius reduces to %.3e\n'], times, Delta);
                end
            end
        end
    end
    timecost = toc;
    if(params.debug > 0)
        fprintf('Num. of Iter.: %d, time cost: %f seconds, Num. of Fun, Gra, VT and R: %d, %d, (%d+%d), %d \n', times, timecost, nf, ng, nV, nVp, nR)
        fprintf('f: %e, |gradf|: %e, |gradf/gradf0|: %e \n', F(end), G(end), G(end)/G(1))
    end
    if(params.isplot)
        figure(1);clf
        scatter(1 : length(F), F, '.');
        ylabel('f(x_i)');
        xlabel('iteration number');
        figure(2);clf
        semilogy(1 : length(G), G, '.');
        ylabel('|grad(x_i)|');
        xlabel('iteration number');
    end
    varargout{1} = X;
    varargout{2} = F;
    varargout{3} = G;
    varargout{4} = timecost;
    varargout{5} = nf;
    varargout{6} = ng;
    varargout{7} = T;
    varargout{8} = nV;
    varargout{9} = nVp;
    varargout{10} = nR;
    varargout{11} = params;
    
end

function [SSnew, SYnew] = updateSS_SY(fns, x, S, Y, SS, SY, start_index)
    n1 = length(S);
    n2 = size(SS, 1);
    SSnew = zeros(n1, n1);
    SYnew = zeros(n1, n1);
    if(n2 < n1)
        SSnew(1 : n2, 1 : n2) = SS;
        SYnew(1 : n2, 1 : n2) = SY;
        SSnew(n1, n1) = fns.inpro(x, S{n1}, S{n1});
        SYnew(n1, n1) = fns.inpro(x, S{n1}, Y{n1});
        for i = 1 : n2
            SYnew(n1, i) = fns.inpro(x, S{n1}, Y{i});
            SYnew(i, n1) = SYnew(n1, i);
            SSnew(n1, i) = fns.inpro(x, S{n1}, S{i});
            SSnew(i, n1) = SSnew(n1, i);
        end
    else
        %n1 = m in this case
        SSnew(1 : n1 - 1, 1 : n1 - 1) = SS(2 : n1, 2 : n1);
        SYnew(1 : n1 - 1, 1 : n1 - 1) = SY(2 : n1, 2 : n1);
        end_ind = start_index - 1;
        if(end_ind < 1)
            end_ind = end_ind + n1;
        end
        SSnew(n1, n1) = fns.inpro(x, S{end_ind}, S{end_ind});
        SYnew(n1, n1) = fns.inpro(x, S{end_ind}, Y{end_ind});
        for i = 1 : n1 - 1
            ind = i - 1 + start_index;
            if(ind > n1)
                ind = ind - n1;
            end
            SYnew(n1, i) = fns.inpro(x, S{end_ind}, Y{ind});
            SYnew(i, n1) = SYnew(n1, i);
            SSnew(n1, i) = fns.inpro(x, S{end_ind}, S{ind});
            SSnew(i, n1) = SSnew(n1, i);
        end
    end
end
