function varargout = MROM(fns, params)
% Solver of modified Riemannian optimization method

% INPUT
% fns: a struct that contains required function handles
% required:
%       fns.f(x) : return objective function value at x.
%       fns.gf(x) : return the gradient of local function at x.
%       fns.gf_full(x) : return the full gradient of function at x.
%       fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two
%       tangent vectors v1 and v2 on T_x M_r.
%       fns.inpro_full(x, v1, v2) : return the inner product g_x(v1, v2) of two
%       tangent vectors v1 and v2 on T_x M.
%       fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element 
%       on the manifold M_r and eta is a tangent vector of T_x M_r.
%       fns.R_r(x, eta) : return R_x(eta), where R_r is a rank-related retraction, 
%       x is an element on the manifold M and eta is a rank-related vector of T_x M_{<=r}.
%       fns.check_rank(x, tol) : return the numerical rank of x w.r.t.
%       the given tolerence.
%       fns.reduce_rank(x, r): truncate the rank of x to numerical rank r.
%       fns.resetHB(x0) : reset the parameters of H and B for the new
%       initial point x0.
%       fns.increase_r(x) : return the new rank point and rank-related
%       vector
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
%
%
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates
%     varargout{2} : function values of all iterates
%     varargout{3} : norm of gradient of all iterates
%     varargout{4} : total computational time
%     varargout{5} : different ranks in all iterates
%     varargout{6} : computational time when rank changes
%     varargout{7} : number of function evaluations
%     varargout{8} : number of function evaluations in line search when
%     change rank
%     varargout{9} : number of gradient evaluations
%     varargout{10} : number of retraction actions on fixd-rank manifold
%     varargout{11} : number of rank-related retraction actions
%     varargout{12} : number of hessian actions

    if nargin < 2,
        error('Invalid arguments: the number of arguments is invalid.');
    end
    
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'StopCriterion', 3, 'int', 0, 4);
    params = set_default_para(params, 'm', 4, 'int', 1, inf);
    params = set_default_para(params, 'restart', 1, 'int', 0, 1);
    params = set_default_para(params, 'error', 1e-6, 'float', 0, 1);
    params = set_default_para(params, 'err_x', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'num_Grads', 10, 'int', 0, inf);
    params = set_default_para(params, 'max_t', 300, 'int', 1, inf);
    params = set_default_para(params, 'debug', 1, 'int', 0, 3);
    params = set_default_para(params, 'alpha', 1e-2, 'float', 0, 0.5);
    params = set_default_para(params, 'beta', 0.99, 'float', 0, 1);
    params = set_default_para(params, 'ratio', 0.25, 'float', 0, 1);    
    params = set_default_para(params, 'minstepsize', 1e-10, 'float', 0, 0.1);
    params = set_default_para(params, 'maxstepsize', 200, 'int', 2, inf);
    params = set_default_para(params, 'keepnumX', 2100000000, 'int', 1, inf);
    params = set_default_para(params, 'isplot', 0, 'int', 0, 1);
    params = set_default_para(params, 'initstepsize', 1, 'float', 0, inf);   
    params = set_default_para(params, 'tol_rank', 1e-14, 'float', 0, inf);
    params = set_default_para(params, 'total_steps', 10, 'float', 1, inf);
    params = set_default_para(params, 'method', 4, 'int', 1, 8);
    params = set_default_para(params, 'epsilon_1', sqrt(3), 'float', 0, inf);
    params = set_default_para(params, 'epsilon_2', 1e-04, 'float', 0, inf);
    params = set_default_para(params, 'epsilon_3', 1, 'float', 0, inf);
    params = set_default_para(params, 'delta', 1e-08, 'float', 0, inf);
    params = set_default_para(params, 'tau11', 0.1, 'float', 0, inf);
    params = set_default_para(params, 'tau22', 0.1, 'float', 0, inf);
    
    % Initialization
    total_time = 0;
    nf_total = 0;
    ng_total = 0;
    nR_total = 0;
    nH_total = 0;
    nV_total = 0;
    nR_r_total = 0;
    nf_ls_total = 0;
    X_total = [];
    F_total = [];
    G_total = [];
    T_total = [];
 
    % check the initial rank
    if(~isfield(params, 'x0') || ~isfield(params, 'delta'))
        error('Invalid arguments: missing initial x0 or delta');
    end
    params.x0 = find_numerical_rank(params.x0, params.delta); 
    % 2020-09-14 The previous version
%     params.x0_New = find_numerical_rank(params.x0, params.delta); 
         
    [fgradx_full, ~] = fns.gf_full(params.x0);
    Gfull0 = sqrt(fns.inpro_full(params.x0, fgradx_full, fgradx_full));
    
    niter = 0;
    while (niter <= params.total_steps)
        if(params.method == 1)
            [X, F, G, time, nf, ng, T, nR, params] = RTR_SD(fns,params);
        elseif(params.method == 2)
            [X, F, G, time, nf, ng, nH, T, nV, nR, params] = RTR_SR1(fns,params);
            nH_total = nH_total + nH;
            nV_total = nV_total + nV;
        elseif(params.method == 3)
            [X, F, G, time, nf, ng, T, nV, nVp, nR, params] = LRTR_SR1(fns,params);
            nV_total = nV_total + nV;
        elseif (params.method == 4)
            [X, F, G, time, nf, ng, nH, T, nR, params] = RTR_Newton(fns,params);
            nH_total = nH_total + nH;
        elseif (params.method == 5)
            [X, F, G, time, nf, ng, nH, T, nV, nR, params] = RBFGS(fns,params);
            nH_total = nH_total + nH;
            nV_total = nV_total + nV;
        elseif (params.method == 6)
            [X, F, G, time, nf, ng, T, nV, nVp, nR, params] = LRBFGS(fns,params);
            nV_total = nV_total + nV;
        elseif(params.method == 7)
            [X, F, G, time, nf, ng, T, nV, nR, params] = RSD(fns, params); 
            nV_total = nV_total + nV;
        elseif(params.method == 8)
            [X, F, G, time, nf, ng, T, nV, nR, params] = RCG(fns, params); 
            nV_total = nV_total + nV;
        elseif(params.method == 9)
            [X, F, G, time, nf, ng, nH, T, nV, nR, params] = RNewton(fns, params); 
            nH_total = nH_total + nH;
            nV_total = nV_total + nV;
        end
        
        X_total = [X_total, X];
        F_total = [F_total, F];
        G_total = [G_total, G];
        if(isempty(T_total))
            T_total = [T_total, T];
        else
            lastT = T_total(end);
            T_total = [T_total, lastT + T];
        end
        nf_total = nf_total + nf;
        ng_total = ng_total + ng;
        nR_total = nR_total + nR;
        total_time = time + total_time;
        niter = niter + 1;
        
        tic
       	x = X{end};
        if(~exist('delta_ref'))
            delta_ref = F(1) - F(end);
            f_ref = F(1);
            r_ref = x.r;
        end
        
        % 11PA - Old version: 
        % if(~isfield(params, 'flag'))
        %     if(~isfield(x, 'r') || ~isfield(x, 'l'))
        %         [x.r, x.l] = fns.check_rank(x, params.delta);
        %         if (x.r < x.l)
        %             params.flag = 0;
        %         end
        %     elseif (G(end) < params.epsilon_3)
        %         params.flag = 1;
        %     end 
        % end
        
        % 11PA - New version:
        [x.r, x.l] = fns.check_rank(x, params.delta);
                
        if (x.r < x.l)
            params.flag = 0;
        else
            % In principle we should then have (G(end) < params.epsilon_3)
            params.flag = 1;
        end

        if (params.flag == 1)
            [fgradx_full, x] = fns.gf_full(x);
            Gfull = sqrt(fns.inpro_full(x, fgradx_full, fgradx_full));
            [fgradx, x] = fns.gf(x);
            Gend = sqrt(fns.inpro(x, fgradx, fgradx));
            
            % check whether to increase the rank
            norm_perp = sqrt(Gfull^2 + Gend^2 - 2 * fns.inpro_full(x, fgradx_full, fgradx));
            nR_r = 0;
            if (norm_perp > max(params.epsilon_1 * Gend, params.epsilon_2) && x.r < params.k_max)
                [xTemp, eta] = fns.increase_r(x);
                step_size = 1; %fns.inpro_full(xTemp, fgradx_full, eta) / fns.inpro(xTemp, eta, eta); 
                eta = fns.STV(x, -1, eta);
                xNew = fns.R_r(xTemp, eta);
                [fNew, xNew] = fns.f(xNew);
                slope = fns.inpro_full(xTemp, fgradx_full, eta);
                nR_r = nR_r + 1;
                nf_ls = 0;
                while (F(end) - fNew < - params.alpha * slope * step_size)
                    step_size = params.ratio * step_size;
                    eta = fns.STV(xTemp, step_size, eta);
                    xNew = fns.R_r(xTemp, eta);
                    [fNew, xNew] = fns.f(xNew);
                    slope = fns.inpro_full(xTemp, fgradx_full, eta);
                    nf_ls = nf_ls + 1;
                    if (step_size < params.minstepsize)
                        step_size = params.minstepsize;
                        xNew = fns.R(xTemp, fns.STV(xTemp, step_size, eta));
                        [fNew, xNew] = fns.f(xNew);
                        nR_r = nR_r + 1;
                        [gfNew, xNew] = fns.gf(xNew);
                        eta = fns.STV(xTemp, step_size, eta);
                        if(params.debug > 0)
                            fprintf('warning: linesearch reaches the min step size, it may need to be diminished \n')
                        end
                        return;
                    end
                end
                niter = niter + 1;
                if(params.debug >= 0)
                    if(xNew.r > x.r)
                        fprintf('increase the rank to %d\n', xNew.r);
                    end
                end
            	x = xNew;
                delta_ref = F(end) - fNew;
                f_ref = F(end);
                r_ref = x.r;
                nR_r_total = nR_r_total + nR_r;
                nf_ls_total = nf_ls_total + nf_ls;
                X_total = [X_total, x];
                F_total = [F_total, fns.f(x)];
                [fgradx, x] = fns.gf(x);
                G_total = [G_total, sqrt(fns.inpro(x, fgradx, fgradx))];
                time_increase = toc;
                T_total = [T_total, T_total(end) + time_increase];
            else
                while params.epsilon_3 > Gend
                    params.epsilon_3 = params.epsilon_3 * params.tau_epsilon;
                end
                if params.epsilon_3 < params.tol
                    break;
                end
            end

            params.x0 = x;
            if(params.method == 2 || params.method == 5)
                [params.H0, params.B0] = fns.resetHB(params.x0);
            end       
        elseif(params.flag == 0)
            xNew = fns.reduce_rank(x, x.r);
            fNew = fns.f(xNew);
            while (f_ref - fNew <= params.C_R * delta_ref)
                params.delta = params.tau_r * params.delta;
                xNew = find_numerical_rank(x, params.delta); 
             	fNew = fns.f(xNew);
            end
            x = xNew;
            niter = niter + 1;
            X_total = [X_total, x];
            F_total = [F_total, fns.f(x)];
            [fgradx, x] = fns.gf(x);
            G_total = [G_total, sqrt(fns.inpro(x, fgradx, fgradx))];
            time_decrease = toc;
            T_total = [T_total, T_total(end) + time_decrease];
            nf_total = nf_total + 1;
            ng_total = ng_total + 1; 
            if(params.debug >= 0)
                if (x.r < X{end}.l)
                    fprintf('reduce the rank to %d at time %s\n', x.r,T_total(end));
                end
            end
            
            % reset the parameters
            params.x0 = x;
            if(params.method == 2 || params.method == 5)
                [params.H0, params.B0] = fns.resetHB(params.x0);
            end         
        end
        time_rank = toc;
        total_time = total_time + time_rank;
        
        x.d
    end
            
    x = X_total{end};
	[fgradx, x] = fns.gf(x);
    Gend = sqrt(fns.inpro(x, fgradx, fgradx)); 
    
 
    
    fprintf('Total time is %f\n', total_time);
   	fprintf('(f: %e, |gradf|: %e, |gradr/gradf0|: %e \n', full(F(end)), full(Gend), full(Gend/Gfull0))
   	fprintf('Num. of Fun(+linesearch) and Gra and Hv and R and rank-related R: %d, %d, %d, %d, %d, %d \n', nf_total, nf_ls_total, ng_total, nH_total, nR_total, nR_r_total)
    varargout{1} = X_total;
    varargout{2} = F_total;
    varargout{3} = G_total;
    varargout{4} = total_time;
    varargout{5} = T_total;
    varargout{6} = nf_total;
    varargout{7} = nf_ls_total;
    varargout{8} = ng_total;
    varargout{9} = nR_total; 
    varargout{10} = nR_r_total;
    varargout{11} = nH_total; 
end
    

function output = find_numerical_rank(x, tol)
    if(~all(isfield(x, {'u', 'd', 'v'}) == 1))
        [x.u, x.d, x.v] = svd(x);
    end
    [md, nd] = size(x.d);
    l = min(md, nd);
    r = 1;
    for i = 2 : l
        if x.d(i,i) / x.d(1,1) > tol
            r = r + 1;
        end
    end 
    output.u = x.u(:, 1:r);
    output.d = x.d(1:r, 1:r);
    output.v = x.v(:, 1:r);
    output.r = r;
    output.l = r;
    output.delta = tol;
end