function varargout = RGS(fns,params)
% Solver of Riemannian gradient sampling method
% 
% INPUT:
% fns : a struct that contains required function handles
% required:
%     fns.f(x) : return objective function value at x.
%     fns.gf(x) : return the gradient of objection function at x.
%     fns.gf_exists(x) : return bool to show whether gradient exists or not.
% not required if params.manifold, params.retraction and params.vector_transport is provided.
%     fns.inpro(x, v1, v2) : return the inner product g_x(v1, v2) of two tangent vectors v1 and v2 on T_x M.
%     fns.R(x, eta) : return R_x(eta), where R is a retraction, x is an element on the manifold and eta is a tangent vector of T_x M.
%     fns.SRTV(x, m, seed) : return m random tangent vectors in T_x M, seed is the random seed.
%     fns.invTranv(x1, d, x2, v) : return a tangent vector on T_x1 M, which is given by vector transport that transport v \in T_x2 M to T_x1 M. Here x2 = R_x(d).
% 
% params : a struct that contains parameters that are used in the solver.
% required:
%     params.x0 : initial approximation of minimizer.
% not required:
%     params.error [1e-5]       : tolerance of stopping criterion.
%     params.StopCriterion [3]  : stopping criterion, 0 means call fns.IsStopped to check
%                                 1 means stop when relative error of objective function is less than tolerance,
%                                 2 means stop when norm of gradient is less than tolerance.
%                                 3 means stop when norm of gradient over initial norm of gradient is less than tolerance,
%                                 4 means stopping criterion is nothing with norm of gradient
%     params.max_t [300]        : the maximum number of iterations.
%     params.debug [1]          : '0' means almost silence, '1' means some output. '2' means a lot of output, '3' means more than need to know.
%     params.ns [20]             : the number of samples
%     params.epsilon [1e-10]    : sampling radius
%     params.nu [0]             : Initial optimality tolerance
%     params.theta [1]          : Optimality tolerance reduction factor
%     params.mu [1]             : Sampling radius reduction factor
%     params.gsbeta [0]         : Armijo parameter
%     params.gamma [0.5]        : Backtracking reduction factor
%     params.max_ls [40]        : maximum number of iteration for backtracking in linesearch
%     params.nu_err[1e-10]      : tolerance of optimality tolerance
%     params.nd_err[1e-10]      : tolerance for checking whether it is close to a Clark stationary point
% 
% The following give some examples of manifold. For each manifold, some retractions and vector transports are provided.
%     params.manifold : the manifold which objective function is on.
%         '1', sphere : S^{n - 1}
%         '2', Stiefel manifold : St(p, n)
%         '3', orthogonal group : O(n)
%         '4', grassmann manifold : Gr(p, n)
%
% OUTPUT:
% varargout : contains output parameters
%     varargout{1} : all iterates
%     varargout{2} : function values of all iterates
%     varargout{3} : norm of gradient of all iterates
%     varargout{4} : total computational time
%     varargout{5} : number of function evaluations
%     varargout{6} : number of gradient evaluations
%     varargout{7} : accumulated computational time for iterations
%     varargout{8} : number of vector transport actions
%     varargout{9} : number of retraction actions
% 
% By Wen Huang

    fprintf('RGS\n')
    if nargin < 2,
       error('Invalid arguments: the number of arguments is invalid.');
    end
    % set default parameters and check the legal range of parameters
    params = set_default_para(params, 'StopCriterion', 4, 'int', 0, 4);
    params = set_default_para(params, 'ns', 20, 'int', 2, inf);
    params = set_default_para(params, 'epsilon', 1e-10, 'float', 0, inf);
    params = set_default_para(params, 'nu', 0, 'float', 0, 1);
    params = set_default_para(params, 'nu_err', 1e-10, 'float', 0, 1);
    params = set_default_para(params, 'nd_err', 1e-10, 'float', 0, 1);
    params = set_default_para(params, 'theta', 1, 'float', 0, 1);
    params = set_default_para(params, 'mu', 1, 'float', 0, 1);
    params = set_default_para(params, 'gsbeta', 0, 'float', 0, 1);
    params = set_default_para(params, 'gamma', 0.5, 'float', 0, 1);
    params = set_default_para(params, 'error', 1e-5, 'float', 0, 1);
    params = set_default_para(params, 'max_t', 300, 'int', 1, inf);
    params = set_default_para(params, 'max_ls', 40, 'int', 1, inf);
    params = set_default_para(params, 'debug', 1, 'int', 0, 3);
    params = set_default_para(params, 'keepnumX', 2100000000, 'int', 1, inf);
    params = set_default_para(params, 'isplot', 0, 'int', 0, 1);

    if(isfield(params, 'manifold') && isfield(params, 'retraction') && isfield(params, 'vector_transport'))
        fs = choose_manifold(params.manifold, params.retraction, params.vector_transport);
        fns.R = fs.R;
        fns.SRTV = fs.SRTV;
        fns.inpro = fs.inpro;
        fns.invTranv = fs.invTranv;
    end
    % Check whether function handles are legal
    check_function_handle(fns, 'f');
    check_function_handle(fns, 'gf');
    check_function_handle(fns, 'gf_exists');
    check_function_handle(fns, 'R');
    check_function_handle(fns, 'SRTV');
    check_function_handle(fns, 'inpro');
    check_function_handle(fns, 'invTranv');
    if(params.StopCriterion == 0)
        check_function_handle(fns, 'IsStopped');
    end
    % Check whether required parameters are legal
    if(~isfield(params, 'x0'))
        error('Invalid arguments: missing initial x0');
    end
    % Initialization
    err = inf;
    times = 0;
    x1 = params.x0;
    [f1, x1] = fns.f(x1);
    f2 = f1;
    [g_all{1}, x1] = fns.gf(x1);
    X{1} = x1;
    F(1) = f1;
    G(1) = sqrt(fns.inpro(x1, g_all{1}, g_all{1}));
    nf = 1;
    ng = 1;
    nV = 0;
    nR = 0;
    m = params.ns;
    epsilon = params.epsilon;
    nu = params.nu;
    theta = params.theta;
    mu = params.mu;
    beta = params.gsbeta;
    gamma = params.gamma;
    max_ls = params.max_ls;
    timeinfo = '';
    fprintf('times : %d, f(x) : %e\n', times, f1);
    tic
    T(1) = 0;
    while(err > params.error && times < params.max_t)
        if(params.debug > 1)
            getgfstimestart = toc;
        end
        % compute gradient samples
        u = fns.SRTV(x1, m, times + 1);
        for i = 1 : m
            eu{i}.TV = epsilon * u{i}.TV;
            x_nearby{i} = fns.R(x1, eu{i});
            if(fns.gf_exists(x_nearby{i}))
                [g_nearby{i}, x_nearby{i}] = fns.gf(x_nearby{i});
            else
                fprintf('the gradient of a sample does not exist \n')
                break;
            end
        end
        if(params.debug > 1)
            timeinfo = sprintf('time, get gf samples:%.2e,', toc - getgfstimestart);
            invtransporttimestart = toc;
        end
        % make gradients in a same tangent space
        for i = 1 : m
            [g_all{i + 1}, eu{i}, x1, x_nearby{i}] = fns.invTranv(x1, eu{i}, x_nearby{i}, g_nearby{i});
        end
        if(params.debug > 2)
            cprintf([1,0,1], sprintf('\t\tCheck isometry of inverse vector transport: |gf| / |inv T gf|: %e/%e \n', ...
                sqrt(fns.inpro(x_nearby{1}, g_nearby{1}, g_nearby{1})), sqrt(fns.inpro(x1, g_all{2}, g_all{2}))));
        end
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('inverse transports:%.2e,', toc - invtransporttimestart)];
            convexprogrammingtimestart = toc;
        end
        nV = nV + m;
        nR = nR + m;
        ng = ng + m;
        % find direction
        for i = 1 : (m + 1)
            g_x(:, i) = reshape(g_all{i}.TV, [], 1);
        end
        [w, dv] = qpsubprob(g_x, 1);
        d.TV = reshape(dv, size(g_all{1}.TV));
        nd = sqrt(fns.inpro(x1, d, d));
        if(params.debug > 1)
            timeinfo = [timeinfo sprintf('convex prog:%.2e,', toc - convexprogrammingtimestart)];
        end
        if(nd < params.nd_err && nu < params.nu_err)
            fprintf('nd: %e, nu: %e, epsilon: %e. stop. \n', nd, nu, epsilon)
            break;
        end
        if(nd <= nu)
            t = 0;
            nu = theta * nu;
            epsilon = mu * epsilon;
            if(params.debug > 0)
                fprintf('reduction optimality tolerance to %e and sampling radius to %e. \n', nu, epsilon);
            end
        else
            if(params.debug > 1)
                linesearchtimestart = toc;
            end
            d.TV = d.TV / nd;
            % step 3
            [t, lnf, lng, lnR, f2] = linesearchgs(fns, x1, d, beta, gamma, nd, 1, max_ls, f1, params.debug);
            nf = nf + lnf;
            ng = ng + lng;
            nR = nR + lnR;
            if(params.debug > 1)
                timeinfo = [timeinfo sprintf('linesearch total:%.2e,', toc - linesearchtimestart)];
            end
        end
        td.TV = t * d.TV;
        x2 = fns.R(x1, td);
        nR = nR + 1;
        % we guarantee f is differentiable at x2 because the line search algorithm.
        % So step 4 is not necessary.
        
        % check stoping cretirion
        err_r = abs((f2 - f1) / f2);
        x1 = x2;
        f1 = f2;
        [g_all{1}, x1] = fns.gf(x1);
        ng = ng + 1;
        times = times + 1;
        X{times + 1} = x1;
        if(times + 1 - params.keepnumX > 0)
            X{times + 1 - params.keepnumX} = [];
        end
        F(times + 1) = f1;
        G(times + 1) = sqrt(fns.inpro(x1, g_all{1}, g_all{1}));
        T(times + 1) = toc;
        err_g = G(times + 1);
        if(params.StopCriterion == 0)
            err = 1 - fns.IsStopped(X, F, G, T, params, fns);
        elseif(params.StopCriterion == 1)
            err = err_r;
        elseif(params.StopCriterion == 2)
            err = err_g;
        elseif(params.StopCriterion == 3)
            err = err_g / G(1);
        end
        if(params.debug > 1)
            cprintf('k', ['\t' timeinfo '\n']);
        end
        if(params.debug > 0)
            fprintf('i:%d,f:%.3e,df/f:%.3e,|gf|:%.3e,nd:%.3e,nu:%.3e,stepsize:%.3e,time:%.2e\n', ...
                times, f2, err_r, err_g, nd, nu, t, T(end));
        end
    end
    timecost = toc;
    fprintf('Num. of Iter.: %d, time cost: %f seconds, Num. of Fun, Gra, VT and R: %d, %d, %d, %d \n', times, timecost, nf, ng, nV, nR)
    fprintf('f: %e, |g|: %e |g_f|/|g_0|: %e, nd: %e \n', F(end), G(end), G(end) / G(1), nd)
    if(params.isplot)
        figure(1);clf
        scatter(1 : length(F), F, '.');
        ylabel('f(x_i)');
        xlabel('iteration number');
        figure(2);clf
        semilogy(1 : length(G), G, '.');
        ylabel('|grad(x_i)|');
        xlabel('iteration number');
    end
    varargout{1} = X;
    varargout{2} = F;
    varargout{3} = G;
    varargout{4} = timecost;
    varargout{5} = nf;
    varargout{6} = ng;
    varargout{7} = T;
    varargout{8} = nV;
    varargout{9} = nR;
end

function [t, lnf, lng, lnR, f2] = linesearchgs(fns, x, d, beta, gamma, nd, t, max_ls, f1, debug)
    xnew = fns.R(x, d);
    times = 0;
    lnf = 0;
    lng = 0;
    lnR = 0;
    [f2, xnew] = fns.f(xnew);
    while(f2 >= f1 - beta * gamma * nd && times < max_ls)
        t = gamma * t;
        [f2, xnew] = fns.f(xnew);
        td.TV = t * d.TV;
        xnew = fns.R(x, td);
        times = times + 1;
    end
    lnf = lnf + times + 1;
    lnR = lnR + times + 1;
    while(~fns.gf_exists(xnew))
        t = t * 1.001;
        td.TV = t * d.TV;
        xnew = fns.R(x, td);
        lnR = lnR + 1;
    end
    if(times >= max_ls && debug > 0)
        fprintf('warning: linearch reaches max iterations, %e \n', t);
%         t = 0;
%         f2 = f1;
    end
end
